/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* DHCP client program */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>
#define	KERNEL	/* suppress a warning message */
#include <netkey/key.h>
#include <netinet/if_mvif.h>
#undef	KERNEL
#include <arpa/inet.h>
#include <sys/time.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "mnconfig.h"
#include "mnstate.h"
#include "dhcp.h"
#ifdef IPSEC
#include "ipsec.h"
#include "route.h"
#endif

#define	BROADCAST	1
#define	UNICAST		0
#define	TRUE		1
#define	FALSE		0

#define DHCPS_PORT	67
#define DHCPC_PORT	68

#define DHCP_STATE_INIT		0
#define DHCP_STATE_SELECTING	1
#define DHCP_STATE_REQUESTING	2
#define DHCP_STATE_BOUND	3
#define DHCP_STATE_RENEWING	4
#define DHCP_STATE_REBINDING	5

#define DHCP_DEFAULT_RENEWING_TIME	10
	/* time interval lease be renewed. a short renewing time allows for
	   effective movement detection. client includes lease suggestions
	   in DHCPDISCOVER and DHCPREQUEST to request a short lease */
#define DHCP_DEFAULT_SELECTING_TIME	5
	/* time to wait for and select OFFERs. the best to select servers
	   according to signal strength. select first server in this impl. */
#define DHCP_DEFAULT_RETRY_LIMIT	3

int backoff[] = { 4, 8, 16 };

char * dhcp_message_types[] = { "BOOTP", "DHCPDISCOVER", "DHCPOFFER", 
				"DHCPREQUEST", "DHCPDECLINE", "DHCPACK", 
				"DHCPNAK", "DHCPRELEASE", "DHCPINFORM" };
struct option {
	u_char type;
	u_char length;
	u_char value[255];
};

extern struct mn_config mn_config;
extern int udpSocket;

struct dhcp_packet packet, pkt;
int dhcpSocket;
int dhcpState;
int lastxid, serverxid;
struct in_addr client, subnet_mask, router, server;
int T1, T2, lease_time, msgtype;
int selecting_timer, retry_timer, retry_count, T1_timer, T2_timer, lease_timer;

void send_discover();
void send_request(int, int);
void error(char * fmt, ...);
void info(char *fmt, ...);
extern void mnStateMachine(int, struct alist *, struct in_addr);

initDhcpSocket() 
{
	struct sockaddr_in name;
	struct timeval tv;
	int flag;

	gettimeofday(&tv, NULL);
	srandom(tv.tv_sec);

	dhcpSocket = socket(AF_INET, SOCK_DGRAM, 0);
	if (dhcpSocket < 0) {
		error("socket");
	}

	flag = 1;
	if (setsockopt(dhcpSocket, SOL_SOCKET, SO_BROADCAST, 
					(char *)&flag, sizeof(flag)) < 0) {
		error("setsockopt");
	}

	name.sin_family = AF_INET;
	name.sin_addr.s_addr = INADDR_ANY;
	name.sin_port = htons(DHCPC_PORT);
	name.sin_len = sizeof(name);
	bzero(name.sin_zero, sizeof(name.sin_zero));
	
	if (bind(dhcpSocket, (struct sockaddr *)&name, sizeof(name)) < 0) {
		error("bind");
		exit(1);
	}
}

/* DHCP State Machine
   if event, then the state machine is driven by an incoming message,
   otherwise, the state machine is driven by mnStateTick */

dhcpStateMachine(int event)
{
	struct sockaddr_in from;
	int fromlen;

	if (mns.state != NOWHERE && mns.state != CO_LOCATED) {
		/* this if clause should never be executed because we 
		   are in AGENT_MODE when in ATHOME or AWAY state */
		if (mns.mode == DHCP_MODE) {
		    	mns.mode = AGENT_MODE;
			network_config(FALSE);
		}
		return;
	}
	else {
		if (mns.mode == AGENT_MODE) {
			mns.mode = DHCP_MODE;
			dhcpState = DHCP_STATE_INIT;
		}
	}

	/* if message pending, receive the message */

	if (event) {
		bzero(&pkt, sizeof(pkt));
		if (recvfrom(dhcpSocket, &pkt, sizeof(pkt), 0,
				(struct sockaddr *)&from, &fromlen) < 0)
			error("recvfrom");

	        msgtype = dhcp_msg_type(&pkt);

		/* if not a DHCP message (BOOTP message) or Transaction_ID
		   mismatch, ignore the message */

	        if (msgtype == 0 || ntohl(pkt.xid) != lastxid) return;
	}

	switch (dhcpState) {
	    case DHCP_STATE_INIT :
		if (event) {
		} /* silently discard messages received in INIT state */
		send_discover();
#ifdef	DEBUG
		info("leaving INIT state, broadcasted DISCOVER message\n");
#endif
		dhcpState = DHCP_STATE_SELECTING;
		selecting_timer = DHCP_DEFAULT_SELECTING_TIME;
		break;
	    case DHCP_STATE_SELECTING :
		if (event) {
		    if (msgtype == DHCPOFFER) {
#ifdef	DEBUG
			info("got OFFER from %s\n", inet_ntoa(from.sin_addr));
#endif
		        parse_packet(&pkt);
#ifdef	DEBUG
			info("IP = %s\n", inet_ntoa(client));
			info("Subnet_mask = %s\n", inet_ntoa(subnet_mask));
			info("First router = %s\n", inet_ntoa(router));
			info("Lease time = %d\n", lease_time);
			info("T1 = %d\n", T1);
			info("T2 = %d\n", T2);
#endif
			send_request(BROADCAST, TRUE);
#ifdef	DEBUG
			info("broadcasted REQUEST, entering "
							"REQUESTING state\n");
#endif
			dhcpState = DHCP_STATE_REQUESTING;
			retry_count = 0;
			retry_timer = backoff[retry_count];
		    }
		}
		else {
		    selecting_timer--;
		    if (selecting_timer <= 0) 
			dhcpState = DHCP_STATE_INIT;
		}
		break;
	    case DHCP_STATE_REQUESTING :
		if (event) {
		    if (msgtype == DHCPACK) {
#ifdef	DEBUG
			info("got ACK from %s, entering BOUND state\n", 
						inet_ntoa(from.sin_addr));
#endif
			parse_packet(&pkt);
			network_config(TRUE);
			dhcpState = DHCP_STATE_BOUND;
			T1_timer = T1;
			T2_timer = T2;
			lease_timer = lease_time;
			/* change state to CO_LOCATED and send out a reg. */
			mnStateMachine(udpSocket, NULL, client);
		    }
		    else if (msgtype == DHCPNAK) {
#ifdef	DEBUG
			info("got NAK from %s, go back to INIT state\n", 
						inet_ntoa(from.sin_addr));
#endif
			dhcpState = DHCP_STATE_INIT;
		    }
		}
		else {
		    retry_timer--;
		    if (retry_timer <= 0) {
			if (retry_count < DHCP_DEFAULT_RETRY_LIMIT-1) {
			    send_request(BROADCAST, TRUE);
#ifdef	DEBUG
			    info("time out, retrying...\n");
#endif
			    retry_count++;
			    retry_timer = backoff[retry_count];
			}
			else {
			    dhcpState = DHCP_STATE_INIT;
			}
		    }
		}
		break;
	    case DHCP_STATE_BOUND :
		if (event) {
		} /* silently discard messages received in BOUND state */
		else {
		    T1_timer--;
		    T2_timer--;
		    lease_timer--;
		    if (T1_timer <= 0) {
			send_request(UNICAST, FALSE);
#ifdef	DEBUG
		 	info("T1 expired, sent REQUEST, "
						"entering RENEWING state\n");
#endif
			dhcpState = DHCP_STATE_RENEWING;
			retry_count = 0;
			retry_timer = (T2-T1)/2;
			/* don't retry if retry_timer is too small */
			if (retry_timer < 3) retry_count++;
		    }
		}
		break;
	    case DHCP_STATE_RENEWING :
		if (event) {
		    if (msgtype == DHCPACK) {
#ifdef	DEBUG
			info("got ACK from %s, entering BOUND state\n", 
						inet_ntoa(from.sin_addr));
#endif
			parse_packet(&pkt);
			/* network parameters should be the same */
			/* network_config(TRUE); */
			dhcpState = DHCP_STATE_BOUND;
			T1_timer = T1;
			T2_timer = T2;
			lease_timer = lease_time;
		    }
		    else if (msgtype == DHCPNAK) {
#ifdef	DEBUG
			info("got NAK from %s, go back to INIT state\n", 
						inet_ntoa(from.sin_addr));
#endif			/*
			network_config(FALSE);
			dhcpState = DHCP_STATE_INIT; */
			enterNowhereState(mns.state, FALSE);
		    }
		}
		else {
		    T2_timer--;
		    lease_timer--;
		    if (!retry_count) {
		        retry_timer--;
		        if (retry_timer <= 0) {
			    send_request(UNICAST, FALSE);
#ifdef	DEBUG
			    info("time out, retrying...\n");
#endif
			    retry_count++;
			}
		    }
		    if (T2_timer <= 0) {
			/* enter REBINDING state only if plenty of time left */
			if (lease_timer >= 4) {
			    send_request(BROADCAST, FALSE);
#ifdef	DEBUG
			    info("T2 expired, broadcasted REQUEST, entering"
							" REBINDING state\n");
#endif
			    dhcpState = DHCP_STATE_REBINDING;
			    retry_count = 0;
			    retry_timer = (lease_time - T2) / 2;
			    /* don't retry if retry_timer is too small */
			    if (retry_timer < 3) retry_count++;
			}
			else {
#ifdef	DEBUG
			    info("T2 expired, go back to INIT state "
								"directly\n");
#endif			    /*
			    network_config(FALSE);
			    dhcpState = DHCP_STATE_INIT; */
			    enterNowhereState(mns.state, FALSE);
			}
		    }
		}
		break;
	    case DHCP_STATE_REBINDING :
		if (event) {
		    if (msgtype == DHCPACK) {
#ifdef	DEBUG
			info("got ACK from %s, entering BOUND state\n", 
						inet_ntoa(from.sin_addr));
#endif
			parse_packet(&pkt);
			/* network parameters should be the same */
			/* network_config(TRUE); */
			dhcpState = DHCP_STATE_BOUND;
			T1_timer = T1;
			T2_timer = T2;
			lease_timer = lease_time;
		    }
		    else if (msgtype == DHCPNAK) {
#ifdef	DEBUG
			info("got NAK from %s, go back to INIT state\n",
						inet_ntoa(from.sin_addr));
#endif			/*
			network_config(FALSE);
			dhcpState = DHCP_STATE_INIT; */
			enterNowhereState(mns.state, FALSE);
		    }
		}
		else {
		    lease_timer--;
		    if (!retry_count) {
		        retry_timer--;
		        if (retry_timer <= 0) {
			    send_request(BROADCAST, FALSE);
#ifdef	DEBUG
			    info("time out, retrying\n");
#endif
			    retry_count++;
			}
		    }
		    if (lease_timer <= 0) {
#ifdef	DEBUG
			info("lease expired, go back to INIT state\n");
#endif			/*
			network_config(FALSE);
			dhcpState = DHCP_STATE_INIT; */
			enterNowhereState(mns.state, FALSE);
		    }
		}
		break;
	    default :
		error("this should not happen, DHCP entered an unknown state");
		break;
	}
}

/* DHCP state machine enters INIT state */

dhcpEnterInitState() {
	if (dhcpState == DHCP_STATE_BOUND || dhcpState == DHCP_STATE_RENEWING
			|| dhcpState == DHCP_STATE_REBINDING)
		network_config(FALSE);
	dhcpState = DHCP_STATE_INIT;
}

/* ifconfig: configure the interface with an address 
   if address is zero, delete the address */
      
myifconfig(char * ifname, struct in_addr ip, struct in_addr mask)
{       
        struct ifreq ifr;
        struct ifaliasreq ifra;
        struct sockaddr_in * ifa; 
 
	/* get address of the interface */
       
        strcpy(ifr.ifr_name, ifname);
        if (ioctl(dhcpSocket, SIOCGIFADDR, &ifr) < 0) {
		if (errno != EADDRNOTAVAIL) 
                	error("ioctl: SIOCGIFADDR: %s", strerror(errno));
        }
	else errno = 0;

        /* delete old address first */

        if (!errno && ioctl(dhcpSocket, SIOCDIFADDR, &ifr) < 0) {
                error("ioctl: SIOCDIFADDR");
        }

        /* add new address */

	if (ip.s_addr) {
        	bzero(&ifra, sizeof(ifra));
        	strcpy(ifra.ifra_name, ifname);
        	ifa = (struct sockaddr_in *)(&ifra.ifra_addr);
        	ifa->sin_family = AF_INET;
        	ifa->sin_addr.s_addr = ip.s_addr;
        	ifa->sin_len = sizeof(struct sockaddr_in);
        	ifa = (struct sockaddr_in *)(&ifra.ifra_mask);
        	ifa->sin_family = AF_INET;
        	ifa->sin_addr.s_addr = mask.s_addr;
        	ifa->sin_len = sizeof(struct sockaddr_in);
        	ifa = (struct sockaddr_in *)(&ifra.ifra_broadaddr);
        	ifa->sin_family = AF_INET;
        	ifa->sin_addr.s_addr = ip.s_addr | ~mask.s_addr;
        	ifa->sin_len = sizeof(struct sockaddr_in);
        	if (ioctl(dhcpSocket, SIOCAIFADDR, &ifra) < 0) {
                	error("ioctl: SIOCAIFADDR");
        	}
	}
}

/* get the IP address of interface ifname */

void getifaddr(char * ifname, struct in_addr * addr)
{
	struct ifreq ifr;
        struct sockaddr_in * ifa; 

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCGIFADDR, &ifr) < 0) {
		if (errno != EADDRNOTAVAIL) 
			error("ioctl: SIOCGIFADDR: %s", strerror(errno));
	}
	else errno = 0;
	ifa = (struct sockaddr_in *)(&ifr.ifr_addr);
	if (!errno)
		addr->s_addr = ifa->sin_addr.s_addr;
	else
		addr->s_addr = 0;
}

/* get the subnet mask of interface ifname */

void getifmask(char * ifname, struct in_addr * mask)
{
	struct ifreq ifr;
        struct sockaddr_in * ifa; 

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCGIFNETMASK, &ifr) < 0) {
		if (errno != EADDRNOTAVAIL)
			error("ioctl: SIOCGIFNETMASK");
	}
	else errno = 0;
	ifa = (struct sockaddr_in *)(&ifr.ifr_addr);
	if (!errno)
		mask->s_addr = ifa->sin_addr.s_addr;
	else
		mask->s_addr = 0;
}

getInnerIP(char * ifname)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCGMVINIP, &ifr) < 0) {
		error("ioctl: SIOCGMVINIP");
	}
	return((int)ifr.ifr_data);
}

getOuterIP(char * ifname)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCSMVOUTIP, &ifr) < 0) {
		error("ioctl: SIOCSMVOUTIP");
	}
	return((int)ifr.ifr_data);
}

void setInnerIP(char * ifname, struct in_addr inip)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	ifr.ifr_data = (caddr_t) inip.s_addr;
	if (ioctl(dhcpSocket, SIOCSMVINIP, &ifr) < 0) {
		if (errno != EINVAL)
			error("ioctl: SIOCSMVINIP");
		else
			errno = 0;
	}
}

void setOuterIP(char * ifname, struct in_addr outip)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	ifr.ifr_data = (caddr_t) outip.s_addr;
	if (ioctl(dhcpSocket, SIOCSMVOUTIP, &ifr) < 0) {
		error("ioctl: SIOCSMVOUTIP");
	}
}

/* variables used to save the old address of mvif interface */

static struct in_addr mvif_old_addr, mvif_old_mask;
static struct in_addr mvif_old_inner_ip, mvif_old_outer_ip;

/* config the network to be used by DHCP_MODE and AGENT_MODE */

network_config(int configured)
{
    char command[256], temp[20];
    struct in_addr zero;
#ifdef IPSEC
    struct keylist	*key;
    int			sectype;
    int			useRouteSA;
#endif

    if (configured) {

	/* configure wlp0 with temporary address */

	myifconfig(mn_config.ifname, client, subnet_mask);

	/* configure mvif with home address */

	getifaddr(mn_config.mvifname, &mvif_old_addr);
	getifmask(mn_config.mvifname, &mvif_old_mask);
	myifconfig(mn_config.mvifname, mn_config.mn, mn_config.netmask);

	/* set the inner source address of mvif */

	mvif_old_inner_ip.s_addr = getInnerIP(mn_config.mvifname);
	setInnerIP(mn_config.mvifname, mn_config.mn);

	/* set the outer source address of mvif */

	mvif_old_outer_ip.s_addr = getOuterIP(mn_config.mvifname);
	setOuterIP(mn_config.mvifname, client);

	/* set the default route as a tunneled route to home agent */
	/*
	sprintf(command, "/sbin/route add default %s -tunnel > /dev/null", 
					inet_ntoa(mn_config.ha));
	system(command); */

#ifdef IPSEC
	zero.s_addr = 0;
	sectype = MN_SECTYPE_BYPASSFA;
	useRouteSA = (mn_config.ipsec[sectype].inuse
		&& (key = search_keylist(mn_config.ipsec[sectype].kl,
			mn_config.ha.s_addr)) && key->inuse);
	if ((useRouteSA && !mns.rosa.transform) ||
				(!useRouteSA && mns.rosa.transform))
	{
		if (useRouteSA) {
			set_rosa(TRUE, key, mn_config.mn, &mns.rosa,
				0, mn_config.ha);
		}
		else {
			set_rosa(FALSE, NULL, zero, &mns.rosa,
				0, mn_config.ha);
		}
	}
	mn_route_add_tunnel(mn_config.ha, &mns.rosa);
#else
	mn_route_add_tunnel(mn_config.ha, NULL);
#endif 

	/* add an auxilliary route to temporary router */
	/*
	strcpy(temp, inet_ntoa(mn_config.ha));
	sprintf(command, "/sbin/route add %s %s > /dev/null",
					temp, inet_ntoa(router));
	system(command); */

	route_add_host(mn_config.ha, router, client, NULL);
    }
    else {

	/* configure mvif with old address */

	myifconfig(mn_config.mvifname, mvif_old_addr, mvif_old_mask);

	/* configure wlp0 with home address */

	myifconfig(mn_config.ifname, mn_config.mn, mn_config.netmask);

	/* set the inner source address of mvif */

	setInnerIP(mn_config.mvifname, mvif_old_inner_ip);

	/* set the outer source address of mvif */

	setOuterIP(mn_config.mvifname, mvif_old_outer_ip);

	/* delete default route */
	/*
	sprintf(command, "/sbin/route delete default > /dev/null");
	system(command); */

	mn_route_delete_default(mn_config.ha, FALSE);

	/* delete the auxilliary route to temporary router */
	/*
	sprintf(command, "/sbin/route delete %s > /dev/null", 
					inet_ntoa(mn_config.ha));
	system(command); */

	route_delete_host(mn_config.ha, router);
    }
}

/* get the broadcast address of interface */

u_long get_broadcast_address()
{
	struct ifreq ifr;
	struct sockaddr_in * ifa;

	strcpy(ifr.ifr_name, mn_config.ifname);
	bzero(&ifr.ifr_addr, sizeof(struct sockaddr));
	if (ioctl(dhcpSocket, SIOCGIFBRDADDR, &ifr) < 0) {
		error("ioctl");
	}
	ifa = (struct sockaddr_in *)(&ifr.ifr_addr);
	return(ifa->sin_addr.s_addr);
}

/* broadcast a DHCP discover message */

void send_discover()
{
	struct sockaddr_in to;
	struct option * opt;
	int optlen, packetlen, rc;

	/* setup address */

	to.sin_family = AF_INET;
	to.sin_addr.s_addr = get_broadcast_address();
	to.sin_port = htons(DHCPS_PORT);
	to.sin_len = sizeof(to);
	bzero(to.sin_zero, sizeof(to.sin_zero));

	bzero(&packet, sizeof(packet));

	/* fillin BOOTP parameters */

	packet.op = BOOTREQUEST;
	packet.htype = HTYPE_ETHER;
	packet.hlen = 6;
	packet.hops = 0;
	lastxid = random();	/* generate a random transaction id */
	packet.xid = htonl(lastxid);
	packet.secs = 0;
	packet.flags = 0;
	packet.ciaddr.s_addr = 0;
	packet.yiaddr.s_addr = 0;
	packet.siaddr.s_addr = 0;
	packet.giaddr.s_addr = 0;
	bcopy(mn_config.mac, packet.chaddr, 6);

	/* put in COOKIE */

	bcopy(DHCP_OPTIONS_COOKIE, packet.options, 4);
	optlen = 4;
	opt = (struct option *)(packet.options + optlen);

	/* put in message type */

	opt->type = DHO_DHCP_MESSAGE_TYPE;
	opt->length = 1;
	opt->value[0] = DHCPDISCOVER;
	optlen += 3;
	opt = (struct option *)(packet.options + optlen);

	/* tell the server we are interested in these parameters */

	opt->type = DHO_DHCP_PARAMETER_REQUEST_LIST;
	opt->length = 2;
	opt->value[0] = DHO_SUBNET_MASK;
	opt->value[1] = DHO_ROUTERS;
	optlen += 4;
	opt = (struct option *)(packet.options + optlen);

	/* suggest the server with a lease time suitable for 
	   movement detection */

	if (mn_config.dhcp_renewing_period > 0) {
		opt->type = DHO_DHCP_LEASE_TIME;
		opt->length = 4;
		*((u_long *)opt->value) = 
				htonl(mn_config.dhcp_renewing_period * 2);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* mark the end of options */

	opt->type = DHO_END;
	optlen += 1;

	packetlen = DHCP_FIXED_LEN + optlen;

	/* send the message */

	rc = sendto(dhcpSocket, &packet, packetlen, 0, 
					(struct sockaddr *)&to, sizeof(to));

	if (rc < 0) {
		error("Can not send discover message");
	}
}

/* send a DHCPREQUEST message to server
   if broadcast, the message will be broadcasted
   if requesting, the message is sent in REQUESTING state */

void send_request(int broadcast, int requesting)
{
	struct sockaddr_in to;
	struct option * opt;
	int optlen, packetlen, rc;

	/* setup address */

	to.sin_family = AF_INET;
	if (broadcast)
		to.sin_addr.s_addr = get_broadcast_address();
	else
		to.sin_addr.s_addr = server.s_addr;
	to.sin_port = htons(DHCPS_PORT);
	to.sin_len = sizeof(to);
	bzero(to.sin_zero, sizeof(to.sin_zero));

	bzero(&packet, sizeof(packet));

	/* fillin BOOTP parameters */

	packet.op = BOOTREQUEST;
	packet.htype = HTYPE_ETHER;
	packet.hlen = 6;
	packet.hops = 0;
	lastxid = serverxid;	/* according to RFC, use 'xid' from server
				   DHCPOFFER message */
	packet.xid = htonl(serverxid);
	packet.secs = 0;
	packet.flags = 0;
	if (requesting)		/* use client IP if we are configured */
		packet.ciaddr.s_addr = 0;
	else
		bcopy(&client, &packet.ciaddr, sizeof(client));
	packet.yiaddr.s_addr = 0;
	packet.siaddr.s_addr = 0;
	packet.giaddr.s_addr = 0;
	bcopy(mn_config.mac, packet.chaddr, 6);

	/* put in COOKIE */

	bcopy(DHCP_OPTIONS_COOKIE, packet.options, 4);
	optlen = 4;
	opt = (struct option *)(packet.options + optlen);

	/* put in message type */

	opt->type = DHO_DHCP_MESSAGE_TYPE;
	opt->length = 1;
	opt->value[0] = DHCPREQUEST;
	optlen += 3;
	opt = (struct option *)(packet.options + optlen);

	/* include requested address only if we are requesting an 
	   initial lease, no requested address if extending lease */

	if (requesting) {
		opt->type = DHO_DHCP_REQUESTED_ADDRESS;
		opt->length = 4;
		bcopy(&client, opt->value, 4);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* include server ID to indicate the server being selected */

	if (requesting) {
		opt->type = DHO_DHCP_SERVER_IDENTIFIER;
		opt->length = 4;
		bcopy(&server, opt->value, 4);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* tell the server we are interested in these parameters */

	opt->type = DHO_DHCP_PARAMETER_REQUEST_LIST;
	opt->length = 2;
	opt->value[0] = DHO_SUBNET_MASK;
	opt->value[1] = DHO_ROUTERS;
	optlen += 4;
	opt = (struct option *)(packet.options + optlen);

	/* suggest the server with a lease time suitable for 
	   movement detection */

	if (mn_config.dhcp_renewing_period > 0) {
		opt->type = DHO_DHCP_LEASE_TIME;
		opt->length = 4;
		*((u_long *)opt->value) = 
				htonl(mn_config.dhcp_renewing_period * 2);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* mark the end of options */

	opt->type = DHO_END;
	optlen += 1;

	packetlen = DHCP_FIXED_LEN + optlen;

	/* send the message */

	rc = sendto(dhcpSocket, &packet, packetlen, 0, 
					(struct sockaddr *)&to, sizeof(to));

	if (rc < 0) {
		error("Can not send request message");
	}
}

/* parse an incoming DHCP message
   TBD: return a value indicating result of parsing (could have problems)
	initialize those values to be overwritten by parser so that upon
	return, calling function can examine what options are available,
	avoiding erronously using old values not setup in this call.
	Should also deal with the situation when "sname" and "file"
	fields are overloaded */

parse_packet(struct dhcp_packet * pkt)
{
	struct option * opt;

	serverxid = ntohl(pkt->xid);

	bcopy(&(pkt->yiaddr), &client, sizeof(client));

	opt = (struct option *)(pkt->options + 4);

	while(opt->type != DHO_END) {
	    switch (opt->type) {
		case DHO_PAD : break;
		case DHO_SUBNET_MASK :
			bcopy(opt->value, &subnet_mask, sizeof(subnet_mask));
			break;
		case DHO_ROUTERS :
			/* use the first router */
			bcopy(opt->value, &router, sizeof(router));
			break;
		case DHO_DHCP_MESSAGE_TYPE :
			msgtype = opt->value[0];
			break;
		case DHO_DHCP_SERVER_IDENTIFIER :
			bcopy(opt->value, &server, sizeof(server));
			break;
		case DHO_DHCP_LEASE_TIME :
			lease_time = ntohl(*((u_long *)opt->value));
			break;
		case DHO_DHCP_RENEWAL_TIME :
			T1 = ntohl(*((u_long *)opt->value));
			break;
		case DHO_DHCP_REBINDING_TIME :
			T2 = ntohl(*((u_long *)opt->value));
			break;
		default :
			/* skip other options */
			break;
	    }
	    if (opt->type != DHO_PAD)
		opt = (struct option *)((char *)opt + opt->length + 2);
	    else
		opt = (struct option *)((char *)opt + 1);
	}
}

/* determine the message type of current incoming message */

dhcp_msg_type(struct dhcp_packet * pkt)
{
	struct option * opt;

	opt = (struct option *)(pkt->options + 4);
	while (opt->type != DHO_END) {
		if (opt->type == DHO_DHCP_MESSAGE_TYPE)
			return(opt->value[0]);
		if (opt->type != DHO_PAD)
			opt = (struct option *)((char *)opt + opt->length + 2);
		else
			opt = (struct option *)((char *)opt + 1);
	}
	return 0;
}

void error(char * fmt, ...)
{
	va_list list;
	char msg[256];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);
	syslog(LOG_ERR, "%s", msg);
	exit(1);
}

void info(char * fmt, ...)
{ /*
	va_list list;
	char msg[256];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);
	miplog(MLOG_INFO, "%s", msg); */
}

dhcpdump(struct dhcp_packet * pkt)
{
	struct option * opt;
	int i;

	printf("BOOTP fields:\n");
	printf("\tOpcode = %d\n", pkt->op);
	printf("\tHardware type = %d\n", pkt->htype);
	printf("\tHardware addr length = %d\n", pkt->hlen);
	printf("\tHops = %d\n", pkt->hops);
	printf("\tTransaction id = %u\n", ntohl(pkt->xid));
	printf("\tSeconds = %d\n", ntohs(pkt->secs));
	printf("\tFlags = 0x%x\n", ntohs(pkt->flags));
	printf("\tClient IP = %s\n", inet_ntoa(pkt->ciaddr));
	printf("\tYour IP = %s\n", inet_ntoa(pkt->yiaddr));
	printf("\tServer IP = %s\n", inet_ntoa(pkt->siaddr));
	printf("\tGateway IP = %s\n", inet_ntoa(pkt->giaddr));
	printf("\tClient hardware address = %02x:%02x:%02x:%02x:%02x:%02x\n",
		pkt->chaddr[0], pkt->chaddr[1], pkt->chaddr[2],
		pkt->chaddr[3], pkt->chaddr[4], pkt->chaddr[5]);
	printf("\tServer name = %s\n", pkt->sname);
	printf("\tBoot file = %s\n", pkt->file);

	printf("DHCP options:\n");
	opt = (struct option *)(pkt->options + 4);
	while (opt->type != DHO_END) {
		switch (opt->type) {
		    case DHO_PAD : break;
		    case DHO_SUBNET_MASK :
			printf("\tSubnet mask = %s\n", 
				inet_ntoa(*((struct in_addr *)opt->value)));
			break;
		    case DHO_ROUTERS :
			printf("\tRouters = %s",
				inet_ntoa(*((struct in_addr *)opt->value)));
			for (i=4; i<opt->length; i+=4)
			    printf(",%s", 
				inet_ntoa(*((struct in_addr *)(opt->value+i))));
			printf("\n");
			break;
		    case DHO_DHCP_MESSAGE_TYPE :
			printf("\tDHCP message type = %s\n", 
				dhcp_message_types[opt->value[0]]);
			break;
		    case DHO_DHCP_SERVER_IDENTIFIER :
			printf("\tServer identifier = %s\n",
				inet_ntoa(*((struct in_addr *)opt->value)));
			break;
		    case DHO_DHCP_LEASE_TIME :
			printf("\tLease time = %u\n", 
				ntohl(*((u_long *)opt->value)));
			break;
		    default :
			printf("\tOption type = %d, length = %d, value = 0x",
				opt->type, opt->length);
			for (i=0; i<opt->length; i++)
				printf("%02x", opt->value[i]);
			printf("\n");
			break;
		}
		if (opt->type != DHO_PAD)
			opt = (struct option *)((char *)opt + opt->length + 2);
		else
			opt = (struct option *)((char *)opt + 1);
	}
}
