/*
 * icmptest.c
 *	standalone test for WRITING beacons
 *
 *	syntax: icmptest dev dev_ip
 * e.g.,
 *	dev: eth0
 *	dev_ip: ip address associated with eth0
 */
/*
 * Copyright (c) 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * icmpinit.c
 *
 * initialization for handling icmp beacon
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/ethernet.h>
#include <linux/if.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
/*#include <netinet/ip_var.h> */
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#include "mnstate.h"
#include "mnstats.h"
#include "unixsock.h"
#include "mip_log.h"
#include "mndversion.h"

#ifdef IPSEC
#include <netsec/ipsec.h>
#include "ipsec.h"
extern int ipsecKeysInuse;
#endif

#define TRUE	1
#define FALSE	0

int debug = 1;

static void sendBroadcast(int);
static void sendMulticast(int);
static void recvMulticast(int, struct in_addr, struct in_addr);

int beaconSd;

/*
 * initIcmp
 *
 * open raw icmp socket so that we can send and recv router adverts
 * on it.  should be able to send solicitations too.
 *
 * Note: if and when we can send multicast (224.0.0.1) as well as
 * limited broadcast, we can't do both on the same socket descriptor.
 * The code in netinet/ip_output.c won't support that.
 * 
 * returns:
 *	int beaconSd - socket descriptor
 */ 
int
initIcmp(sendBcast, sendMcast, maddr, mif, dev)
int sendBcast;	/* want to send broadcast only */
int sendMcast;   /* want to send multicast only */
struct in_addr maddr;  /* multicast address for output */
struct in_addr mif;  /* if non-zero, a default multicast ip interface */
char *dev;   /* mobile device name */
{
	struct protoent *proto;
	int on;
	int rc;

	if (!(proto = getprotobyname("icmp"))) {
 		syslog(LOG_ERR, "initIcmp(): unknown protocol icmp (%m).");

		exit(1);
	}
	/* icmp raw socket for sending out ICMP adverts
	*/
	if ((beaconSd = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
 		syslog(LOG_ERR, "initIcmp(): socket (%m)");
		exit(1);
	}
	
	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(beaconSd, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initIcmp(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	rc = setsockopt(beaconSd, SOL_SOCKET, SO_BINDTODEVICE, dev, IFNAMSIZ);
	if (rc < 0) {
		syslog(LOG_ERR, "initIcmp(): SOL_SOCKET (%m)");
		exit(1);
	} 

	/* if no multicast addr set, we still want to enable
	 * default reception of multicast
	 */
	if (sendMcast == FALSE) {
		maddr.s_addr = inet_addr("224.0.0.1");
	}
	recvMulticast(beaconSd, maddr, mif);

	/* configure send side of socket, either broadcast or multicast
	*/
	if (sendBcast)  {
		sendBroadcast(beaconSd);
	}
	else if (sendMcast) {
		sendMulticast(beaconSd);
	}
	else {
		/* has to be one or the other */
		syslog(LOG_ERR, "internal icmp configuration error");
		exit(1);
	}

	/* configure socket to receive incoming i/f index
	 * which allows us to determine which i/f a packet came in on
	 */
	/*rc = setsockopt(beaconSd, IPPROTO_IP, IP_RECVIF, (char *)&on, sizeof(on));
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(IP_RECVIF) (%m)");
		exit(1);
	} */

#ifdef IPSEC
	if (ipsecKeysInuse) {
		int	ipsec_level = IPSEC_LEVEL_REQUIRE;

		if (setsockopt(beaconSd, SOL_SOCKET,
			SO_SECURITY_PLAINTEXT_ROUTE,
			(char *)&ipsec_level, sizeof(int)) < 0)
		{
			syslog(LOG_ERR, "initIcmp(): sockopt SO_SECURITY_PLAINTEXT_ROUTE failed");
			if (debug) {
				printf("initIcmp(): sockopt SO_SECURITY_PLAINTEXT_ROUTE failed");
			}
			exit(1);
		}
	}
#endif /* IPSEC */

	return(beaconSd);
}

/*
 * sendBroadcast
 *
 * done when we want to send 255.255.255.255 icmp router messages
 * on any interface.  We set ROUTETOIF so that the destination
 * ip address used with sendto will determine the interface.
 * We set BROADCAST because we want to broadcast.  We set
 * LIMITBCAST, so that our ip dst will be munged into 255.255.255.255
 * after ROUTETOIF has determined the interface.
 * ROUTETOIF hardwires the ttl to 1.
 */
static void
sendBroadcast(int beaconSd)
{
	int on;
	int rc;
	int ttl;

	/* set ROUTETOIF for this socket.  packets sent to i/f ip
	 * address are not routed but delivered directly
	 */
  	on = 1;
#ifdef BJORN_DEBUG
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_DONTROUTE, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
 		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_DONTROUTE) (%m)");
		exit(1);
	}
#endif
  	ttl = 1;
	rc = setsockopt(beaconSd, IPPROTO_IP, IP_TTL, &ttl, 1);
	if (rc < 0) {
 		syslog( LOG_ERR, "initIcmp(): setsockopt(IPPROTO_IP) (%m)");
		exit(1);
	}

	/* so we can broadcast on this socket
	*/
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_BROADCAST, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_BROADCAST) (%m)");
		exit(1);
	}
}

/*
 * recvMulticast 
 *	1. merely enables the reception of
 * 	multicast addresses on the MIP default multicast port.
 * 	This allows us to recv icmp advert/solicitations from
 * 	hosts using multicast.
 *
 *	2. disables loopback of multicast packets
 *
 * Note: this code does NOT cause a fatal exit if the attempt
 * to add a multicast group on a particular interface fails.
 * It is possible that a given i/f might not support multicast.  
 * A message will be logged however.
 */
static void
recvMulticast(int beaconSd, struct in_addr maddr, struct in_addr mif)
{
	struct ip_mreq mreq;
	unsigned char loop = 0;

 	mreq.imr_multiaddr.s_addr = maddr.s_addr;
	if (mif.s_addr)
		mreq.imr_interface.s_addr = mif.s_addr;
	else
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);

        if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0){
 		syslog(LOG_ERR, "initIcmp(): IP_ADD_MEMBERSHIP failed (%m).");
		return;
	}
        if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_MULTICAST_LOOP, &loop, sizeof(loop))) {
 		syslog(LOG_ERR, "initIcmp(): IP_MULTICAST_LOOP failed (%m).");
		exit(1);
	}
}

/*
 * sendMulticast
 *
 * setup send side.  recvMulticast already joined the mcast group.
 * We set the TTL to 1.  Setting the i/f either is only done
 * in the recv routine BECAUSE we only support one interface
 * for mns (mnd).  mipd has to explicitly set the output i/f
 * for each interface a multicast beaocn is sent out on.
 */
static void
sendMulticast(int beaconSd)
{
	unsigned char ttl = 1;

 	if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0) {
 		syslog(LOG_ERR, "initIcmp(): IP_MULTICAST_TTL failed (%m).");
		exit(1);
	}
}

/*
 * aim multicast socket at a given interface
 */
setMcastIf(struct in_addr addr)
{
	int rc;
     	rc = setsockopt(beaconSd, IPPROTO_IP, IP_MULTICAST_IF, &addr, sizeof(addr));
	if (rc < 0) {
 		syslog(LOG_ERR, "setMcastIF(): IP_MULTICAST_IF failed (%m).");
		exit(1);
	}
}

/* prototypes
*/ 
static 	in_cksum(u_short *, int);
unsigned short getSequence(u_short *);

extern int beaconSd;
static int resendTimer;

static int current_lifetime;


/*
 * sendBeacon --
 *
 * See RFC 1256 for ICMP router adverts, etc.
 */

void
sendBeacon(sd, router, dst, coa, mse_lifetime, identity, bothAgents, seq)
int sd;
struct sockaddr_in *router;	/* router in ADVERT packet */
struct sockaddr_in *dst;        /* local ip dest for packet */
struct sockaddr_in *coa;	/* COA for FA */
int mse_lifetime;		/* lifetime to mobile host */
int identity;			/* true if fa, else ha */
int bothAgents;			/* TRUE if both ha AND fa */
u_short seq;
{
	int errno;
	register struct icmp_advert *icp;
	register struct mip_mseicmp *mse;
	register struct mip_mse_icmp_mac *msemac;
	register struct mip_mse_icmp_info *mseinfo;
	register struct mip_mse_icmp_auth *mseauth;
	int rc;
	extern int debug;
	int offset=0;
	u_short flags = 0;
	u_char outpack[MAXICMPPACKET];
	char *location;
	u_char *key;
	int spi;
	int infosize = 0;
	char ifname[MAXIFNAME];
	int doArp;

	/* convert router string to ip address */

	/* pointer to packet buffer */
	icp = (struct icmp_advert *)outpack;

	/* setup icmp part of packet
	*/
	icp->icmp_type = ICMP_ROUTERADVERT;
	icp->icmp_code = 16;
	icp->icmp_cksum = 0;
	icp->icmp_naddrs = 1;
	icp->icmp_asize = 2;

	/* the "lifetime" of the advertisement
	*/
	icp->icmp_adlifetime = htons(120);

	/* the router is most likely to be the mobile interface address
	 */
	icp->icmp_router = router->sin_addr.s_addr;
	icp->icmp_preference = htonl(0);

	/* setup mse part of packet */
	mse = (struct mip_mseicmp *) (outpack + sizeof(struct icmp_advert));
	mse->mse_type = MIPAGENTAD;
	/* note that all flags are off by default */
	if (identity == SELF_MN) {
		mse->mse_len = 6;
		SET(flags, MNBIT);
		mse->mse_flags = htons(flags);
		offset = -4;
	}
	/* set beacon fields for HA and/or FA characteristics.  if agent
	 * plays both roles, the FA characteristics will be used 
	 * as they are a superset of the HA characteristics in this context.
	 */
	if (identity==SELF_FA || bothAgents) {
		SET(flags, FABIT);
		mse->mse_len = 10;   /* FA has coa */
		mse->mse_coa = coa->sin_addr.s_addr;
		offset = 0;
	}
	mse->mse_flags = htons(flags);

	/* each interface has associated sequence number 
	 */
	mse->mse_sequence = htons(seq);
	mse->mse_lifetime = htons(mse_lifetime);

	/* offset is size so far into packet (outpack) */
	offset += sizeof(struct icmp_advert) + sizeof(struct mip_mseicmp);

	doArp = 1;

	/* compute ICMP checksum here */
	icp->icmp_cksum = in_cksum((u_short *)icp, offset);

	/* send the beacon
	*/
	errno = 0;
	rc = sendto(sd, (char *)outpack, offset, 0, 
		(struct sockaddr *)dst, sizeof(struct sockaddr));
	if (debug) {
		printf("Wrote icmp advert, %d chars, ret=%d\n", offset, rc);
		dumpme(outpack, offset); 
		printf("End of dump\n");
		printf("send packet to %s\n", 
			inet_ntoa( dst->sin_addr ));
		printf("rc, %d errno %d, strerr=%s\n", rc, errno, strerror(errno) );
	}

	if (rc < 0 || rc != offset)  {
		if (rc < 0) {
			/* if out of buffers continue, try again later
			*/ 
			if (errno == ENOBUFS)
				return;
			/* linux is not thrilled about sending ICMP
			 * router adverts ...  This is a bug that needs
			 * to be fixed, probably by pushing the socket
			 * mechanism down to include the IP hdr
			 */
			if (errno == ENETUNREACH) {
				return;
			}
		}
		/* else fatal error
		*/
		syslog( LOG_ERR, "sendBeacon(): sendto failed (%m).");
		exit(1);
	}
	if (debug){
		fprintf( stderr, "beacon sent successfully\n" );
	}

}

/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 *	(from BSD ping code)
 */
static
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

/*
 * generate next mse sequence number for router advert message
 */
unsigned short
getSequence(u_short *seq)
{
	u_short next = *seq;	/* return previous value */
	/* increment for next round
	*/
	if (next == 0xffff) {
		*seq = 256;	/* non reboot rollover */
	}
	else {
		*seq = *seq + 1;
	}
	return(next);
}

dumpme(char *packet, int size)
{
	/* dump the packet */
	int i = 0;
	char *p = packet;
	printf("packet dump ------------------------\n");
	for ( ; i < size; i++) {
		if ( (i != 0) && ((i % 16) == 0)) {
			printf("\n");
		}
		printf("%2.2x ", *p++ & 0xff);
	}
	printf("\nEND: packet dump ------------------------\n");
}

/*
 * syntax: icmptest eth0 1.2.3.4
 *
 * string name of device, followed by IP address of device
 */

main(argc, argv)
int argc;
char **argv;
{
	int sd, sd2;
	char *dev = argv[1];  /* device name as string */
	char *ifaddr = argv[2];
	struct in_addr mif;
	struct sockaddr_in router;
	struct sockaddr_in dst;
	/* beacon dest is either 255.255.255.255 or 224.0.0.1
	*/
	struct in_addr maddr;	/* beacon destination address */

	mif.s_addr = inet_addr(ifaddr);

	/* do broadcast first
	*/
	maddr.s_addr = inet_addr("255.255.255.255");
	sd = initIcmp(1, 0, maddr, mif, dev);

	/* write one broadcast beacon
	*/
	router.sin_addr.s_addr  = inet_addr("1.2.3.4");
	dst.sin_addr = maddr;
        dst.sin_family = AF_INET;
        dst.sin_port = htons(434);

	maddr.s_addr = inet_addr("224.0.0.1");
	sendBeacon(sd, &router, &dst, &router, 1,  1, 0, 1);

	/* write one multicast beacon
	*/
	sd2 = initIcmp(0, 1, maddr, mif, dev);
	dst.sin_addr = maddr;
	sendBeacon(sd2, &router, &dst, &router, 1, 1, 0, 1);
}

