/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipconfig.c
 * 
 * purpose: read/parse/store tag values from /etc/mipd.conf mipd configuration
 * file.
 * 
 * file syntax:
 * 
 * ascii lines, one tag per line, may have N values depending on tag skip
 * lines/and comments, # in 1st column is comment
 * 
 * tag syntax:
 * 
 * 1. send router advert beacons out if ifname with limited broadcast
 * 
 * fa <ifname>	<ipaddr>	[<arp>|<noarp>]
 * 
 * ha <ifname>	<ipaddr>	[<arp>|<noarp>|<proxyarp>]
 * 
 * both may be given. at least one MUST be given.  <ifname> is name of
 * the interface, e.g., ed0.  <ipaddr> may be "255.255.255.255" (limited
 * broadcast) or "224.0.0.1" (multicast).
 * 
 * Example:
 * 
 * fa wlp0	255.255.255.255 
 *
 * fa wlp0	255.255.255.255  noarp
 * 
 * Also may give "nobeacon" as <ipaddr>, which means don't beacon on that
 * interface, but allow registration packets to cross it.
 * One might choose to have a combined HA/FA interface and have
 * only the FA beacon on that interface (MNs would have to understand
 * though that FA beacons from their HA, should be interpreted as HA
 * beacons).
 *
 * Note that if a single interface is used to define a HA *and* a FA, when
 * characteristics overlap between the two definitions, the values defined
 * for the HA are given preference.  Currently this includes only whether
 * ARP is on or off, the sequence counter, and whether the beacon is
 * broadcast or multicast.  Global values which might
 * eventually be associated with interfaces include: advertised Mobile-IP
 * lifetime, COAs, and compression/encapsulation bits.
 *
 * With the current implementation, at least one agent defined for a given
 * interface *must* beacon.  It is not allowed that an agent(s) be defined
 * for an interface which does not beacon.  Further, agent definitions on 
 * a given interface marked as "nobeacon" still must match its co-defined
 * agent (if any) wrt whether "noarp" or "arp" is set.
 * 
 * ICMP router adverts.  The main part of an ICMP router advert contains
 * a router address. For the FA, this is the address bound to the interface
 * here (and may or may not be the COA, see below). For the HA, we will also
 * use the address bound to the interface for the router address.  This
 * address MAY be used by MNs for setting a default route. If for some reason
 * you are using aliases on a given interface, most likely the first one will
 * be pulled off and used.
 * 
 * arp is the default, if nothing follows <ipaddr>.
 * 
 * "noarp" means the arp function is turned off in the kernel on a per interface
 * basis.  If you turn arp off, you must use authenticated ICMP beacons.
 * Both MNs and agents must use the netauth_key (default) to share
 * a MD5 authentication secret.  This will allow the routing daemons
 * to install link-layer routes for only those systems knowing the secret.
  *
  * "proxyarp" is identical to "arp" except that the HA will also construct
 * proxy arp entries for mobile nodes which are away.
 * This switch may only be used with HA interfaces.
 * other RFC2002 behaviors with regard to ARP have not yet been
 * implemented, notably, gratuitous arp by HA and MN when returning
 * home, the restricted arp replies for co-located MN's,
 * and the specialized arp behavior between
 * the MN and FA.

 * 
 * 
 * 2. beacon timer in seconds
 * 
 * beacon	<seconds>
 * 
 * send a router advertisement every N seconds.  This is for all ICMP beacons
 * specified with ha/fa lines.  1 second is a minimum and the default.
 * 
 * 3. list entry timeout
 * 
 * mobile_lifetime <seconds> | max
 * 
 * MN must refresh within this time.  This is time that HA/FA hold onto visitor
 * list entry and routes implied.
 * 
 * if lifetime is sent to max, MNs will not timeout.  Recommend this only be
 * used for testing as if HA/FA crashes, MN will lose connectivity (or be
 * forced to reregister).
 * 
 * Example:
 * 
 * mobile_lifetime	180
 * mobile_lifetime      max
 *
 * Note: max is not the best idea unless you somehow have agents that
 * never crash.
 * 
 * 4. coa ipaddr
 * 
 * coa <ipaddr/dnsname>
 * 
 * Care Of Address for FA ifs. Not used by a HA.  
 * Note this address MUST be routeable (unlike the FA interface ip
 * address, which can be private); e.g., you should be able to ping
 * the coa from the HA.  If a FA entry exists, a coa must be
 * provided too.  There is currently only one coa per system.
 * 
 * 5. MIP spi/keys
 * 
 * mnha_key <ipaddr> <transform> <spi> <hexstring>
 * faha_key <ipaddr> <transform> <spi> <hexstring>
 * famn_key <ipaddr> <transform> <spi> <hexstring>
#ifdef HARP
#ifdef MIPSEC
 * haha_key <ipaddr> <transform> <spi> <hexstring>
#endif
#endif 
 * 
 * The above lines as appropriate should be the same in the mn config file. 
 * The existance of one of the lines above means a security association 
 * is required between  the two parties involved.  Keyed MD5 is used as 
 * the authentication and * integrity mechanism.  If ANY key is supplied 
 * for one relationship,  we assume that keys are required for all 
 * entities in that relationship.
 *
 * The MN HA key relationship is required and is non-optional.  
 *
 * If a ha requires that fas have keys, all fas must have keys, and
 * packets arrivied from fas without a key relationship will be ignored.
 * If the faha_key or famn_key are not specified, keys are not required
 * and key relationships will be ignored if hash/mses show up in a UDP
 * registration/replay packet.
 * 
 * For now, the transform is ignored as there is only one. The hexstring is
 * transform dependent.  For md5 it must consist of 32 bytes of hex nibbles
 * (128 bits).
 * 
 * Example:
 * 
 * mnha_key 1.2.3.4 md5 256 0x9700ca3a82f7cf9b2d0cc739abd5f33f
 * 
 * 6. location "string ...."
 * 
 * The location information is optional.  If present, the agent will append a
 * mobile service extention of type 20 to the ICMP router advertisements sent
 * out.  These can be used by mobile nodes to id a particular agent; e.g.,
 * 
 * location "PCAT 128, Machine Room SE Corner"
 *
 * This option is not specified in Mobile-IP and is a PSU hack.
 * However there is reason to believe that non-PSU MIP implementations
 * would ignore it.
 * 
 * 7. timestamp deviation allowed by HA
 * 
 * ntp_slop  seconds
 * 
 * If not specified, slop tolerated is 2 seconds.  If a TS arrives from the MN
 * (in a UDP registration packet) and it is off by greater than this value,
 * the HA will send a NAK with error code 133 (ID MISMATCH) and a time update
 * to the MN.  The time slop has to be set to the worst case latency
 * across the network possible.  That is hard to predict.
 *
 * 8. replay  timestamp | nonce
 *
 * timestamp is default and nonce is not implemented.  This switch is
 * optional.
 * 
 * 9. NEW: netauth_key - icmp beacons sent by mipd MAY be authenticated.
 * This is optional.  If authentication is desired,
 * 
 * netauth_key <transform> <spi> <hexstring>
 * 
 * should be put into the config file, and arp should be turned off
 * on at least one interface.  If a key exists, and arp is off, ICMP
 * beacons on that interface will be authenticated.  The MAC address
 * will be put in the beacon in such a way that the IP address, and MAC
 * address of the interface are authenticated.  Nodes recv. the beacon
 * can install a link-layer route based on shared authentication. 
 * The mn config file should have the same line.
 * This is a simple form of ad hoc routing.  Basically the key defines
 * the logical link.  Those who hear the key and share it can exchange
 * network messages.  arp is ignored.
 * 
 * 10. NEW: adhoc_key - icmp beacons sent by agents and mns may have a secondary
 * optional authentication key if so desired.  This key is used to
 * allow temporary establishment of a logical link.  It might be
 * manufactured on the spot, and shared with guests and then destroyed
 * when the guests leave.  You could add/delete to the config file
 * and restart a FA daemon.
 * 
 * adhoc_key <transform> <spi> <hexstring>
 * 
 * should be put into the config file.  MN's if configured to pay attention,
 * will only trust icmp beacons authenticated with the above key as long as
 * the md5 key matches for both sending and receiving systems.  The mn config
 * file should have the same line.
 * 
 * 11. NEW: ICMP nonce or timestamp replay protection.  
 * (nonce not supported) If not specified, timestamp is assumed.  
 * NOT CURRENTLY IMPLEMENTED.
 * 
 * beacon_replay <timestamp | nonce>
 * 
 * 12.ICMP timeslop allowed in replay protection for address authentication.
 * NOT CURRENTLY IMPLEMENTED.
 * 
 * beacon_timeslop <seconds>
 * default is 2 seconds
 *
 * 13. logIPOnly <off | on>
 *
 * default is "on".  setting this to "on" prevents IP addresses printed to
 * the logfile from being resolved to DNS names, avoiding a lookup with the
 * local DNS server which may be inaccessible in situations of low-bandwidth
 * or lacking connectivity.  this tag is optional.
 *
#ifdef IPSEC
 * 14.  ipsec_ha mn <ha src IP> <mn dst IP> ah|esp|ah+esp	\
 *		<spi_o> <spi_i> [<spi_o> <spi_i>] 
 *
 *	ipsec policy (with transform/spi) to use with the given MN.
 *	The policy is dynamically determined by a request from the MN.
 *	The config line does NOT force the mn to use a particular
 *	policy (possibly it should).
 *
 * 15.  ipsec_ha mn_policy strict|lax                           
 *	If mn_policy is set to strict, ALL mns must have a ipsec_ha mn
 *	line, else they will be ignored when they register.  
 *	Any missing MN will have a log message put in the log.  
 *	The default here is lax.  This can be used as a strong
 *	consistency check to make sure that keys are loaded for
 *	all mns in a "family" of mns.
 *
 * 16.  ipsec_tun ha <ha src ip> <coa ip> ah|esp|ah+esp		\
 *		<spi_ah> | <spi_ah><spi_esp>
 *      ipsec_tun coa <ha src ip> <coa ip> ah|esp|ah+esp 	\
 *		<spi_ah> | <spi_ah><spi_esp>
 *
 *	These switches are used to specify a security association 
 *	between a MIP Home Agent and a COA (typically a Foreign Agent,
 *	since a Mobile Node with a COA would be unlikely to have a
 *	IP address that could be statically determined, however the MN
 *	possibility is not ruled out).  The first form is used at a HA.
 *	The second form is used at a FA (for example).  All of the
 *	remaining parameters should be the same.  
 *
 *	These entries cause a system acting as a Home Agent to have
 *	a list of Foreign Agents with which authenticated tunnels
 *	will be formed.  Every route installed using the mvif driver
 *	where a ipsec_tun line eixsts will form packets that are 
 *	IP<ipsec>IP packets as opposed to IPIP packets, where <ipsec>
 *	is AH, ESP, or AH+ESP.   With AH, we call this construction
 *	an authenticated MIP tunnel.  The mvif driver, instead of
 *	merely appending an IP header, will append an IP<ipsec> combination.
 *	Note that the tunnel's SA is one-way (as one might expect with
 *	routing).  There is only one SPI value, and it must be the same
 *	at both the tunnel entrance and tunnel exit.  At a HA,
 *	mipd will dynamically installed authenticated tunnels if such
 *	a ipsec_tun ha entry exists in the configuration file.  At the FA,
 *	mipd does nothing but check for the existance of an a priori loaded
 *	key/SA in the kernel at boot (loaded via keyadmin(8)).  The kernel
 *	itself will do all of the processing on input.
 *
 *	All HA/FA tunnels may thus be made more secure.
 *	One does not need to rule out unauthenticated packets, however there
 *	is an input switch associated with the mvif driver that can be
 *	used to rule out any non-IPSEC packets.  There mvifconfig(8) for
 * 	more information.
 *
 *	SETUP
 *	=====
 *
 *	at HA --
 *
 *	/etc/keys has line like:
 *		ah	spi	HA(src)   FA(dst)  transform key
 *	/etc/mipd.conf has line like:
 *		ipsec_tun ha <HA_ip> <FA_ip> ah <spi>
 *	action:
 *		when MN requests COA to FA, tunnel is installed HA/FA
 *
 *	at FA --
 *
 *	/etc/keys has line like:
 *		ah	spi	HA(src)   FA(dst)  transform key
 *	/etc/mipd.conf has line like:
 *		ipsec_tun coa <HA_ip> <FA_ip> ah <spi>
 *
 *	action:
 *		at boot, mipd verifies that key is loaded. if not, it exits.
 *
 * 17.  ipsec_fa mn <fa src IP> <mn dst IP> ah|esp|ah+esp
 *		<spi_o> <spi_i> [<spi_o> <spi_i>] 
 *	NOT IMPLEMENTED.
 *
 * 18.  ipsec_fa mn_policy strict | lax
 *	NOT IMPLEMENTED.
 *
#endif
#ifdef HARP
 * 19. coha < co-HA ip/dns name >
 *	 Only used when HARP is in effect.  The coha config line indicates the
 *	 address of the co-HA.
 *	 
 * 20.  harp_tunnel < yes | no >
 *	Indicates whether the home subnet is partitioned.  Functionally this
 *	option indicates whether AT_COHA state exists. 
 *
 *	In the case that the home subnet is partitioned a Mobile Node
 *	may be physically located on one portion of the home subnet but
 *	may not be directly reachable by both Home Agents.  In this case the
 *	HA which is not located on the same physical link as the MN must
 *	forward packets addressed to the MN at the IP address of the other
 *	Home Agent as if it were a COA.
 *		
 * 21.  harp2_arp_disable < if_name >
 *	Indicates whether the secondary home agent should dynamically  
 *	disable ARP on its MIP interface.  This option only makes sense if the 
 *	home-subnet is not partitioned and both HAs have multiple 
 *	interfaces.  This option should be used only if harp_tunnel=no.
 *	
 *	The purpose of this option is to allow two Home Agents on the 
 *	same physical link to share the HA IP address while avoiding
 *	ARP conflicts which would arise if both HAs were simultaneously 
 *	responding to ARP requests for that IP address.
endif

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "mipconfig.h"
#include "mip_log.h"
#include "mipstats.h"
#ifdef IPSEC
#include "ipsec.h"
extern int ipsecKeysInuse;
#endif
#ifdef HARP
#ifndef IPSEC
#define MIPSEC
#endif
#include "harp.h"
#endif

/* global config variables *************************** */

struct mip_config mip_config;

static int globalArp = 0;	/* true if arp set on any interface */
extern int debug;


/*************************************************/

/*
 * function prototype declarations
 */
static          switchOnOp();
int             resolveAddress(char *, struct sockaddr_in *);
static          doAgentLine(char *, char *, int, struct mip_ifs *, int);
static          doCoa(char *op, char *linebuf, int line);
int             pack_string(char *, unsigned char *, int);
static          dokey(char *, char *, int, int, int);
int      	extractString(char *linebuf, char *output, int max);
u_char          *getmac(char *ifname); 

#ifdef IPSEC
void            ipsecConfig(char *tag, char *linebuf, int linenum, int haflag);
#endif

#ifdef HARP
static          doCoHa(char *op, char *linebuf, int line);
static		doHarpTunnel(char *op, char *linebuf, int line);
static		doHarp2( char *op, char *linebuf, int line );
static		doHarpPing( char *op, char *linebuf, int line );
#endif

void
configInit()
{
	mip_config.sawFa = mip_config.sawHa = mip_config.sawCoa = FALSE;
	mip_config.faha_security = mip_config.famn_security = FALSE;
	mip_config.mnha_security = FALSE;
	mip_config.netauth_security = mip_config.adhoc_security = FALSE;
	mip_config.beaconRate = BEACONTIMER;
	mip_config.mobileLifetime = MOBILELIFETIME;
	mip_config.ntp_slop = 2;
	mip_config.doTimestamp = TRUE;
	mip_config.beacon_timeslop = 2;
	mip_config.beacon_replay = TRUE;
	mip_config.beacon_timestamp = TRUE;
        mip_config.logIPOnly = TRUE;
#ifdef HARP
	mip_config.harp_tunnel = FALSE;
	mip_config.harp = FALSE;
	mip_config.harp2_arp_disable = FALSE;
	mip_config.harp_ping = HARP_PING_INTERVAL;
#endif

	/* initialize keylists
	*/
	mip_config.sa[MNHASECURITY].kl = init_keylist();
	mip_config.sa[FAMNSECURITY].kl = init_keylist();
	mip_config.sa[FAHASECURITY].kl = init_keylist();
#ifdef HARP
#ifdef MIPSEC
	mip_config.sa[HAHASECURITY].kl = init_keylist();
#endif
#endif
#ifdef IPSEC
        mip_config.ipsec[HA_SECTYPE_MN].kl = init_keylist();
        mip_config.ipsec[HA_SECTYPE_MN].inuse = FALSE;
        mip_config.ipsec[HA_SECTYPE_MN].strict = FALSE;
        mip_config.ipsec[FA_SECTYPE_MN].kl = init_keylist();
        mip_config.ipsec[FA_SECTYPE_MN].inuse = FALSE;
        mip_config.ipsec[FA_SECTYPE_MN].strict = FALSE;
        mip_config.ipsec[TUN_SECTYPE].kl = init_keylist();
        mip_config.ipsec[TUN_SECTYPE].inuse = FALSE;
        mip_config.ipsec[TUN_SECTYPE].strict = FALSE;
#endif
        mip_config.access_control_order = 0;
        mip_config.access_control_default_permission = ALLOWED;
        mip_config.denylist = NULL;
        mip_config.allowlist = NULL;
}

/*
 * read the config file and load its values
 * 
 * return: 0: ok -1: fatal error (although we return).  All errors logged prior
 * to returning.
 */
int
configRead(configName)
	char           *configName;	/* name of config file */
{
	FILE           *fd;
	char            linebuf[1024];
	char            opbuf[1024];
	char           *opword;
	char           *result;
	int             lineNumber = 1;

	fd = fopen(configName, "r");
	if (fd == NULL) {
		syslog(LOG_ERR, "configRead(): fopen of config file \"%s\" failed.\n",
		       configName);
		return (-1);
	}

	/*
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		bzero(linebuf, sizeof(linebuf));
		result = fgets(linebuf, MAXLINE, fd);
		if (result == NULL)
			break;
		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		/*
		 * between the opword  and the following parameter, there
		 * should be a "\t"
		 */
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0))
			continue;
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */
		switchOnOp(opword, result, lineNumber);
		lineNumber++;
	};
	fclose(fd);
#ifdef IPSEC
	if (ipsecKeysInuse) {
		closeKeySocket();
	}
#endif

	/*
	 * check semantics
	 */

	/* must have reasonable beacon setting
	*/
	if (mip_config.beaconRate <= 0) {
		syslog(LOG_ERR,
		       "configRead(): beacon rate must be > 0.\n");
		return (-1);
	}
	/*
	 * must have at least one fa or ha line
	 */
	if (mip_config.sawFa == 0 && mip_config.sawHa == 0) {
		syslog(LOG_ERR, "configRead(): must have at least one FA or HA line.\n");
		return (-1);
	}
	/*
	 * if we saw a FA line, we must have a COA too
	 */
	if (mip_config.sawFa && mip_config.sawCoa == 0) {
		syslog(LOG_ERR, "configRead(): saw FA line, but no COA.\n");
		return (-1);
	}
	/* if noarp is set,  there must be at least one logical
	 * network key(netauth_key or adhoc_key) 
	 */
	if ( (globalArp == 0) && (mip_config.netauth_security == 0)
	   && (mip_config.adhoc_security == 0)) {
		syslog(LOG_ERR, "configRead(): noarp is set, but neither netauth_key or adhoc_key exists");
		return (-1);
	}

	/* can't currently have both multicast and broadcast mixed
	 * across interfaces.  we only have one socket descriptor
	 */
	if (mip_config.sawBroadcast && mip_config.sawMcast) {
		syslog(LOG_ERR, "cannot support mixed broadcast/multicast ifs -- either one or the other");
		return (-1);
	} 

#ifdef IPSEC
	/* if mn_policy is "strict," every {mnha,famn}_key must be paired
	 * with an appropriate ipsec_{ha,fa} mn key
	 */
{
	struct ipsec_policy	*iskeys;	/* ipsec keys */
	struct sec_assoc	*mikeys;	/* mobileIP keys */
	struct keylist		*k;
	int	i;

	for (i = 0; i < (mip_config.sawHa>0)+(mip_config.sawFa>0); i++ ) {
		if (i==0 && mip_config.sawHa) {
			iskeys = &mip_config.ipsec[HA_SECTYPE_MN];
			mikeys = &mip_config.sa[MNHASECURITY];
		}
		else if (i==1 && mip_config.sawFa) {
			iskeys = &mip_config.ipsec[FA_SECTYPE_MN];
			mikeys = &mip_config.sa[FAMNSECURITY];
		}
		else {
			continue;
		}

		/* ASSUMPTION: strict is only defined when inuse is defined */
		if (!iskeys->strict) {
			continue;
		}

		klist_initwalk(mikeys->kl);
		while (k = klist_nextwalk()) {
			if (! search_keylist(iskeys->kl, k->ipaddr)) {
				syslog(LOG_ERR, "configRead(): mn_policy is strict and some %s_key is not paired with an ipsec_%s key.",
				    (i==0)?"mnha":"famn", (i==0)?"ha":"fa" );
				exit(1);
			}
		}
	} /* endfor */
}
#endif /* IPSEC */

	return (0);
}

static
switchOnOp(op, linebuf, line)
	char           *op;
	char           *linebuf;
	int             line;
{
	char            dc[1024];
	char            v1[1024];
	char            v2[1024];
	int             rc;
	static struct mip_ifs *mifs = (struct mip_ifs *) & mip_config.mip_ifs[0];

	/* beacon secs */
	if (strcmp(op, "beacon") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		mip_config.beaconRate = (unsigned short) atoi(v1);
	/* mobile_lifetime secs */
	} else if (strcmp(op, "mobile_lifetime") == 0) {
		/* TBD -- differentiate on a per/interface basis? */
		sscanf(linebuf, "%s %s", dc, v1);
		if (strcmp(v1, "max") == 0) {
			mip_config.mobileLifetime = 0xffff;
		} else {
			mip_config.mobileLifetime = (unsigned short) atoi(v1);
			if (mip_config.mobileLifetime > 0xffff) {
				syslog(LOG_ERR, "switchOnOp(): Config error line %d, mobile lifetime cannot be greater than 0xffff\n", line);
				exit(1);
			}
		}
	/* fa ifname ip  arp | noarp  */
	} else if (strcmp(op, "fa") == 0) {
		mip_config.sawFa = 1;
		doAgentLine(op, linebuf, line, mifs, 1);
		mifs++;
	/* ha ifname ip  arp | noarp | proxyarp */
	} else if (strcmp(op, "ha") == 0) {
		mip_config.sawHa = 1;
		doAgentLine(op, linebuf, line, mifs, 0);
		mifs++;
	/* coa ip */
	} else if (strcmp(op, "coa") == 0) {
		mip_config.sawCoa = 1;
		doCoa(op, linebuf, line);
#ifdef HARP
	/* coha ip */
	} else if ( strcmp( op, "coha" ) == 0){
		mip_config.harp = TRUE;
		doCoHa( op, linebuf, line );
	/* partitionedSubnet */
	} else if ( strcmp( op, "harp_tunnel" ) ==0){
		doHarpTunnel( op, linebuf, line );
	} else if ( strcmp( op, "harp2_arp_disable" ) == 0){
		mip_config.harp2_arp_disable = TRUE;
		doHarp2( op, linebuf, line );
	} else if ( strcmp( op, "harp_ping" )==0){
		doHarpPing( op, linebuf, line );
#ifdef MIPSEC
	} else if ( strcmp( op, "haha_key" )==0){
		mip_config.haha_security = 1;
		dokey( op, linebuf, line, HAHASECURITY, 1);
#endif
#endif
	/* mnha_key transform spi key */
	} else if (strcmp(op, "mnha_key") == 0) {
		mip_config.mnha_security = 1;
		dokey(op, linebuf, line, MNHASECURITY, 1);
	/* famn_key transform spi key */
	} else if (strcmp(op, "famn_key") == 0) {
		mip_config.famn_security = 1;
		dokey(op, linebuf, line, FAMNSECURITY, 1);
	/* faha_key transform spi key */
	} else if (strcmp(op, "faha_key") == 0) {
		mip_config.faha_security = 1;
		dokey(op, linebuf, line, FAHASECURITY, 1);
	} else if (strcmp(op, "access_policy") == 0) {
		access_def(linebuf, line);
	/* location "string" */
	} else if (strcmp(op, "location") == 0) {
		rc = extractString(linebuf, mip_config.location,
				   MAXLOCATION);
		if (rc < 0) {
			syslog(LOG_ERR, "switchOnOp(): Config error line [%d]: bad location value\n", line);
			exit(1);
		}
		mip_config.sawLocation = 1;
	/* UDP registration msg: ntp_slop secs */
	} else if (strcmp(op, "ntp_slop") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		mip_config.ntp_slop = (unsigned long) atoi(v1);
	/* UDP replay timestamp | nonce */
        } else if (strcmp(op, "replay") == 0) {
                sscanf(linebuf, "%s %s", dc, v1);  
                if (strcmp(v1, "timestamp") == 0) {
                        mip_config.doTimestamp = 1;
                }
                else {
                        syslog( LOG_ERR, "switchOnOp(): config error line [%d]: nonce replay not supported\n", line);
                        exit(1);
                }
	/* netauth_key transform spi key */
	} else if (strcmp(op, "netauth_key") == 0) {
		mip_config.netauth_security = 1;
		dokey(op, linebuf, line, NETAUTHSECURITY, 0);
	/* adhoc_key transform spi key */
	} else if (strcmp(op, "adhoc_key") == 0) {
		mip_config.adhoc_security = 1;
		dokey(op, linebuf, line, ADHOCSECURITY, 0);
	}
	/* beacon_replay timestamp | nonce  on | off */
	else if (strcmp(op, "beacon_replay") == 0) {
		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		if (strcmp(v1, "timestamp") == 0) {
			mip_config.beacon_timestamp = 1;
		} else if (strcmp(v1, "nonce")==0) {
			syslog(LOG_ERR, "switchOnOp(): config error line [%d]: nonce replay not supported\n", line);
			return(-1);
		}
		else  {
			mip_config.beacon_timestamp = 1;
		}
		if ((strcmp(v1, "off") ==0)||(strcmp(v2, "off")==0))  {
			mip_config.beacon_replay=0;
		}
		else {
			mip_config.beacon_replay=1;
		}
	/* beacon_timeslop seconds */
	} else if (strcmp(op, "beacon_timeslop") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		mip_config.beacon_timeslop = (unsigned long) atoi(v1);
        /* logIPOnly <on|off> */
        } else if (strcmp(op, "logIPOnly") == 0) {
                sscanf(linebuf, "%s %s", dc, v1);
                if (strcmp(v1, "off") == 0) {
                        mip_config.logIPOnly = FALSE;
                } else if (strcmp(v1, "on") == 0) {
                        mip_config.logIPOnly = TRUE;
                } else {
                        syslog(LOG_ERR, "Usage: logIPOnly <on | off>\n");
                        exit(1);
                }
#ifdef IPSEC
	/* ipsec_ha mn <ip src> <ip dst> <policy> <spi> 
	 * ipsec_ha mn_policy strict | lax
	 */
 	} else if (strcmp(op, "ipsec_ha") == 0) {
                ipsecConfig(op, linebuf, line, HA_SECTYPE_MN); 
	/* ipsec_fa mn <ipaddr> <policy> <spi> 
	 * ipsec_fa mn_policy strict | lax
	 */
 	} else if (strcmp(op, "ipsec_fa") == 0) {
                ipsecConfig(op, linebuf, line, FA_SECTYPE_MN); 
 	} else if (strcmp(op, "ipsec_tun") == 0) {
                ipsecConfig(op, linebuf, line, TUN_SECTYPE); 
#endif
	/* fatal */
	} else {
		syslog(LOG_ERR, "switchOnOp(): config error line [%d]: unknown config op in config file\n", line);
		return(-1);
	}
}
    
/* reverse copy a string */
    
char * strrcpy(char *dst, const char *src)
{   
        char *temp = dst;
        int len = strlen(src);
        while (len > 0) { 
                *dst++ = src[len-1];
                len--;  
        }
        *dst = 0;
        return(temp);
}   
    
int isip(char *s)
{   
        int yes = TRUE, stop = FALSE;
    
        while (*s && !stop) {
                if (!isdigit(*s) && !(*s == '.')) {
                        yes = FALSE;
                        stop = TRUE;
                }
                s++;
        }
        return yes;
}

/* access control definitions
   access_policy order <deny_allow | allow_deny>        # default: deny_allow
   access_policy default_permission <denied | allowed>  # default: allowed
   access_policy deny  <ip | network | hostname | domain>
   access_policy allow <ip | network | hostname | domain>
*/
 
int access_def(char * linebuf, int line)
{
        char v1[80], v2[80];
        struct element * node;

        if (sscanf(linebuf, "access_policy %79s %79s", v1, v2) != 2) {
                fprintf(stderr, "config error line %d, invalid access_policy "
                                "definition\n", line);
                exit(1);
        }

        if (strcmp(v1, "order") == 0) {
                if (strcmp(v2, "deny_allow") == 0)
                        mip_config.access_control_order = DENY_ALLOW;
                else if (strcmp(v2, "allow_deny") == 0)
                        mip_config.access_control_order = ALLOW_DENY;
                else {
                        fprintf(stderr, "config error line %d, unknown access "
                                        "control order (deny_allow or "
                                        "allow_deny)\n", line);
                        exit(1);
                }
        }
        else if (strcmp(v1, "default_permission") == 0) {
                if (strcmp(v2, "denied") == 0)
                        mip_config.access_control_default_permission = DENIED;
                else if (strcmp(v2, "allowed") == 0)
                        mip_config.access_control_default_permission = ALLOWED;
                else {
                        fprintf(stderr, "config error line %d, unknown access "
                                        "control default permission (denied or "
                                        "allowed)\n", line);
                        exit(1);
                }
        }
        else if (strcmp(v1, "deny") == 0) {
                node = (struct element *) malloc(sizeof (struct element));
                if (node == NULL) {
                        fprintf(stderr, "Panic -- out of memory\n");
                        exit(2);
                }
                bzero(node, sizeof(struct element));
                if (isip(v2)) {
                        strcpy(node->address, v2);
                        node->type = ACCESS_CONTROL_IP_TYPE;
                }
                else {
                        strrcpy(node->address, v2);
                        node->type = ACCESS_CONTROL_DOMAIN_TYPE;
                }
                node->next = mip_config.denylist;
                mip_config.denylist = node;
        }
        else if (strcmp(v1, "allow") == 0) {
                node = (struct element *) malloc(sizeof (struct element));
                if (node == NULL) {
                        fprintf(stderr, "Panic -- out of memory\n");
                        exit(2);
                }
                bzero(node, sizeof(struct element));
                if (isip(v2)) {
                        strcpy(node->address, v2);
                        node->type = ACCESS_CONTROL_IP_TYPE;
                }
                else {
                        strrcpy(node->address, v2);
                        node->type = ACCESS_CONTROL_DOMAIN_TYPE;
                }
                node->next = mip_config.allowlist;
                mip_config.allowlist = node;
        }
        else {
                fprintf(stderr, "config error line %d, unrecognized "
                                "access_policy directive %s\n", line, v1);
                exit(1);
        }
}


/*
 * parse the ha/fa agent line
 *
 * reject any configuration that has same interface set
 * twice to ha or to fa.  ha/fa on same interface is ok.
 * In that case, we set the "both" field of the HA interface
 * to 1 and the field of the FA interface to 2.  In this way,
 * one interface may be preferred if the two are configured
 * differently.  Currently, the HA i/f determines whether arp is
 * on or off.  Currently all interfaces share a COA, lifetime, and a choice
 * between multicast or broadcast beacons.
 */

static
doAgentLine(char *op, char *linebuf, int line, struct mip_ifs * mifs, int amfa)
{
	char            dc[1024];
	char            v1[1024];
	char            v2[1024];
	char            v3[1024];
	int             rc;
	char           *e;
	int		i;
	char		*ptr;
	struct mip_ifs  *mips = &mip_config.mip_ifs[0];

	bzero(v1, sizeof(v1));
	bzero(v2, sizeof(v1));
	bzero(v3, sizeof(v1));

	if (amfa) {
		e = "fa";
	} else {
		e = "ha";
	}

	sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);
	if (mip_config.mip_ifcount >= (MAX_INTERFACES-1)) {
		syslog(LOG_ERR, "doAgentLine(): too many interfaces, max %d\n", MAX_INTERFACES);
		exit(1);
	}
	if (strlen(v1) == 0 || strlen(v2) == 0) {
		syslog(LOG_ERR, "doAgentLine(): config error line %d, agent line needs 3 parameters: %s <ifname> <router advertisement address>\n", line, e);
		exit(1);

	}
	mifs->ifname = malloc(strlen(v1) + 1);
	mifs->router = malloc(strlen(v2) + 1);
	strncpy(mifs->ifname, v1, 1024);
	strncpy(mifs->router, v2, 1024);
	if (strcmp(mifs->router, "nobeacon") == 0) {
		mifs->router = NULL;
	} else if (strcmp(mifs->router, "255.255.255.255") == 0) {
		mip_config.sawBroadcast = 1;
	} else if (strcmp(mifs->router, "224.0.0.1") == 0) {
		mip_config.sawMcast = 1;
	} else {
		syslog(LOG_ERR, "doAgentLine(): config error line %d: only support 255.255.255.255 or 224.0.0.1 as interface router advertisement\n", line);
		exit(1);

	}
	if (mifs->router != NULL) {
		rc = resolveAddress(mifs->router, &mifs->r_sin);
		if (rc <= 0) {
			syslog(LOG_ERR, "doAgentLine(): invalid route advert address %s\n", mifs->router);
			exit(1);
		}
	}
	if (amfa)
		mifs->amfa = 1;
	else
		mifs->amfa = 0;
	/* noarp or arp */
	if (strcmp(v3, "noarp") == 0) {
		mifs->doArp = FALSE;
		set_arp(mifs->ifname, "-arp", 128);
	} else {
		globalArp = 1;
		mifs->doArp = TRUE;
		set_arp(mifs->ifname, "arp", -128);
		if (strcmp(v3, "proxyarp") == 0) {
			mifs->doProxy = TRUE;
			if (mips->amfa) {
				syslog(LOG_ERR, "config error line %d, can't do proxyarp on fa interface", line);
				exit(1);
			}
#ifdef IPSEC
			mip_config.proxyexists++;
#endif
		}
	}
	mip_config.mip_ifcount++;
	/* get the mac address dynamically */
	ptr=getmac(mifs->ifname);
	for (i=0;i<6;i++)
		mifs->mac[i]=ptr[i];

	/* for our last gasp, search the list here and note
	 * whether or not we have a duplicate.  It's an error
	 * if you see the same thing twice.  It's not an error
	 * if you see ha and this fa.  It's a possibility ...
	 */
	mifs->both = 0;
	for (i = 0; i < (mip_config.mip_ifcount-1); i++, mips++) {
                if ( strcmp(mips->ifname, mifs->ifname) == 0) {
			/* if same pointer, we are done */
			if (mips == mifs)
				break;
			/* not the same pointer. make some checks */
			/* can't have two fa lines for same if */
			if (mifs->amfa && mips->amfa ) {
				syslog(LOG_ERR, "config error line %d, can't have two fas, same interface", line);
				exit(1);
			}
			/* can't have two ha lines for same if 
			 * ( needs work - should be able to do this
			 */
			if ((mifs->amfa == 0) && (mips->amfa == 0)) {
				syslog(LOG_ERR, "config error line %d, can't have two has, same interface", line);
				exit(1);
			}

			/* sanity check shared fields which should be
			 * identical: broadcast type and whether arp is on
			 * or off.
			 */
			if ( (mips->doArp != mifs->doArp) ||
				(mips->router && mifs->router &&
				strcmp(mips->router, mifs->router)) )
			{
				syslog(LOG_ERR, "config error line %d: conflict in b-cast or ARP setting for shared i/f\n", line);
				exit(1);
			}
				
			/* mark the interfaces slightly differently
			 * according to their agent function: 1 == ha,
			 * 2 == fa.
			 */
			if (mifs->amfa) {
				mips->both = 1;  
				mifs->both = 2;  
				mifs->nextAgent = mips;
				mips->nextAgent = mifs;
			}
			else {
				mips->both = 2;  
				mifs->both = 1;  
				mifs->nextAgent = mips;
				mips->nextAgent = mifs;
			}
		}
        }
}

static
doCoa(char *op, char *linebuf, int line)
{
	char            dc[1024];	/* coa line, don't care */
	char            v1[1024];	/* wannabe ip or dns */
	int             rc;

	sscanf(linebuf, "%s %s", dc, v1);
	if (strlen(v1) == 0) {
		syslog(LOG_ERR, "doCoa(): config error line %d, coa line needs 2 parameters: coa <ip-address>\n", line);
		exit(1);

	}
	rc = resolveAddress(v1, &mip_config.fa_coa);
	if (rc <= 0) {
		syslog(LOG_ERR, "doCoa(): invalid coa address %s\n", v1);
		exit(1);
	}
}

#ifdef HARP
static
doCoHa(char *op, char *linebuf, int line)
{
       char            dc[1024];       /* coha line, don't care */
       char            v1[1024];       /* wannabe ip or dns */
       int             rc;

       sscanf(linebuf, "%s %s", dc, v1);
       if (strlen(v1) == 0) {
	       syslog(LOG_ERR, "doCoHa(): config error line %d, coha line needs2 parameters: coha <ip-address>\n", line);
	       exit(1);
       }
       rc = resolveAddress( v1, &mip_config.coha );
       if (rc < 0) {
	       syslog(LOG_ERR, "doCoHa(): invalid coha address %s\n", v1);
	       exit(1);
       }
}
static
doHarpTunnel( char *op, char *linebuf, int line ) 
{
	char 	dc[1024];	
	char	v1[1024];

	sscanf( linebuf, "%s %s", dc, v1 );
	if ( strlen( v1 ) == 0) {
		syslog(LOG_ERR, "doHarpTunnel(): config error, no argument" );
		exit(1);
	}
	if ( v1[0] == 'y' ){
		mip_config.harp_tunnel = TRUE;
		if ( debug ){
			printf( "subnet is partitioned\n" );
		}
	}
	else{
		mip_config.harp_tunnel = FALSE;
		if ( debug ){
			printf( "subnet is not partitioned\n" );
		}
	}
}
static
doHarp2( char *op, char *linebuf, int line )
{
	char dc[1024];
	char v1[1024];
	sscanf( linebuf, "%s %s", dc, v1 );

	if ( strlen( v1 ) == 0) {
		syslog(LOG_ERR, "doHarp2(): config error, no argument" );
		exit(1);
	}
	strcpy( mip_config.harp2_if, v1 );
	if ( debug ){
		printf( "Arp is disabled at the secondary\n" );
	}
}

static
doHarpPing( char *op, char *linebuf, int line )
{
	char dc[1024];
	char v1[1024];
	sscanf( linebuf, "%s %s", dc, v1 );

	if ( strlen( v1 ) == 0) {
		syslog(LOG_ERR, "doHarpPing(): config error, no argument" );
		exit(1);
	}
	mip_config.harp_ping = atoi( v1 );
	if ( debug ){
		printf( "Harp ping interval set to %d\n", atoi( v1 ));
	}
}
#endif	/* HARP */


/*
 * take in security association key
 * if keyflag is set assume syntax is:
 * mnha_key ipaddr md5 256 9700ca3a82f7cf9b2d0cc739abd5f33f
 * else
 * mnha_key md5 256 9700ca3a82f7cf9b2d0cc739abd5f33f
 * 
 */
static
dokey(char *op, char *linebuf, int line, int type, int keyflag)
{
	char            dc[1024];	/* don't care */
	char		ipline[1024];   /* ipaddr */
	char            v1[1024];	/* transform */
	char            v2[1024];	/* spi */
	char            v3[1024];	/* key */
	int             l;
	int             rc;
	int             spi_wannabe;
	char           *p;
	struct 		sockaddr_in sa;
	extern int debug;

	v1[0] = v2[0] = v3[0] = ipline[0] = 0;
	if (keyflag) {
		sscanf(linebuf, "%s %s %s %s %s", dc, ipline, v1, v2, v3);
		if (strlen(ipline) == 0) {
			syslog(LOG_ERR, "config error line %d, invalid key line: \n", line);
			syslog(LOG_ERR, "example: mnha_key ipaddr md5 256 hexkey...\n");
			exit(1);
		}
		rc = resolveAddress(ipline, &sa);
		if (rc <= 0) {
			syslog(LOG_ERR, "config error line %d, invalid key line:\n", line);
			syslog(LOG_ERR, "ipaddr %s failed to resolve", ipline);
			exit(1);
		}
	}
	else {
		sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);
	}
	if (strlen(v1) == 0 || strlen(v2) == 0 || strlen(v3) == 0) {
		syslog(LOG_ERR, "config error line %d, invalid key line\n",line);
		syslog(LOG_ERR, "example: mnha_key md5 256 hexkey...\n");
		exit(1);

	}
	/*
	 * spi cannot be in range 0..255
	 */
	spi_wannabe = atoi(v2);
	if (spi_wannabe <= 255) {
		syslog(LOG_ERR, "config error line %d, bad spi:\n", line);
		exit(1);
	}
	l = strlen(v3);
	p = v3;
	/*
	 * skip over initial hex 0x if present
	 */
	if (v3[0] == '0' && v3[1] == 'x' || v3[1] == 'X') {
		p += 2;
		l -= 2;
	}
	/*
	 * WARNING: this is md5 specific
 	 *
	 * length of key string should be 32
	 */
	if (l != 32) {
		syslog(LOG_ERR, "doKey(): config error line %d, key string should be 32 chars long\n", line);
		exit(1);
	}
	rc = pack_string(p, mip_config.sa[type].key, 32);
	if (rc < 0) {
		syslog(LOG_ERR, "config error line %d, invalid key string\n", line);
		exit(1);
	}
	mip_config.sa[type].spi = spi_wannabe;

	/* for per ipaddr types, the sa/key variable
	 * functions as temp placeholder.
	 */
	if (keyflag) {
		u_long	spi[4];
		bzero(spi, sizeof(u_long)*4);
		spi[0] = spi_wannabe;

		rc = enter_keylist(
			mip_config.sa[type].kl,	/* keylist pointer */
			sa.sin_addr.s_addr,     /* ip addr in net order */
			mip_config.sa[type].key,  /* the key */
			16, 	/* HARDWIRED to MD5 for now */
			spi, 
			1,	/* HARDWIRED to MD5 for now */
			0L,	/* ipaddr2: ip src address (not used here) */
			1	/* inuse: field not used for mobileIP keys */
			);
		if (rc <= 0) {
			syslog(LOG_ERR, "config error line %d, invalid key string\n", line);
			if (rc < 0) {
				syslog(LOG_ERR, "malloc - out of space");
			}
			else {
				syslog(LOG_ERR, "duplicate ip addr entry");
				if (debug)
					printf("duplicate ip addr entry\n");
			}
			exit(1);
		}
	}
}

#ifdef IPSEC
/*
 * handle ipsec_ha, ipsec_fa lines
 */
void
ipsecConfig(char *tag, char *linebuf, int linenum, int ptype)
{
	int	rval,
		sectype,
		ahflag = 0,
		espflag = 0,
		transform,
		base,
		index,
		spiOffset,
		numSpiArgs,
		i,
		lio_saved;	/* saved value of logIPOnly config switch */
	char 	dummy_key[4];
	u_long	spi[4];			/* ah/esp outbound/inbound spis */
	char	v[8+1][MAXLINE];	/* 8 arguments & 1 don't care */
	struct sockaddr_in	src;
	struct sockaddr_in	dst;
	int policy;
        char *who = tag;
        int tuncoa = 0;

	bzero(dummy_key, 4);
	bzero(spi, sizeof(u_long)*4);

	if (!ipsecKeysInuse) {
		ipsecKeysInuse = TRUE;
		openKeySocket();
	}

	/* ipsec_mn 2nd_tag <ip src> <ip dst> (ah|esp|ah+esp)		\
	 *		<spi_o> <spi_i> [<spi_o> <spi_i>]
	 */
	rval = sscanf(linebuf, "%s %s %s %s %s %s %s %s %s",
		    v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8]);

	if (rval < 3) {
		syslog(LOG_ERR, "%s config error: incorrect number of arguments (line %d).", who, linenum);
		exit(1);
	}
        /* deal with strict|lax possibility and get it out of the way
	 */
	if (strcmp(v[1], "mn_policy") == 0) {
		if (strcmp(v[2], "strict") == 0) {
			policy = TRUE;
		} else if (strcmp(v[2], "lax") == 0) {
			policy = FALSE;
		} else { 
			syslog(LOG_ERR, "%s config error: unrecognized mn_policy (line %d).", who, linenum);
			exit(1);
		}
		/* set strict flag and turn on inuse if strict was
		 * set to true
		 */
		switch (ptype) {
		case HA_SECTYPE_MN:
			mip_config.ipsec[HA_SECTYPE_MN].strict = policy;
			if (policy) {
				mip_config.ipsec[HA_SECTYPE_MN].inuse 
					= TRUE;
			}
			break;
		case FA_SECTYPE_MN:
			mip_config.ipsec[FA_SECTYPE_MN].strict = policy;
			if (policy) {
				mip_config.ipsec[FA_SECTYPE_MN].inuse
					= TRUE;
			}
			break;
		default:
			syslog(LOG_ERR, "%s config error: no such combination (line %d).", who, linenum);
			exit(1); 
		}
		return;
	}

	/* long form from here on out 
	*/
	if ((ptype == TUN_SECTYPE) && (rval < 6)) {
		syslog(LOG_ERR, "%s config error: too few arguments (line %d).", who, linenum);
		exit(1); 
	}
	else if (ptype != TUN_SECTYPE && rval < 7) {
		syslog(LOG_ERR, "%s config error: too few arguments (line %d).", who, linenum);
		exit(1); 
	}

	/* check to see if tun coa command
	*/
	if (ptype == TUN_SECTYPE) {
		if (strcmp(v[1], "coa") == 0) {
			tuncoa = TRUE;
		}
	}

	mip_config.ipsec[ptype].inuse = TRUE;

	bzero(&dst, sizeof(dst));
	bzero(&src, sizeof(src));
	if (!resolveAddress(v[2], &src)) {
		syslog(LOG_ERR, "%s config error: could not resolve ip src (line %d).", who, linenum);
		exit(1);
	}
	if (!resolveAddress(v[3], &dst)) {
		syslog(LOG_ERR, "%s config error: could not resolve ip dst (line %d).", who, linenum);
		exit(1);
	}

	/* Check that there is no other key defined for this node under the
	 * given policy type.  Note that ipsec_tun coa keys still cannot be
	 * checked in this way because the FA does not cache them.
	 */
	if ( search_keylist(mip_config.ipsec[ptype].kl, 
		(ptype == TUN_SECTYPE && !tuncoa)
			? src.sin_addr.s_addr:dst.sin_addr.s_addr) )
	{
		if (debug) {
			printf("ipsecConfig(): (config err, line %d) \"%s %s\" already defined for node %s.\n",
				linenum, v[0], v[1], inet_ntoa(dst.sin_addr));
		}
		syslog(LOG_ERR,
			"ipsecConfig(): (config err, line %d) \"%s %s\" already defined for node %s.\n",
			linenum, v[0], v[1], inet_ntoa(dst.sin_addr));

		exit(1);
	}

	/* determine security type
	*/
	sectype = NOSEC;
	numSpiArgs = 2;
	if (strcmp(v[4], "ah") == 0) {
		sectype = AHSEC;
	}
	else if (strcmp(v[4], "esp") == 0) {
		sectype = ESPSEC;
	}
	else if (strcmp(v[4], "ah+esp") == 0) {
		sectype = (AHSEC|ESPSEC);
		numSpiArgs = 4;
	}
	else {
		syslog(LOG_ERR, "%s config error: unsupported transform type (line %d).", who, linenum);
		exit(1);
	}

        /* tun type has only 1/2 as many spis
	 * (XXX hack alert!  battle stations!)
         */
        if (ptype == TUN_SECTYPE) {
                numSpiArgs /= 2;
        }

	/* get spi
	 *
	 * store spis in order: ah out, ah in, esp out, esp in.
	 */
	spiOffset = (sectype==ESPSEC)?2:0;
	for (i = 5; i < 5+numSpiArgs; i++ ) {
		base = 10;
		index = i-5+spiOffset;

		if ( *(v[i]+1) == 'x' || *(v[i]+1) == 'X' ) base = 16;
		spi[index] = strtoul(v[i], NULL, base);
		if (spi[index] < MINSPI) {
			syslog(LOG_ERR, "Config error: spi #%d is too small (tag=%s, line %d).", i-4, tag, linenum);
			exit(1);
		}
                if (index % 2) {
			/* Simple spi value sanity check, incoming and
			 * outgoing spis must differ for ah or esp.
			 * However, with ipsec_tun keys the pair represents
			 * ah AND esp, an thus may be equal, so skip the test.
			 */
                        if ( (ptype != TUN_SECTYPE)
					&& spi[index] == spi[index-1] 
					&& spi[index])
			{
                                syslog(LOG_ERR, "ipsecConfig(): spis for a given transform may not be equal (line=%d).",
                                        linenum);
                                exit(1);
                        }
                }
	}

	/* (XXX hack alert!)
	 *
	 * Redistribute the stored spi values for tunnel routes depending
	 * upon whether it (they) represent ah and/or esp and whether
	 * at the FA (incoming) or at the HA (outgoing).  Thusly, the
	 * keys will be checked properly below.
	 */
	if (ptype == TUN_SECTYPE) {
		/* If ah+esp, esp must be moved. */
		if (sectype == (AHSEC|ESPSEC)) {
			spi[2] = spi[1];
			spi[1] = 0;
		}

		/* If we are a FA, shift spis from outgoing position to
		 * incoming position (doesn't matter how many transforms
		 * or which).  Else if we are a HA, the spi(s) are already
		 * positioned properly.
		 */
		if (tuncoa) {
			spi[1] = spi[0];
			spi[3] = spi[2];
			spi[0] = spi[2] = 0;
		}
	}


        /* in all cases, except for tun coa type (where we don't want
         * keylist entry because we only want to verify that the key exists), 
         * stick tuple in keylist.  we are indexed by IP dst (MN) address,
	 * so each policy list can not have the same ip address
	 * in it more than once.  we don't care about the key (yet?).
	 * we store out transform value acc. to our own non-NRL sectype
	 * value.  we also store the ip src which anchors this end.
	 */
	if (!tuncoa) {
		rval = enter_keylist(
			mip_config.ipsec[ptype].kl, /* keylist pointer */
			dst.sin_addr.s_addr,    /* dst addr of SA */
			dummy_key,		/* the key. DONT CARE actually */
			4,     		 	/* keysize */
			spi,			/* SA spi */
			sectype,		/* overload transform for now */
			src.sin_addr.s_addr,	/* src side of SA */
			1			/* mipd keys are always "inuse" */
			);

		if (rval <= 0) {
			syslog(LOG_ERR, "%s config error, line %d\n", who, linenum);
			if (rval < 0)
				syslog(LOG_ERR, "malloc - out of space");
			else
				syslog(LOG_ERR, "duplicate ip address");
			exit(-1);
		}
	}

	/* access key socket and make sure that key is actually
	 * loaded at config time.  Note we do this with DNS
	 * reverse address lookups OFF.
	 *
	 * Because the spi[] array orders the values according to
	 * type and direction, the position in the array can be used
	 * to communicate this same information to verifyKey().
	 */
	lio_saved = mip_config.logIPOnly;
	mip_config.logIPOnly = TRUE;

	for (i = 0; i < 4; i++) {
		if (spi[i] == 0)
			continue;

		if (tuncoa) {
			rval = verifyKey(dst.sin_addr, src.sin_addr,
				spi[i], i<2, (i+1)%2);
		}
		else {
			rval = verifyKey(src.sin_addr, dst.sin_addr,
				spi[i], i<2, (i+1)%2);
		}

		if (rval < 0) {
			syslog(LOG_ERR, "ipsecConfig(): No such key in kernel (tag=%s, line %d).", tag, linenum);
			exit(1);
		}
	}

	mip_config.logIPOnly = lio_saved;
}
#endif /* IPSEC */


/*
 * debug - dump configuration variables
 */
configDump()
{
	int             i;
	int		j;

	printf("---------config dump-------------------\n");
	printf("logIPOnly: %s\n", mip_config.logIPOnly ? "on" : "off");
	printf("sawHa %d\n", mip_config.sawHa);
	printf("sawFa %d\n", mip_config.sawFa);
	printf("sawCoa %d\n", mip_config.sawCoa);
	printf("beaconRate %d\n", mip_config.beaconRate);
	printf("mip_ifcount %d\n", mip_config.mip_ifcount);
	printf("UDP ntp_slop %d\n", mip_config.ntp_slop);
	if (mip_config.sawLocation) {
		printf("location %s\n", mip_config.location);
	}
	printf("mobile lifetime %d\n", mip_config.mobileLifetime);
	if (mip_config.sawCoa) {
		printf("coa in hex=(%x)\n", mip_config.fa_coa.sin_addr.s_addr);
	}
	printf("doTimestamp %d\n", mip_config.doTimestamp);
	printf("beacon timeslop %d\n", mip_config.beacon_timeslop);
	printf("beacon timestamp %d\n", mip_config.beacon_timestamp);
	printf("beacon Replay protection %d\n", mip_config.beacon_replay);
#ifdef IPSEC
	printf("proxyexists flag %d\n", mip_config.proxyexists);
#endif
	/*
	 * print out interface values
	 */
	for (i = 0; i < mip_config.mip_ifcount; i++) {
		if (mip_config.mip_ifs[i].amfa) {
			printf("fa\tifname=%s\nrouter advert addr=%s\n",
			       mip_config.mip_ifs[i].ifname,
			       mip_config.mip_ifs[i].router);
		} else {
			printf("ha\tifname=%s\trouter advert addr=%s\n",
			       mip_config.mip_ifs[i].ifname,
			       mip_config.mip_ifs[i].router);
		}
		printf("found %d\n", mip_config.mip_ifs[i].found);
		printf("doArp %d\n", mip_config.mip_ifs[i].doArp);
		printf("doProxy %d\n", mip_config.mip_ifs[i].doProxy);
		printf("netmask %s\n", 
			inet_ntoa(mip_config.mip_ifs[i].netmask.sin_addr));
		printf("cloneip %s\n", 
			inet_ntoa(mip_config.mip_ifs[i].cloneip.sin_addr));
 
 		printf("mac addr: "); 
		for (j = 0; j < 6; j++) {
			printf("%02x%c", 
				mip_config.mip_ifs[i].mac[j], (j<5)?'.':' ');
		}
		printf("\n");
	}
	/*
	 * print out security values
	 */
	if (mip_config.mnha_security) {
		printf("mnha keylist\n");
		dump_keylist(mip_config.sa[MNHASECURITY].kl);
	}
	if (mip_config.faha_security) {
		printf("faha keylist\n");
		dump_keylist(mip_config.sa[FAHASECURITY].kl);
	}
	if (mip_config.famn_security) {
		printf("famn keylist\n");
		dump_keylist(mip_config.sa[FAMNSECURITY].kl);
	}
	if (mip_config.netauth_security) {
		i = 0;
		printf("netauth_key spi %d, key=", mip_config.sa[NETAUTHSECURITY].spi);
		while (i < 16) {
			printf("%2.2x", mip_config.sa[NETAUTHSECURITY].key[i]);
			i++;
		}
		printf("\n");
	}
	if (mip_config.adhoc_security) {
		i = 0;
		printf("adhoc_key spi %d, key=", mip_config.sa[ADHOCSECURITY].spi);
		while (i < 16) {
			printf("%2.2x", mip_config.sa[ADHOCSECURITY].key[i]);
			i++;
		}
		printf("\n");
	}
#ifdef IPSEC
	if (mip_config.ipsec[HA_SECTYPE_MN].inuse) {
		printf("ipsec_ha mn keylist, policy (strict) %d\n",
			mip_config.ipsec[HA_SECTYPE_MN].strict);
                dump_keylist(mip_config.ipsec[HA_SECTYPE_MN].kl);
	}
	if (mip_config.ipsec[FA_SECTYPE_MN].inuse) {
		printf("ipsec_fa mn keylist, policy (strict) %d\n",
			mip_config.ipsec[FA_SECTYPE_MN].strict);
                dump_keylist(mip_config.ipsec[FA_SECTYPE_MN].kl);
	}
        if (mip_config.ipsec[TUN_SECTYPE].inuse) {
                printf("ipsec_tun keylist, policy (strict) %d\n",
                        mip_config.ipsec[TUN_SECTYPE].strict);
                dump_keylist(mip_config.ipsec[TUN_SECTYPE].kl);
        }
#endif
	printf("---------config dump-------------------\n");
}


/* configInterface()
 * called from getIfs() routine that walks kernel ifconfigs. 
 * checks to see if ifname is in mip_ifs table, 
 * if so, marks it (found) and stores ip address for i/f, netmask,
 * and clone route address.
 */
configInterface(char *ifname, struct sockaddr_in *sin, struct sockaddr_in *netmask, int mcast, int link_index)
{
	struct mip_ifs *mips;
	extern int debug;

	for (mips = &mip_config.mip_ifs[0];
		mips < &mip_config.mip_ifs[mip_config.mip_ifcount];
		mips++) {
		if (strcmp(ifname, mips->ifname) == 0) {
			mips->found = 1;	/* check it off */
			/* store sockaddrs netmask net link_index */
			bcopy(sin, &mips->if_sin, sizeof (struct sockaddr_in));
			if (debug) {
				printf("%lx\n", mips->if_sin.sin_addr);
				printf("ifname %s, ip %s\n", mips->ifname, inet_ntoa(mips->if_sin.sin_addr));
			}
			bcopy(netmask, &mips->netmask, sizeof (struct sockaddr_in));
			mips->cloneip.sin_addr.s_addr = sin->sin_addr.s_addr &
				netmask->sin_addr.s_addr;

			mips->link_index = link_index;
			/* blow up if multicast requested and no
			 * multicast on device
			 */
			if (mip_config.sawMcast && !mcast) {
				if (debug) {
					printf("configInterfaces - no multicast on device\n");
				}
				syslog(LOG_ERR, "multicast beacons wanted but no multicast on %s - use 255.255.255.255", ifname);
				exit(1);
			}
		}
	}
}

void
mapIfname(char *ifname, int *index, unsigned long *ifaddr)
{
	struct mip_ifs *mips;
	
	for (mips = &mip_config.mip_ifs[0];
			mips < &mip_config.mip_ifs[mip_config.mip_ifcount];
			mips++) {
		if (strcmp(ifname, mips->ifname) == 0) {
			*index = mips->link_index;
			*ifaddr = mips->if_sin.sin_addr.s_addr;
			return;
		}
	}
	syslog(LOG_ERR, "mipIfname: interface doesn't exist");
	exit(1);
}

/*
 * verifyIfConfig()
 * 
 * make sure that all config-file mip interfaces are indeed real interfaces
 */
verifyIfConfig()
{
	struct mip_ifs *mips;
	int             fatal = 0;

	for (mips = &mip_config.mip_ifs[0];
			mips < &mip_config.mip_ifs[mip_config.mip_ifcount];
			mips++) {
		if (mips->found == 0) {
			fatal++;
			syslog(LOG_ERR, "verifyIfConfig(): config i/f %s not available in kernel ifconfigs\n", mips->ifname);
		}
		mips++;
	}
	if (fatal) {
		syslog(LOG_ERR, "verifyIfConfig(): daemon terminated\n");
		exit(1);
	}
}

/*
 * life is interesting.  we need to check the if
 * structure and see if any interfaces are BOTH ha/fa
 * and if so note that we do zap the clone route once
 * and then install a REJECT version of it for HAness.  
 * in summary:
 *	if interface is HA (even if FA)
 *		zap clone route
 *		install REJECT subnet route so that packets to
 *			that subnet are at least sinked.
 *	if interface is FA only
 *		zap clone route
 *
 * NOTE: the reason we install the REJECT route on a HA i/f
 * is because we MUST have a routing entry for the HA/subnet
 * else packets would be sent (wrongly) to the default route.
 */
int     
zapClones()
{
	int             i;
	struct mip_ifs *mips = &mip_config.mip_ifs[0];

	for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
                if ( mips->doArp == FALSE) {
			switch(mips->both) {
			case 0:
				clone_route_delete(
					mips->cloneip.sin_addr.s_addr);
				/* if single HA, we must put in reject
				 * route
				 */
				if (mips->amfa == 0) {
					reject_route_add(
					mips->cloneip.sin_addr.s_addr,
					mips->if_sin.sin_addr.s_addr,
					mips->netmask.sin_addr.s_addr
					);
				}
				break;
			case 1:
				/* do nothing */
				break;
			/* for combined ha/fa we must zap clone route
			 * and put in reject route of same form; i.e.,
			 * do ha setup.
			 */
			case 2:
				/* zap clone */
				clone_route_delete(
					mips->cloneip.sin_addr.s_addr);
				/* install reject route */
				reject_route_add(mips->cloneip.sin_addr.s_addr,
					mips->if_sin.sin_addr.s_addr,
					mips->netmask.sin_addr.s_addr
					);
				break;
			}
		}
        }
}

/*
 * put clone routes back when daemon is shutdown
 */
int
unzapClones()
{
	int             i;
	struct mip_ifs *mips = &mip_config.mip_ifs[0];

	for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
                if ( mips->doArp == FALSE) {
			switch(mips->both) {
			case 0:
				/* if it was a ha only, zap the reject route
				*/
				if (mips->amfa == 0) {
					clone_route_delete(
						mips->cloneip.sin_addr.s_addr);
				}
				clone_route_add(mips->cloneip.sin_addr.s_addr,
					mips->if_sin.sin_addr.s_addr,
					mips->netmask.sin_addr.s_addr
#ifdef IPSEC
					, NULL	/* always NULL */
#endif
					);
				break;
			case 1:  /* do nothing */
				break;
			case 2:
				/* delete reject route */
				clone_route_delete(
						mips->cloneip.sin_addr.s_addr);
				/* install clone route */
				clone_route_add(mips->cloneip.sin_addr.s_addr,
					mips->if_sin.sin_addr.s_addr,
					mips->netmask.sin_addr.s_addr
#ifdef IPSEC
					, NULL	/* always NULL */
#endif
					);
				break;
			}
		}
        }
}  /* end unzapClones() */

/*
 * get location string  if no such string, return NULL, else return string
 */
char           *
configLocation()
{
	if (mip_config.sawLocation) {
		return (mip_config.location);
	}
	return (NULL);
}

/*
 * test for icmp security association
 * 
 * returns: NULL - no such association *spi indirect, and pointer to key
 */

unsigned char  *
configIcmpAuth(int *spi)
{
	if (mip_config.netauth_security) {
		*spi = mip_config.sa[NETAUTHSECURITY].spi;
		return (mip_config.sa[NETAUTHSECURITY].key);
	}
	return (NULL);
}

unsigned char  *
configAdhocAuth(int *spi)
{
	if (mip_config.adhoc_security) {
		*spi = mip_config.sa[ADHOCSECURITY].spi;
		return (mip_config.sa[ADHOCSECURITY].key);
	}
	return (NULL);
}

int getBeaconRate()
{
	return(mip_config.beaconRate);	
}

int getNetauthSecurity()
{
	return(mip_config.netauth_security);	
}

int getAdhocSecurity()
{
	return(mip_config.adhoc_security);	
}

/*
 * return 1 if netauth_security not in place.
 *
 * these functions for now return 1 if
 * adhoc_security or netauth_security switches are set to 0
 */
checkNetauthReject()
{
	if (mip_config.netauth_security)
		return(0);
	return(1);
}

checkAdhocReject()
{
	if (mip_config.adhoc_security)
		return(0);
	return(1);
}

char	*getNetauthKey()
{
	if (mip_config.netauth_security)
		return(mip_config.sa[NETAUTHSECURITY].key);	
	else return(NULL);
}

char	*getAdhocKey()
{
	if (mip_config.adhoc_security)
		return(mip_config.sa[ADHOCSECURITY].key);	
	else return(NULL);
}

int	getNetauthSpi()
{
	if (mip_config.netauth_security)
		return(mip_config.sa[NETAUTHSECURITY].spi);	
	else return(0);
}

int	getAdhocSpi()
{
	if (mip_config.adhoc_security)
		return(mip_config.sa[ADHOCSECURITY].spi);	
	else return(0);
}

/*
 * getConfigArp() called from routine that need to know whether arp is
 * on or off on a particular ifaddr side
 *
 * returns:
 *	1: arp is on
 *	0: arp is not on
 *     -1: no such interface
 */
int     getConfigArp(struct in_addr  *ifaddr)
{
	int             i;
	struct mip_ifs *mips = &mip_config.mip_ifs[0];

	for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
                if ( mips->if_sin.sin_addr.s_addr == ifaddr->s_addr) {
			/* if an interface is being used in the role 
			 * of both HA and FA, return the value configured
			 * for the HA.
			 */
			if (mips->both == 2) {
				if (!mips->nextAgent) {
					return(-1);
				}
				return(mips->nextAgent->doArp);
			}
			else {
                        	return(mips->doArp);
			}
		}
        }
	return(-1);
}

/*
 * getConfigIf() called from routine to match ifaddr to
 * config structure.
 */
struct mip_ifs *
getConfigIf(struct in_addr  *ifaddr)
{
	int             i;
	struct mip_ifs *mips = &mip_config.mip_ifs[0];

	for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
                if ( mips->if_sin.sin_addr.s_addr == ifaddr->s_addr) {
                        return(mips);
		}
        }
	return(NULL);
}

getConfiglogIPOnly()
{
        return(mip_config.logIPOnly); 
}

/*
 * turn arp back on for interfaces
 * that need it.  called on daemon shutdown.
 */
void
unzapArp()
{
	int             i;
	struct mip_ifs *mips = &mip_config.mip_ifs[0];

	for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
		if (mips->doArp == FALSE){ 
			set_arp(mips->ifname, "arp", -128);
		}
        }
}

/*
 * given link index, return matching ip addr
 *
 *	returns: ip address as int
 *		 0 if no match
 */
int
indextoif(int index)
{
	int             i;
	struct mip_ifs *mips = &mip_config.mip_ifs[0];

	for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
		if (mips->link_index == index) {
			return(mips->if_sin.sin_addr.s_addr);
		}
        }
	return(0);
}

#ifdef MADRP            
int IsAgent(){ return TRUE;} 
#endif  /* MADRP */


#ifdef TESTMAIN
int             debug = 1;
logperror(char *s) 
{
	printf("%s\n", s);
}

main(argc, argv)
	int             argc;
	char          **argv;
{
        /* open a local log file */
        if ( mip_openlog( "./mipd.log.local" ) < 0 ) {
            syslog( LOG_ERR,
                "Failed to open local logfile \"%s\".", "./mipd.log.local" );
            exit(1);
        }

	configInit();
	configRead(argv[1]);
	getIfs();
	configDump();

	mip_closelog();
}
#endif

