/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * route.c
 *
 * strategy:
 *	for now use route command.  eventually replace with
 *	explicit calls to route socket.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef IPSEC
#include "ipsec.h"
#endif

#include "keylist.h"
#include "mip_log.h"
#include "route.h"
#include "alist.h"

/*	getting the interface out of the config file */
#include "mnconfig.h"

#define TRUE	1
#define FALSE	0

static int      noticeFunction = 3;
extern int      debug;


/* prototypes
*/
void flushArp();  /* flush the entire arp table */



/*
 * route_add_link()
 * add the link route of the nature
 *	arp -s <mn ip> mac -if ifname
 *	arp -s <mn ip> NULL -if ifname   
 *
 * This routine may be used by an agent or an mn that simply
 * wishes to put an LLINFO (with a S for static flag) into
 * the routing table.  The caller may choose to fill in the MAC
 * address or let arp do it.  There are NO dependencies on the clone
 * route by the way.
 *
 * The mac address should be set to the NULL pointer if arp is
 * being used.  If the mac address is known, it should be supplied
 * as this will avoid the use of arp.  
 * 
 * This is actually an arp table entry in FreeBSD.
 *
 * If the final argument is non-NULL, it is used to form a tuple for
 * associating a security association with the link by appending the
 * following arguments to the above arp commandline statement:
 *
 *   -type (ah|esp) -id <hex|dec> -itsrc <IPaddr> -itdst <IPaddr>
 *
 */

void
route_add_link(struct in_addr mn, char *mac, char *ifname
#ifdef IPSEC
	, struct routeSA *rosa
#endif
	)
{
	char command[1024];
	char gateway[1024];
	char	netaddr[MAXBUF],
		AHstring[MAXBUF] = "",	/* for IPSEC */
		ESPstring[MAXBUF] = "";	
	int rc;

	if (mac == NULL)
		mac = "NULL";
	else {
		sprintf(gateway, "%s", ether_ntoa(mac));
		mac = gateway;
	} 

#ifdef IPSEC
	/* build security tuple string.
	 * it may be defined as not necessary.
	 */
	if (rosa && (rosa->transform & (AHSEC|ESPSEC))) {
		/* FIX -- yet to use AH on constructed commandline */
		if (rosa->transform & AHSEC) {
			syslog(LOG_ERR, "route_add_link(): SA tuple AH bit not supported yet.");
			return;	
		}
		if (rosa->transform & AHSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_AH].src));
			sprintf(AHstring,
				"-type ah -id 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_AH].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_AH].dst));
		}
		if (rosa->transform & ESPSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_ESP].src));
			sprintf(ESPstring,
				"-type esp -id 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_ESP].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_ESP].dst));
		}
	}
#endif /* IPSEC */

	/*sprintf(command, 
	  "/usr/sbin/arp -s %s %s -if %s %s> /dev/null", 
		inet_ntoa(mn), mac, ifname, ESPstring); */

	sprintf(command, 
	  "/sbin/arp -s %s %s -i %s > /dev/null", 
		inet_ntoa(mn), mac, ifname); 

	if (debug) {
		printf("route_add_link: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog(LOG_ERR, "route_add_link(): route command (%s) failed, rc %d\n", command, rc);
	}
}

/*
 * route_delete_link()
 * delete the route specified by in_addr mn
 * 
 * This is actually an arp table entry in FreeBSD.
 */
void
route_delete_link(struct in_addr mn, char *mac)
{
	char command[1024];
	int rc;

	sprintf(command, "/sbin/route -n delete %s > /dev/null", inet_ntoa(mn));

	if (debug) {
		printf("route_delete_link: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "route_delete_link(): route command (%s) failed, rc %d\n", command, rc);
	}
}

/*
 * route_add_proxy()
 *
 * Add a proxy arp entry for a mobile node.
 */
void
route_add_proxy(struct in_addr mn, char *mac, char *ifname)
{
	char command[1024];
	char gateway[1024];
	int rc;

	if (mac == NULL) {
		syslog(LOG_ERR, "route_add_proxy(): (internal) MAC address is NULL.  Exiting...");
		exit(1);
	}

	sprintf(gateway, "%s", ether_ntoa(mac));
	sprintf(command, "/sbin/arp -s %s %s pub -if %s >/dev/null", 
		inet_ntoa(mn), gateway, ifname);

	if (debug) {
		printf("route_add_proxy: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch arp error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog(LOG_ERR, "route_add_proxy(): arp command (%s) failed, rc %d\n", command, rc);
	}
}

/*
 * route_delete_proxy()
 *
 * Delete the proxy arp entry specified by the moble node mn.
 */
void
route_delete_proxy(struct in_addr mn)
{
	char command[1024];
	int rc;

	sprintf(command, "/sbin/arp -d %s > /dev/null", inet_ntoa(mn));

	if (debug) {
		printf("route_delete_proxy: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch arp error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "route_delete_proxy(): arp command (%s) failed, rc %d\n", command, rc);
	}
}

/*
 * add a host route
 *      route add dst nexthop
 */
void
route_add_host(struct in_addr desti, struct in_addr nexthop)
{
        char command[1024];
        char gateway[1024];
        char netaddr[MAXBUF];
	extern int debug;
        int  rc;

        /* delete the route just in case it is lurking
         * in the routing table from when the mn was home
         */
        route_delete_host(desti, nexthop);

        sprintf(gateway, "%s", inet_ntoa(nexthop));
        sprintf(command, "/sbin/route -n add %s %s > /dev/null",
                inet_ntoa(desti), gateway);

#ifdef DEBUG
	if (debug)
                printf("route_add_host: <%s>\n", command);
#endif
        rc = system(command);
        /* TBD. do something clever to catch route error which
         * will be written to stderr.
         */
        if (rc == 127 || rc < 0) {
                syslog( LOG_ERR,
                        "route_add_host(): route command (%s) failed, rc %d\n",
                        command, rc);
        }
}

void
route_delete_host(struct in_addr desti, struct in_addr nexthop)
{
        char command[1024];
	extern int debug;
        int rc;

        sprintf(command, "/sbin/route -n delete %s > /dev/null",
                         inet_ntoa(desti));

#ifdef DEBUG
	if (debug)
                printf("route_delete_host: <%s>\n", command);
#endif
        rc = system(command);
        /* TBD. do something clever to catch route error which
         * will be written to stderr.
         */
        if (rc == 127 || rc < 0) {
                syslog( LOG_ERR,
                        "route_delete_host():route command (%s) failed, rc 
%d\n",
                        command, rc);
        }
}






/*
*	flush the entire arp table
*/
void flushArp()
{
	/* there is no such option in linux
	system("/usr/sbin/arp -z"); */
	return;
}

/*
 * add a tunnel route from mn_addr to gateway coa_addr
 *
 *	route add -tunnel mn_addr coa_addr
 *
 * If the final arguments are non-NULL, they is used to form tuples for
 * associating a security association with the route by appending the
 * following arguments to the above route commandline statement:
 *
 *   (-ah|-esp) -spi <hex|dec> -itsrc <IPaddr> -itdst <IPaddr>
 *
 * The tunnel route is initially added and then changed once or twice
 * to add the rosa and/or mvifrosa route SAs
 * 
 */
void
ha_route_add_tunnel(struct in_addr mn, struct in_addr coa
#ifdef IPSEC
	, struct routeSA *rosa
        , struct routeSA *mvifrosa  /* ipsec between IP headers */
#endif
	)
{
	char command[1024];
	char gateway[1024];
	char	netaddr[MAXBUF],
		AHstring[MAXBUF],	/* for IPSEC */
		mvif_AHstring[MAXBUF],
		ESPstring[MAXBUF];
	int rc;

	*AHstring = *mvif_AHstring = *ESPstring = 0;

	/* delete the route just in case it is lurking
	 * in the routing table from when the mn was home
	 */
	ha_route_delete_tunnel(mn, coa);

#ifdef IPSEC
	/* build security tuple string.
	 * it may be defined as not necessary.
	 */
	if (rosa && (rosa->transform & (AHSEC|ESPSEC))) {
		/* FIX -- use AH on constructed commandline */
		if (rosa->transform & AHSEC) {
			syslog(LOG_ERR, "ha_route_add_tunnel(): SA tuple transform AH bit not supported yet.");
			return;
		}
		if (rosa->transform & AHSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_AH].src));
			sprintf(AHstring, "-ah -spi 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_AH].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_AH].dst));
		}
		if (rosa->transform & ESPSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_ESP].src));
			sprintf(ESPstring, "-esp -spi 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_ESP].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_ESP].dst));
		}
	}
	/* XXX. for now primary MN binding overrides any HA/FA tunnel
	*/

	if ( mvifrosa && (mvifrosa->transform & AHSEC) ) {
		if (mvifrosa->transform & AHSEC) {
			strcpy(netaddr, inet_ntoa(mvifrosa->tup[ROSA_AH].src));
			sprintf(mvif_AHstring, "-ahtunnel -spi 0x%x -itsrc %s -itdst %s",
				mvifrosa->tup[ROSA_AH].spi,
				netaddr, inet_ntoa(mvifrosa->tup[ROSA_AH].dst));
		}
	}
#endif /* IPSEC */

	sprintf(gateway, "%s", inet_ntoa(coa));
        sprintf(command, "/sbin/route -n add %s %s -tunnel> /dev/null", 
                inet_ntoa(mn), 
                gateway);

	if (debug) {
		printf("ha_route_add_tunnel: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "ha_route_add_tunnel(): route command (%s) failed, rc %d\n", 
			command, rc);
	}

	if (debug) {
		printf("ha_route_add_tunnel: ESPstring=\"%s\"\n", ESPstring );
		printf("ha_route_add_tunnel: mvif_AHstring=\"%s\"\n",
			mvif_AHstring );
	}
	if (*ESPstring != '\0') {
		sprintf(command, "/sbin/route -n change %s %s >/dev/null",
			inet_ntoa(mn),
			ESPstring);
		if (debug) {
			printf("ha_route_add_tunnel: <%s>\n", command);
		}
		rc = system(command);
		if (rc == 127 || rc < 0) {
			syslog( LOG_ERR, "ha_route_add_tunnel(): route command (%s) failed, rc %d\n", 
				command, rc);
		}
	}
	if (*mvif_AHstring != '\0') {
		sprintf(command, "/sbin/route -n change %s %s >/dev/null",
			inet_ntoa(mn),
			mvif_AHstring);
		if (debug) {
			printf("ha_route_add_tunnel: <%s>\n", command);
		}
		rc = system(command);
		if (rc == 127 || rc < 0) {
			syslog( LOG_ERR, "ha_route_add_tunnel(): route command (%s) failed, rc %d\n", 
				command, rc);
		}
	}
}

void
ha_route_delete_tunnel(struct in_addr mn, struct in_addr coa
	)
{
	char command[1024];
	int rc;

	sprintf(command, "/sbin/route -n delete %s > /dev/null", inet_ntoa(mn));

	if (debug) {
		printf("ha_route_delete_tunnel: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "ha_route_delete(): route command (%s) failed, rc %d\n", 
			command, rc);
	}
}

/* used by linux when in NOWHERE state so we can still send beacons
*/
void
mn_route_add_interface_default()
{
	char command[1024];
	int rc;

	sprintf( command, "/sbin/route add default %s", mn_config.ifname );
	rc = system( command );
	if ( rc < 0 ){
		syslog( LOG_ERR, "mn_route_add_interface_default: 
				system() failed\n" );
	}
}

/* used to remove interface default route */
void 
mn_route_del_interface_default()
{
	char command[1024];
	int rc;

	sprintf( command, "/sbin/route delete default" ); 
	rc = system( command );
	if ( rc < 0 ){
		syslog( LOG_ERR, "mn_route_del_interface_default: 
			system() failed\n" );
	}
}
	

/*
 * add a mn default route to an agent
 * 	route add default agent_addr -if ifaddr 
 *
 * In order to be independent of the clone route, we always
 * attempt to delete any link layer address that has popped
 * up because of it.  We then insert the link layer address
 * as a static LLINFO with (or without) a MAC address, depending
 * on whether arp is to be used or not.
 *
 * commands used:
 *	route add default <ipaddr> -ifa <interface ipaddr>
 *	arp -s ipaddr <mac or NULL> -if name
 *
 * If the final argument is non-NULL, it is used to form a tuple for
 * associating a security association with the link by appending the
 * following arguments to the above arp and route commandline statements.
 *
 * for arp:
 *   -type (ah|esp) -id <hex|dec> -itsrc <IPaddr> -itdst <IPaddr>
 * for route:
 *   (-ah|-esp) -spi <hex|dec> -itsrc <IPaddr> -itdst <IPaddr>
 */
void
mn_route_add_default(struct in_addr agent, struct in_addr interface,
	int dolink,  char *mac, char *ifname
#ifdef IPSEC
	, struct routeSA *rosa,		/* ipsec route SA data */
	u_long link_trans		/* transforms to apply to LL route */
#endif
	)
/* agent - agent ip address */
/* interface - interface ip address to associate route with */
/* dolink - tells us to pay attention to next 2 params */
/* mac - mac address or NULL if arp is to be used */
/* ifname - string name of interface */
{
	char command[1024];
	char	netaddr[MAXBUF],
		AHstring[MAXBUF] = "",	/* for IPSEC */
		ESPstring[MAXBUF] = "";
	int rc;
	char ifstring[1024];

	/* evil Linux PORT kludge to make up for the fact that we are
	not getting the interface */
	ifname = mn_config.ifname;

	/* put the link layer address in OUR way, which might
	 * already have the MAC address in it
	 */
	if (dolink) {
		route_delete_link(agent, NULL);

		/* assume that any crypto transform set for LL is identical
		 * to non-LL route entries
		 */
		route_add_link(agent, mac, ifname
	#ifdef IPSEC
			, link_trans ? rosa : NULL
	#endif
			);
	}

#ifdef IPSEC
	/* build security tuple string.
	 * it may be defined as not necessary.
	 */
	if (rosa && (rosa->transform & (AHSEC|ESPSEC))) {
		/* FIX -- use AH on construc6ted commandilne */
		if (rosa->transform & AHSEC) {
			syslog(LOG_ERR, "mn_route_add_default(): SA tuple transform AH bit not supported yet.");
			return;
		}
		if (rosa->transform & AHSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_AH].src));
			sprintf(AHstring, "-ah -spi 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_AH].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_AH].dst));
		}
		if (rosa->transform & ESPSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_ESP].src));
			sprintf(ESPstring, "-esp -spi 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_ESP].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_ESP].dst));
		}
	}
#endif /* IPSEC */

	sprintf(ifstring, "%s", inet_ntoa(interface));
/*	sprintf(command, "/sbin/route -n add default %s -ifa %s %s> /dev/null", 
		inet_ntoa(agent), ifstring, ESPstring); */
	sprintf(command, "/sbin/route -n add  -host %s dev %s> /dev/null", 
		inet_ntoa(agent), ifname); 
	rc = system(command);
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "mn_route_add_default(): route command (%s) failed, rc %d\n", 
			command, rc);
	}

	sprintf(command, "/sbin/route -n add default gw %s > /dev/null", 
		inet_ntoa(agent)); 

	if (debug) {
		printf("mn_route_add_default: <%s>\n", command);
	}
	rc = system(command);
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "mn_route_add_default(): route command (%s) failed, rc %d\n", 
			command, rc);
	}

#ifdef NOWWRONG
	/* if we add the LLINFO/link route first, this isn't a problem.
	 *
	 * delete clone route if it should show up
	 */
	sprintf(command, "/sbin/route -n delete %s > /dev/null", 
		inet_ntoa(agent));
	rc = system(command);
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "mn_route_add_default(): route command (%s) failed, rc %d\n", 
			command, rc);
	}
#endif
}

/*
 * note:
 *	two route deletes must be done.  one to get rid of
 *	the default route itself and the other to delete
 *	the arp stub
 */
void
mn_route_delete_default(struct in_addr agent, int dolink)
{
	char command[1024];
	int rc;

	sprintf(command, "/sbin/route -n delete default > /dev/null 2> /dev/null");
	if (debug) {
		printf("mn_route_delete_default: <%s>\n", command);
	}
	rc = system(command);

	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "mn_route_delete_default(): route command (%s) failed, rc %d\n", 
			command, rc);
	}

	if (!dolink)
		return;

	sprintf(command, "/sbin/route -n delete %s > /dev/null", inet_ntoa(agent));
	if (debug) {
		printf("mn_route_delete_default: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "mn_route_delete_default(): route command (%s) failed, rc %d\n", 
			command, rc);
	}
}

/*
 * put the clone route back.  in order to do this at boot
 * we need to take the specified ifname and determine
 *	the ip address of the interface
 *	the subnet mask
 *	AND them together to get the clone route and store it
 *		in the per interface structure
 *
 * A route entry of the form:
 *	route add -cloning CLONEIP -netmask NETMASKIP -interface IFAIP
 *
 * will put it back.
 *
 * If the final argument is non-NULL, it is used to form a tuple for
 * associating a security association with the link by appending the
 * following arguments to the above route commandline statement:
 *
 *   (-ah|-esp) -spi <hex|dec> -itsrc <IPaddr> -itdst <IPaddr>
 */
void 
clone_route_add( struct in_addr cloneroute,
	struct in_addr interface, 
	struct in_addr netmask
#ifdef IPSEC
	, struct routeSA *rosa
#endif
	)
{
	char command[1024];
	char	netaddr[MAXBUF],
		AHstring[MAXBUF] = "",	/* for IPSEC */
		ESPstring[MAXBUF] = "";
	int rc;
	char clonestring[1024];
	char netmaskstring[1024];

#ifdef IPSEC
	/* build security tuple string.
	 * it may be defined as not necessary.
	 */
	if (rosa && (rosa->transform & (AHSEC|ESPSEC))) {
		/* FIX -- use AH on constructed commandline */
		if (rosa->transform & AHSEC) {
			syslog(LOG_ERR, "clone_route_add(): SA tuple transform AH bit not supported yet.");
			return;
		}
		if (rosa->transform & AHSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_AH].src));
			sprintf(AHstring, "-ah -spi 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_AH].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_AH].dst));
		}
		if (rosa->transform & ESPSEC) {
			strcpy(netaddr, inet_ntoa(rosa->tup[ROSA_ESP].src));
			sprintf(ESPstring, "-esp -spi 0x%x -itsrc %s -itdst %s",
				rosa->tup[ROSA_ESP].spi,
				netaddr, inet_ntoa(rosa->tup[ROSA_ESP].dst));
		}
	}
#endif /* IPSEC */

	sprintf(clonestring, "%s", inet_ntoa(cloneroute));
	sprintf(netmaskstring, "%s", inet_ntoa(netmask));
	sprintf(command, "/sbin/route -n add -cloning %s -netmask %s -interface %s %s> /dev/null",
		clonestring, netmaskstring, inet_ntoa(interface), ESPstring);

	if (debug) {
		printf("clone_route_add: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "clone_route_add(): route command (%s) failed, rc %d\n", 
			command, rc);
	}
}

/*
 * install a clone route with the -reject flag
 */
void 
reject_route_add( struct in_addr cloneroute,
	struct in_addr interface, 
	struct in_addr netmask
	)
{
	char command[1024];
	int rc;
	char clonestring[1024];
	char netmaskstring[1024];


	sprintf(clonestring, "%s", inet_ntoa(cloneroute));
	sprintf(netmaskstring, "%s", inet_ntoa(netmask));
	sprintf(command, "/sbin/route -n add -reject %s -netmask %s -interface %s > /dev/null",
		clonestring, netmaskstring, inet_ntoa(interface));

	if (debug) {
		printf("reject_route_add: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "reject_route_add(): route command (%s) failed, rc %d\n", 
			command, rc);
	}
}

/* 
 * need to extract subnet mask from interface to deduce 
 *	clone addr.
 */
void 
clone_route_delete(struct in_addr cloneaddr, struct in_addr netmask)
{
	char command[1024];
	char buf[1024];
	int rc;

	sprintf(buf, "%s", inet_ntoa(netmask));
	sprintf(command, "/sbin/route -n delete -net %s netmask %s > /dev/null", inet_ntoa(cloneaddr), buf);
	if (debug) {
		printf("clone_route_delete: <%s>\n", command);
	}
	rc = system(command);
	/* TBD. do something clever to catch route error which
	 * will be written to stderr.
	 */
	if (rc == 127 || rc < 0) {
		syslog( LOG_ERR, "clone_route_delete(): route command (%s) failed, rc %d\n", 
			command, rc);
	}
}


#ifdef TESTMAIN

int debug = 1;

main()
{
	struct in_addr mn;
	struct in_addr fa;
	struct in_addr ha;
	struct in_addr interface;
	struct in_addr clone;
	struct in_addr netmask;
	char *ifname;
	char mac[6];

	mac[0] = 1;
	mac[1] = 1;
	mac[2] = 1;
	mac[3] = 2;
	mac[4] = 2;
	mac[5] = 2;

	mn.s_addr = inet_addr("204.203.71.203");
	fa.s_addr = inet_addr("10.0.0.2");
	ha.s_addr = inet_addr("204.203.71.254");
	clone.s_addr = inet_addr("204.203.71.192");
	netmask.s_addr = inet_addr("255.255.255.192");
	interface.s_addr = mn.s_addr;
	ifname = "wlp0";

#ifdef TESTLINK
	/* 1st as if we were using arp */
	route_add_link(mn, NULL,
		ifname
#ifdef IPSEC
		, NULL
#endif
		);
	route_delete_link(mn, NULL);
	/* 2nd as if we know mac address already */
	route_add_link(mn, mac,
		ifname
#ifdef IPSEC
		, NULL
#endif
		);
	route_delete_link(mn, NULL);
#endif

#ifdef TESTDEFAULT
	mn_route_add_default(fa,interface, 1, mac,
		ifname
#ifdef IPSEC
		, NULL, 0
#endif
		);
	mn_route_delete_default(fa, 1);
	mn_route_add_default(ha, interface, 1, NULL,
		ifname
#ifdef IPSEC
		, NULL, 0
#endif
		);
	mn_route_delete_default(ha, 1);
#endif

#ifdef TESTCLONE
	clone_route_delete(clone, netmask);
	clone_route_add( clone, interface,
		netmask
#ifdef IPSEC
		, NULL
#endif
		);
#endif
}

#endif


