/*
 * Copyright (c) 1996,1997 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * routesock.c
 *
 * handle route(4) messages needed by mobile routing daemons.
 *
 * TBD:
 *	LLINFO part
 *	. test mn default route with link layer routes too	DONE.
 *	IPSEC - add rosa					DONE.
 *	LOGGING - put logging in				DONE.
 *	fix up with mip.src/route.h
 *	plug in and test
 *		mapIfname() goes in *config.c
 *	UPCALL -
 *		read RESOLVE and MISS messages sent to us via select
 *			should drive ad hoc routing protocol
 *		other messages to log
 */
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/mbuf.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/route.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netns/ns.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <paths.h>

#include "mip_log.h"

#ifdef IPSEC
#include "ipsec.h"
#include "keylist.h"
#include "route.h"
#endif /* IPSEC */

#include "mnconfig.h"
#include "mnstate.h"

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))

#define NEXTADDR(w, u) \
        if (rtm_addrs & (w)) {\
            l = ROUNDUP(u.sa.sa_len); bcopy((char *)&(u), cp, l); cp += l;\
        }

static
union	sockunion {
	struct	sockaddr sa;
	struct	sockaddr_in sin;
	struct	sockaddr_dl sdl;
	struct  sockaddr_inarp arp;
#ifdef IPSEC
	struct  sockaddr_ipsec sipsec;
#endif
}
#ifdef IPSEC
	so_ah, so_esp, so_ahtunnel, so_esptunnel,
#endif
	so_dst, so_gate, so_mask, so_genmask, so_ifa, so_ifp,
	arp_sin_m, arp_sdl_m;

typedef union sockunion *sup;

/* structure of a routing message
*/
static struct {
	struct	rt_msghdr m_rtm;
	char	m_space[512];
} m_rtmsg;

static struct rt_metrics rt_metrics;
static unsigned long rtm_inits;

/* route socket descriptor
*/
static int routesocket = -1;		/* make sure it gets opened */
static int nofailflag = 0;   /* don't exit on failure */
extern int debug;

/* prototypes */

#ifdef TESTMAIN
void mapIfname(char *ifname, int *index, unsigned long *ifaddr);
#endif

/* internal functions */
static int rtmsg(int cmd, int flags, int rtm_addrs, int arpflag);
static void mask_addr(int rtm_addrs);
static int getaddr(int which, struct in_addr addr);

#ifdef IPSEC
/* used to walk through list of route SAs to add during route creation */
static int cur_sa;
static struct routeSA_tuple *salist[RTSA_MAX];

static void init_rosa_list(struct routeSA *rosa, struct routeSA *mvifrosa);
static struct routeSA_tuple *walk_rosa_list(void);
static void getaddr_ipsec(u_long ttype, struct routeSA_tuple *tinfo,
	u_long *flags, u_long *rtaddr);
#endif /* IPSEC */

#ifdef MADRP
#include "sr_mad.h"
struct SR_hlist * srhlist_search( struct in_addr mn);
#endif
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif


/* boolean functions
 * nofail means that route command CANNOT fail.
 */   
int     
route_nofail(void) 
{
	nofailflag = 1;
} 

route_fail(void)
{
	nofailflag = 0;
}


/*
 * enable/disable loopback
 * must enable for GET message
 */
static void
routeLoopback(int on)
{
	/* suppress readback of our own messages */
        if (setsockopt(routesocket, SOL_SOCKET, SO_USELOOPBACK,
			&on, sizeof(on))) {
 		syslog(LOG_ERR, "routeInit(): SO_USELOOPBACK failed (%m).");
		exit(1);
	}
}

/*
 * init routing socket
 */
int
routeInit()
{

	routesocket = socket(PF_ROUTE, SOCK_RAW, 0);
	if (routesocket < 0) {
		syslog(LOG_ERR, "routeInit(): socket failed (%m).");
		exit(1);
	}

	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(routesocket, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "routeInit(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	/* turn off unless needed
	*/
	routeLoopback(0);
	return(routesocket);
}

/*
 * zero out the socket variables and set various
 * initial values.  every routing function that actually
 * sets a route should call this before it starts plugging
 * values into sockets.
 */
static
void
initaddrs()
{
	rtm_inits = 0;

#define z(so)	(bzero((char *) &(so), sizeof(so)))

	z(so_dst);	z(so_gate);	z(so_mask);	z(so_genmask);
	z(so_ifa);	z(so_ifp);	z(arp_sin_m);	z(arp_sdl_m);
	z(rt_metrics);
#ifdef IPSEC
	z(so_ah);	z(so_esp);	z(so_ahtunnel);	z(so_esptunnel);
#endif

#undef z

	so_dst.sa.sa_family = AF_INET;
	so_gate.sa.sa_family = AF_INET;
}

void route_delete_host(struct in_addr mn, struct in_addr router) ;
void route_delete_link(struct in_addr host, char *mac);
#ifdef MIPD
struct vlist * vlist_getfahead();
struct vlist * vlist_gethahead();
struct vlist * vlist_search(struct vlist *, struct in_addr);
#endif

/*
 * read from the routing socket (after select)
 */

void
routeRead()
{
	struct in_addr dst;
	u_char packetbuf[2048];
	struct rt_msghdr *packet = (struct rt_msghdr *) packetbuf;
	struct in_addr nexthop;
	u_int len ;
	int type;
#ifdef MIPD
	struct vlist * vlisthead=NULL, *thisvlist=NULL;
#else
#ifdef MADRP
	struct vlist * vlisthead=NULL, *thisvlist=NULL;
	struct SR_hlist * ss;
	struct in_addr myip;
	myip.s_addr=getmyaddr();
#endif
#endif

	len = read(routesocket, packetbuf, sizeof packetbuf);
	if (len < 0) {
 		syslog(LOG_ERR, "routeRead(): route socket read failed (%m).");
		mip_log(MLOG_ERR, "routeRead(): read failed, errno %d \n", 
			strerror(errno));
		return;
	}

	dst.s_addr= ((struct sockaddr_in *) (&packet[1]))->sin_addr.s_addr;
	
	switch (packet->rtm_type){

#ifdef MND

	case RTM_RESOLVE :
		/* message was due to cloning with route that has xresolve (X)
	 	 * flag
		 * do a route get on it to get flags
		 */
		initaddrs();
		getaddr(RTA_DST, dst);
		if ( rtmsg('g', 0, RTA_DST,0) < 0) {
#ifdef DEBUG
			if (debug) { 
				printf("routegetinfo failed: dst:%s\terrno:\n", 
					inet_ntoa(dst), errno);
			}
#endif
 		        syslog(LOG_ERR, "routeRead(): get failed on dst %s,"
					" errno (%m).", inet_ntoa(dst));
			return;
		}
		/* if LLINFO flag, then this would be due to
		 * ARP/clone
		 */
		if ( m_rtmsg.m_rtm.rtm_flags & RTF_LLINFO) {
			/* it goes to link layer clone route */
#ifdef DEBUG
			if (debug)
				printf("rtm_resolve message recved, "
				"%s goes to link layer clone route! ignored\n",
				inet_ntoa(dst));
#endif
			return;
		/* else assume default route clone, and drive
		 * multicast ad hoc routing algorithm
		 */
		} else {
			/* TBD. only call if config option indicates
			 * madrp in use
			 */
#ifdef MADRP
			Send_SR_request(dst, TRUE,myip); 
#endif
#ifdef DEBUG
			if (debug)
				printf("rtm_resolve message recved,"
				" %s goes to default route!"
				" send_sr_request\n", inet_ntoa(dst));
#endif
			return;
		}
		break;

	case RTM_MISS : 
		/* route miss.  */
		initaddrs();
		getaddr(RTA_DST, dst);
		if( rtmsg('g', 0, RTA_DST,0)<0) {
			/* failed in looking up the routing table 
			 */
			if (noDefaultroute()) {
				dst.s_addr=((struct sockaddr_in *) (packet
				  +sizeof(struct rt_msghdr)))->sin_addr.s_addr;	
#ifdef DEBUG
				if (debug) {
					printf("rtm_miss recved,no defaultroute"
					"failed in looking up the routing"
					" table of %s ! send_sr_request\n",
					inet_ntoa(dst));
				}
#endif
#ifdef MADRP
				ss=srhlist_search(dst);
				if (ss) {
#ifdef DEBUG
					if(debug)
						printf("strange, who deleted"
					"the route entry, hlist still exists");
#endif
					srhlist_delete(dst);
				}
				Send_SR_request(dst, TRUE,myip);
#endif
				return;
			} else {
				/* ignore this case */
#ifdef DEBUG
				if (debug)
					printf("rtm_miss recved."
					" should have default."
					" failed in looking up the routing"
					" table of %s ! ignore\n",
					inet_ntoa(dst));
#endif
				return;
			}
		}
		/* generated by arp , strange */
		if ( m_rtmsg.m_rtm.rtm_flags & RTF_LLINFO) {
#ifdef DEBUG
			if (debug)
				printf("rtm_miss recved. it is generated by arp"
				" dst: %s ! ignore\n", inet_ntoa(dst));
#endif
			return;
		} else {
			/* not generated by arp 
			*/
			if (noDefaultroute()) {
				/*
				Send_SR_request(dst, TRUE,myip);
				*/
#ifdef DEBUG
				if (debug)
					printf("rtm_miss. no defaultroute."
					"send sr request of %s\n",
					inet_ntoa(dst));
#endif
				return;
			} else {
				/* ignore this case */
#ifdef DEBUG
				if (debug)
					printf("rtm_miss. should have default."
					" dst: %s ! ignore\n",
					inet_ntoa(dst));
#endif
				return;
			}
		}	
		break;

	case RTM_LOSING :  
		/* rtm_losing, it happens when an entry in routing table is not
		 * reac	hable by TCP.
		 * TBD. see netinet/in_pcb.c
		 */
#ifdef DEBUG
		if(debug){
			printf("rtm_losing recved, will send out SR request"
			       "dst is %s.\n", inet_ntoa(dst));
		}
#endif
#ifdef MADRP
		ss=NULL;
		ss=srhlist_search(dst);
		if (ss) {
			route_delete_host(ss->destHost,ss->nextHop);
			srhlist_delete(ss->destHost);
			/* only send request when it was a host route entry ,
			 * ignor the link layer RTM_LOSING msg, wait time out
			 */
			Send_SR_request(dst, TRUE,myip);
		} else {
			if ( (m_rtmsg.m_rtm.rtm_flags & RTF_LLINFO ) &&
			     (! (m_rtmsg.m_rtm.rtm_flags & RTF_WASCLONED))) {
				/* entry of <host, mac>	
				 * TBD , may be we should wait till link entry 
		   		 * to time out
		   		 */
			} else {
				/* include cases of <host,gateway>(not in
				 * hlist),  <cloned link entry> ...
				 */
				route_delete_host(dst,dst); /* TBD */
				Send_SR_request(dst, TRUE,myip);
			}
		}
#endif
		break;

#else /* ifdef MIPD */
	
	case RTM_RESOLVE :
		/* message was due to cloning with route that has xresolve (X)
	 	 * flag
		 */
		printf("strange!, recv RTM_RESOLVE\n");
		mip_log(MLOG_INFO, "RTM_RESOLVE recv,dst %s\n", inet_ntoa(dst));
		break;

	case RTM_MISS : 
		/* route miss.  */
		printf("strange!, agent don't have default route?\n");
		mip_log(MLOG_INFO, "RTM_MISS recv, dst %s\n", inet_ntoa(dst));
		break;

	case RTM_LOSING :  
		/* rtm_losing, it happens when an entry in routing table is not
		 * reac	hable by TCP.
		 * TBD. see netinet/in_pcb.c
		 */
        	if(is_host_route(dst, &nexthop)) {
#ifdef MADRP
#ifdef DEBUG
			if(debug)
		       printf("delete host %s in routing table\n",
			       inet_ntoa(dst));
#endif
		       route_delete_host(dst, nexthop);
	       } else if(is_link_route(dst)) { 
#ifdef DEBUG
			if(debug)
		       printf("delete link %s in routing table\n", 
				inet_ntoa(dst));
#endif
		       route_delete_link(dst,(char *)&dst); /* TBD */
	     	} 
		vlisthead=vlist_getfahead();
                thisvlist=vlist_search(vlisthead,dst);
		if(! thisvlist) {
			vlisthead=vlist_gethahead();
			thisvlist=vlist_search(vlisthead,dst);
		}
		if( thisvlist){ 
#ifdef DEBUG
			if(debug)
			printf("delete %s in vlist/hlist\n",inet_ntoa(dst)); 
#endif
			vlist_delete(&vlisthead, dst);
			srhlist_delete( dst);
#endif /* MADRP */
		}
		break;

#endif /* DEFINE OF MND */

	default:
		/* TBD. should log these for awhile and see what
		 * shows up
		 */
		type = ((struct rt_msghdr *) packet)->rtm_type;
		mip_log(MLOG_INFO, "ROUTEMSG: type %d, dst %s\n", type,
			inet_ntoa(dst));
	}
	return;
}

/*
 * ha_route_add_tunnel
 */

void
ha_route_add_tunnel(struct in_addr mn, struct in_addr coa
#ifdef IPSEC
		, struct routeSA *rosa, struct routeSA *mvifrosa
#endif
		)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_GATEWAY | RTF_TUNNEL | RTF_UP;
	int rtm_addrs = RTA_DST | RTA_GATEWAY;
	int cmd = 'a';
	struct routeSA_tuple *tuple;
	u_long newflag, newaddr;
	int rc;
	int add_attempts = 0;

tryagain_tunnel:
	init_rosa_list(rosa, mvifrosa);
	walk_rosa_list();

	do {
		initaddrs();
		newflag = newaddr = 0;

		getaddr(RTA_DST, mn);
		getaddr(RTA_GATEWAY, coa);
		getaddr_ipsec((u_long) cur_sa, salist[cur_sa],
			&newflag, &newaddr);

		rc = rtmsg(cmd, flags | newflag, rtm_addrs | newaddr, 0);
		if (rc < 0) {
			char coabuf[1024];
			char *attempt_message = "";

			sprintf(coabuf, "%s", inet_ntoa(coa));

			/* distinguish the first failed add attempt
			 * from the second in the log file
			 */
			if (cmd == 'a') {
				if (!add_attempts)
					attempt_message = " (first attempt)";
				else
					attempt_message = " (fatal)";
			}

#ifdef TESTMAIN
			printf("ha_route_add_tunnel(): %s FAILED%s: mn=%s coa=%s, %s\n",
				cmd == 'a' ? "add" : "change", attempt_message,
				inet_ntoa(mn), coabuf, sys_errlist[errno]);
#else
			syslog(LOG_ERR, "ha_route_add_tunnel(): %s FAILED%s: mn=%s coa=%s, %s\n",
				cmd == 'a' ? "add" : "change", attempt_message,
				inet_ntoa(mn), coabuf, sys_errlist[errno]);
#endif
	
			if (cmd == 'a' && !add_attempts) {
				add_attempts += 1;
				ha_route_delete_tunnel(mn, coa);
				goto tryagain_tunnel;
			}

			exit(1);
		}

		cmd = 'c';
	} while (walk_rosa_list());

}


/*
 * zap the tunnel route
 */
void
ha_route_delete_tunnel(struct in_addr mn, struct in_addr coa)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_GATEWAY | RTF_TUNNEL | RTF_UP;
	int rtm_addrs = RTA_DST | RTA_GATEWAY;
	char coabuf[1024];

	initaddrs();

	getaddr(RTA_DST, mn);
	getaddr(RTA_GATEWAY, coa);
#ifdef DEBUG
	if (debug) {
		char rbuf[1024];
		sprintf(rbuf, "%s", inet_ntoa(coa));
		printf("route delete tunnel:mn %s coa %s\n",
			inet_ntoa(mn), rbuf);
	}
#endif
	if (rtmsg('d', flags, rtm_addrs, 0) >= 0) {
		return;
	}

	sprintf(coabuf, "%s", inet_ntoa(coa));
#ifdef TESTMAIN
	printf("ha_route_delete_tunnel(): FAILED: mn %s coa %s, errno %s\n",
		inet_ntoa(mn), coabuf, strerror(errno));
#else
#ifdef DEBUG
		if (debug)
			printf("ha_route_delete_tunnel: FAILED: mn %s coa %s, errno %s\n",
				inet_ntoa(mn), coabuf, strerror(errno));
	syslog(LOG_ERR, "ha_route_delete_tunnel(): FAILED: mn %s coa %s, errno %s\n",
		inet_ntoa(mn), coabuf, strerror(errno));
#endif
#endif
}

/*
 * add host route
 *
 * interface is specified so that it will work with/without default route
 */

void
route_add_host(struct in_addr mn, struct in_addr router,
		struct in_addr interface
#ifdef IPSEC
		, struct routeSA *rosa
#endif
		)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_GATEWAY | RTF_UP;
	int rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFA;
	u_long newflag, newaddr;
	int cmd = 'a';
	int rc;
	
	init_rosa_list(rosa, NULL);
	walk_rosa_list();

	do {
		initaddrs();
		newflag = newaddr = 0;

		getaddr(RTA_DST, mn);
		getaddr(RTA_GATEWAY, router);
		getaddr(RTA_IFA, interface);
		getaddr_ipsec((u_long) cur_sa, salist[cur_sa],
			&newflag, &newaddr);

		rc = rtmsg(cmd, flags | newflag, rtm_addrs | newaddr, 0);
		if (rc < 0) {
			char rbuf[1024];
			sprintf(rbuf, "%s", inet_ntoa(router));
#ifdef TESTMAIN
			printf("route_add_host(): %s FAILED: mn %s router %s, errno %s\n",
				(cmd == 'a') ? "add" : "change",
				inet_ntoa(mn), rbuf, sys_errlist[errno]);
#else
			syslog(LOG_ERR, "route_add_host(): %s FAILED: mn %s router %s, errno %s\n",
				(cmd == 'a') ? "add" : "change",
				inet_ntoa(mn), rbuf, sys_errlist[errno]);
#endif
			exit(1);
		}

		cmd = 'c';
	} while (walk_rosa_list());

}  /* end route_add_host() */

/*
 * delete host route
 */
void
route_delete_host(struct in_addr mn, struct in_addr router)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_GATEWAY | RTF_UP;
	int rtm_addrs = RTA_DST | RTA_GATEWAY;

	initaddrs();
	getaddr(RTA_DST, mn);
	getaddr(RTA_GATEWAY, router);

#ifdef DEBUG
	if(debug) {
		char rbuf[1024];
		sprintf(rbuf, "%s", inet_ntoa(router));
		printf("route delete host:mn %s router %s\n",
			inet_ntoa(mn), rbuf);
	}
#endif

	if (rtmsg('d', flags, rtm_addrs, 0) < 0) {
		char rbuf[1024];

		sprintf(rbuf, "%s", inet_ntoa(router));
#ifdef TESTMAIN
		printf("route_delete_host(): FAILED: mn %s router %s, errno %s\n",
			inet_ntoa(mn), rbuf, strerror(errno));
#else
		syslog(LOG_ERR, "route_delete_host(): FAILED: mn %s router %s, errno %s\n",
			inet_ntoa(mn), rbuf, strerror(errno));
#endif
	}
}

/*
 * add link layer route
 *
 * agent - ip address (could be agent or mn)
 * mac - if NULL, arp is used, else mac address filled in
 * ifname - char * name of i/f (not callback to config code
 *	to get link index, ip addr)
 */
void
route_add_link(struct in_addr agent, char *mac, char *ifname
#ifdef IPSEC
		, struct routeSA *rosa
#endif
		)
{
	struct sockaddr_inarp sin;
	struct sockaddr_dl sdl;
	u_char *ea;
	int expire_time = 0;
	int flags = RTF_STATIC | RTF_HOST ;
	int rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFA;
	u_long newflag, newaddr;
	int cmd = 'a';
	int rc;
	int if_index;
	struct in_addr if_inaddr;
	int add_attempts = 0;
 
tryagain_link:
	bzero(&sin, sizeof(sin));
	bzero(&sdl, sizeof(sdl));


	/* map interface name to index/ip
	*/
	mapIfname(ifname, &if_index, (unsigned long *)&if_inaddr);

	/* initialize the destination
	*/
	sin.sin_len = sizeof(struct sockaddr_inarp);
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = agent.s_addr;  /* set ip */	

	/* initialize the gateway
	*/
	sdl.sdl_len = sizeof(struct sockaddr_dl);
	sdl.sdl_family = AF_LINK;	
	sdl.sdl_index = if_index;	
	sdl.sdl_alen = 0;
	if (mac != NULL) {
		bcopy(mac, LLADDR(&sdl), 6);
		sdl.sdl_alen = 6;
	}


	init_rosa_list(rosa, NULL);
	walk_rosa_list();

	do {
		errno = 0;
		initaddrs();
		newflag = newaddr = 0;
		bcopy(&sin, &arp_sin_m, sizeof(sin));
		bcopy(&sdl, &arp_sdl_m, sizeof(sdl));

		/* set the interface value
		*/
		getaddr(RTA_IFA, if_inaddr);
		getaddr_ipsec((u_long) cur_sa, salist[cur_sa],
			&newflag, &newaddr);

		/* make sure expire is set as a flag and value is 0
		 */
		rt_metrics.rmx_expire = 0;
		rtm_inits = RTV_EXPIRE; 

		rc = rtmsg(cmd, flags | newflag, rtm_addrs | newaddr, 1);
		if (rc < 0) {
			char macaddr[32];
			char *attempt_message = "";

			/* distinguish the first failed add attempt
			 * from the second in the log file
			 */
			if (cmd == 'a') {
				if (!add_attempts)
					attempt_message = " (first attempt)";
				else
					attempt_message = " (fatal)";
			}

			sprintf(macaddr, "%x:%x:%x:%x:%x:%x",
				mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
#ifdef TESTMAIN
			printf("route_add_link(): %s FAILED%s: agent %s MAC %s, errno %s\n",
				(cmd == 'a') ? "add" : "change",
				attempt_message,
				inet_ntoa(agent), macaddr, sys_errlist[errno]);
#else
			syslog(LOG_ERR, "route_add_link(): %s FAILED%s: agent %s MAC %s, errno %s\n",
				(cmd == 'a') ? "add" : "change",
				attempt_message,
				inet_ntoa(agent), macaddr, sys_errlist[errno]);
#endif

                        if (cmd == 'a' && !add_attempts) {
                                add_attempts += 1;
                                route_delete_link(agent, mac);
                                goto tryagain_link;
                        }

			exit(1);
		}

		cmd = 'c';
	} while(walk_rosa_list());

}  /* end route_add_link() */


/*
 * zap the link layer route
 */
void
route_delete_link(struct in_addr host, char *mac)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_UP;
	int rtm_addrs = RTA_DST;

	initaddrs();

	getaddr(RTA_DST, host);

#ifdef DEBUG
	if (debug) {
		printf("route delete link:mn %s \n", inet_ntoa(host));
	}
#endif

	if (rtmsg('d', flags, rtm_addrs, 0) < 0) {
#ifdef TESTMAIN
		printf("route_delete_link(): FAILED: host %s errno %s\n",
			inet_ntoa(host), strerror(errno));
#else
		syslog(LOG_ERR, "route_delete_link(): FAILED: host %s errno %s\n",
			inet_ntoa(host), strerror(errno));
#endif
	}
}

/*
 * add proxy arp entry
 *
 * agent - ip address (could be agent or mn)
 * mac - if NULL, arp is used, else mac address filled in
 * ifname - char * name of i/f (not callback to config code
 *	to get link index, ip addr)
 */
void
route_add_proxy(struct in_addr agent, char *mac, char *ifname)
{
	struct sockaddr_inarp *sin;
	struct sockaddr_dl *sdl;
	u_char *ea;
	int flags = RTF_STATIC | RTF_ANNOUNCE;
	int rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK | RTA_IFA ;
	int if_index;
	struct in_addr if_inaddr;
	struct in_addr allones;

        if (mac == NULL) {
                syslog(LOG_ERR, "route_add_proxy(): (internal) MAC address is NULL.  Exiting...");
                exit(1);
        }

	errno = 0;
	initaddrs();

	/* map interface name to index/ip
	*/
	mapIfname(ifname, &if_index, (unsigned long *)&if_inaddr);

	sin = (struct sockaddr_inarp *) &arp_sin_m;
	sdl = (struct sockaddr_dl *) &arp_sdl_m;

	/* initialize the destination
	*/
	sin->sin_len = sizeof(struct sockaddr_inarp);
	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = agent.s_addr;  /* set ip */

	/* initialize the gateway
	*/
	sdl->sdl_len = sizeof(struct sockaddr_dl);
	sdl->sdl_family = AF_LINK;
	sdl->sdl_index = if_index;
	ea = (u_char *) LLADDR((struct sockaddr_dl *)&arp_sdl_m);
	sdl->sdl_alen = 0;
	if (mac != NULL) {
		bcopy(mac, ea, 6);
		sdl->sdl_alen = 6;
	}

	/* set the interface value
	*/
	allones.s_addr = inet_addr("255.255.255.255");
	getaddr(RTA_NETMASK, allones);
	getaddr(RTA_IFA, if_inaddr);

	/* make sure expire is set as a flag and
	 * value is 0
	 */
	rt_metrics.rmx_expire = 0;
	rtm_inits = RTV_EXPIRE;

	if (rtmsg('a', flags, rtm_addrs, 1) < 0) {
#ifdef TESTMAIN
		printf("route_add_proxy(): FAILED: agent %s MAC %s, errno %s\n",
			inet_ntoa(agent), mac, strerror(errno));
#else
		syslog(LOG_ERR, "route_add_proxy(): FAILED: agent %s MAC %s, errno %s\n",
			inet_ntoa(agent), mac, strerror(errno));
#endif
		exit(1);
	}

}  /* end route_add_proxy() */

/*
 * zap the proxy arp entry
 */
void
route_delete_proxy(struct in_addr host)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_UP;
	int rtm_addrs = RTA_DST;

	initaddrs();
	getaddr(RTA_DST, host);

	if (rtmsg('d', flags, rtm_addrs, 0) < 0) {
#ifdef TESTMAIN
		printf("route_delete_proxy(): FAILED: host %s errno %s\n",
			inet_ntoa(host), strerror(errno));
#else
		syslog(LOG_ERR, "route_delete_proxy(): FAILED: host %s errno %s\n",
			inet_ntoa(host), strerror(errno));
#endif
	}
}  /* end route_delete_proxy() */


/*
 * add default route including link at MN
 */
void 
mn_route_add_default(struct in_addr agent, struct in_addr interface,
	int dolink, char *mac, char *ifname
#ifdef IPSEC
	, struct routeSA *rosa
	, u_long link_trans	/* TRUE: transform also applies to LL route */
#endif
	)
{
	/* XRESOLVE will cause an upcall when cloning occurs
	*/
	int flags = RTF_STATIC | RTF_GATEWAY | RTF_UP | RTF_XRESOLVE;
	int rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_IFA | RTA_NETMASK;
	u_long newflag, newaddr;
	int cmd = 'a';
	int rc;
	struct in_addr mn;

	/* put the link address in BEFORE the default route
	 * if it is there already, log that fact and remove it
	 */
	if (dolink) {
		nofailflag = 1;
		route_add_link(agent, mac, ifname
#ifdef IPSEC
			, link_trans ? rosa : NULL
#endif
			);
		nofailflag = 0;
		/* if the route exists, delete it
		*/
		if (errno == EEXIST) {
			route_delete_link(agent, mac);
			route_add_link(agent, mac, ifname
#ifdef IPSEC
				, link_trans ? rosa : NULL
#endif
			);
			mip_log(MLOG_INFO, "mn_route_add_default(): deleted pre-existing link-layer route (%s) for default.\n",
				agent.s_addr);

		}
	}
	
	mn.s_addr = 0;

	init_rosa_list(rosa, NULL);
	walk_rosa_list();

	do {
		initaddrs();
		newaddr = newflag = 0;
		errno = 0;

		getaddr(RTA_DST, mn);
		getaddr(RTA_GATEWAY, agent);
		getaddr(RTA_IFA, interface);
		getaddr_ipsec((u_long) cur_sa, salist[cur_sa],
			&newflag, &newaddr);

		rc = rtmsg(cmd, flags | newflag, rtm_addrs | newaddr, 0);

		if (rc < 0) {
			char rbuf[1024];
			char ifbuf[1024];
			sprintf(rbuf, "%s", inet_ntoa(agent));
			sprintf(ifbuf, "%s", inet_ntoa(interface));
#ifdef TESTMAIN
			printf("mn_route_add_default(): %s FAILED: mn %s router %s, ifname %s, errno %s\n",
				(cmd == 'a') ? "add" : "change",
				inet_ntoa(mn), rbuf, ifbuf, sys_errlist[errno]);
#else
			syslog(LOG_ERR, "mn_route_add_default(): %s FAILED: mn %s router %s, ifname %s, errno %s\n",
				(cmd == 'a') ? "add" : "change",
				inet_ntoa(mn), rbuf, ifbuf, sys_errlist[errno]);
#endif
			exit(1);
		}

		cmd = 'c';
	} while (walk_rosa_list());

}  /* end mn_route_add_default() */


/*
 * delete the default route
 */
void
mn_route_delete_default(struct in_addr router, int dolink)
{
	int flags = RTF_STATIC | RTF_GATEWAY | RTF_UP;
	int rtm_addrs = RTA_DST | RTA_NETMASK;
	struct in_addr mn;

	initaddrs();
	errno = 0;
	mn.s_addr = 0;

	getaddr(RTA_DST, mn);
	/*
	getaddr(RTA_GATEWAY, router);
	*/

	if (rtmsg('d', flags, rtm_addrs, 0) < 0) {
		char rbuf[1024];

		sprintf(rbuf, "%s", inet_ntoa(router));
#ifdef TESTMAIN
		printf("mn_route_delete_default(): FAILED: mn %s router %s, errno %s\n",
			inet_ntoa(mn), rbuf, strerror(errno));
#else
		syslog(LOG_ERR, "mn_route_delete_default(): FAILED: mn %s router %s, errno %s\n",
			inet_ntoa(mn), rbuf, strerror(errno));
#endif
	}

	/* delete the link
	*/
	if (dolink) {
		route_delete_link(router, NULL);
	}
}

#ifdef DHCP
/* 
 * add tunneled default route to home agent, used in DHCP mode
 */     
void
mn_route_add_tunnel(struct in_addr agent
#ifdef IPSEC
        , struct routeSA *rosa
#endif
        )
{
        int flags = RTF_STATIC | RTF_GATEWAY | RTF_TUNNEL | RTF_UP;  
        int rtm_addrs = RTA_DST | RTA_GATEWAY | RTA_NETMASK;
        u_long newflag, newaddr;
        int cmd = 'a';
        int rc;
        struct in_addr mn;
        mn.s_addr = 0;

        init_rosa_list(rosa, NULL);
        walk_rosa_list();

        do {
                initaddrs();
                newaddr = newflag = 0;
                errno = 0;

                getaddr(RTA_DST, mn);
                getaddr(RTA_GATEWAY, agent);
                getaddr_ipsec((u_long) cur_sa, salist[cur_sa],
                        &newflag, &newaddr);

                rc = rtmsg(cmd, flags | newflag, rtm_addrs | newaddr, 0);

                if (rc < 0) {
                        char rbuf[16];
                        sprintf(rbuf, "%s", inet_ntoa(agent));
                        syslog(LOG_ERR, "mn_route_add_tunnel(): %s FAILED: dst %s router %s, errno %s\n",
                                (cmd == 'a') ? "add" : "change",
                                inet_ntoa(mn), rbuf, sys_errlist[errno]);
                        exit(1);
                }

                cmd = 'c';
        } while (walk_rosa_list());

}  /* end mn_route_add_tunnel() */
#endif

static void
clone_or_reject_route_add(int clone_or_reject,
	struct in_addr cloneroute, struct in_addr interface,
	struct in_addr netmask
#ifdef IPSEC
	, struct routeSA *rosa
#endif
	)
{
/* add XRESOLVE (and BLACKHOLE) for arp under ipsec
	int flags = RTF_STATIC | RTF_UP | RTF_XRESOLVE | RTF_BLACKHOLE | RTF_CLONING;
*/
	int flags = RTF_STATIC | RTF_UP | clone_or_reject;
		/*  RTF_REJECT or RTF_CLONING */
	int rtm_addrs = RTA_DST | RTA_NETMASK | RTA_GATEWAY;

	errno = 0;
	initaddrs();

	/* clone route goes into destination
	*/
	getaddr(RTA_DST, cloneroute);
	getaddr(RTA_NETMASK, netmask);
	/* interface goes in gateway
	*/
	getaddr(RTA_GATEWAY, interface);

	if (rtmsg('a', flags, rtm_addrs, 0) < 0
			&& !(clone_or_reject & RTF_REJECT)) {
		char clonestring[1024];
		char netmaskstring[1024];
		char ifastring[1024];

		sprintf(clonestring, "%s", inet_ntoa(cloneroute));
		sprintf(netmaskstring, "%s", inet_ntoa(netmask));
		sprintf(ifastring, "%s", inet_ntoa(interface));
#ifdef TESTMAIN
		printf("clone_or_reject_route_add(): FAILED: clone %s netmask %s, if %s, errno %s\n",
			clonestring, netmaskstring, ifastring,
			strerror(errno));
#else
		syslog(LOG_ERR, "clone_or_reject_route_add(): FAILED: clone %s netmask %s, if %s, errno %s\n",
			clonestring, netmaskstring, ifastring,
			strerror(errno));
#endif
		exit(1);
	}
}

/*
 * install the link layer clone route.
 */
void
clone_route_add(struct in_addr cloneroute, struct in_addr interface,
	struct in_addr netmask
#ifdef IPSEC
	, struct routeSA *rosa
#endif
	)
{
	clone_or_reject_route_add(RTF_CLONING, cloneroute, interface,
				netmask
#ifdef IPSEC
				, rosa
#endif
				);
}

/*
 * add REJECT route
 */
void
reject_route_add(struct in_addr cloneroute, struct in_addr interface,
	struct in_addr netmask)
{
	clone_or_reject_route_add(RTF_REJECT, cloneroute, interface,
				netmask
#ifdef IPSEC
				, 0
#endif
				);
}

/*
 * delete the link layer clone route
 */
void
clone_route_delete(struct in_addr cloneaddr)
{
	int flags = RTF_STATIC | RTF_HOST | RTF_UP;
	int rtm_addrs = RTA_DST;

	initaddrs();
	getaddr(RTA_DST, cloneaddr);

	if (rtmsg('d', flags, rtm_addrs, 0) < 0) {
#ifdef TESTMAIN
		printf("clone_route_delete(): FAILED: clone %s errno %s\n",
			inet_ntoa(cloneaddr), strerror(errno));
#else
		syslog(LOG_ERR, "clone_route_delete(): FAILED: clone %s errno %s\n",
			inet_ntoa(cloneaddr), strerror(errno));
#endif
		/*
		 * Don't worry if it's not there (especially important when
		 * starting up after a crash).
		 */
		if (errno != ESRCH) {
			exit(1);
		}
	}
}


#ifdef IPSEC
/*
 * init_rosa_list
 *
 * salist is initialized with pointers to all the tuple information kept
 * in the pair of routeSA struct's rosa and mvifrosa.
 *
 * cur_sa is in the set RTSA_* .
 */
static void
init_rosa_list(struct routeSA *rosa, struct routeSA *mvifrosa)
{
	int i;

	for (i = 0; i < RTSA_MAX; i++)
		salist[i] = NULL;

	if (rosa) {
		if (rosa->transform & AHSEC) {
			salist[0] = &rosa->tup[ROSA_AH];
		}
		if (rosa->transform & ESPSEC) {
			salist[1] = &rosa->tup[ROSA_ESP];
		}
	} 
	if (mvifrosa) {
		if (mvifrosa->transform & AHSEC) {
			salist[2] = &mvifrosa->tup[ROSA_AH];
		}
		if (mvifrosa->transform & ESPSEC) {
			salist[3] = &mvifrosa->tup[ROSA_ESP];
		}
	}

	cur_sa = -1;

}  /* end init_rosa_list() */

/* 
 * walk_rosa_list
 *
 * return the next valid pointer to tuple data in the pair of routeSA
 * struct's rosa/mvifrosa, or NULL if there are no more.
 */
static struct routeSA_tuple *
walk_rosa_list(void)
{
	while (cur_sa < RTSA_MAX && !salist[++cur_sa])
		/* empty */ ;

	return( (cur_sa == RTSA_MAX) ?
		(struct routeSA_tuple *)NULL : salist[cur_sa] );

}  /* end walk_rosa_list() */

/*
 * getaddr_ipsec
 *
 * add addresses and bits to the route message for the given route SA.
 * ttype is in the set RTSA_* .
 */
static void
getaddr_ipsec(u_long ttype, struct routeSA_tuple *tinfo,
	u_long *flags, u_long *rtaddr)
{
	u_long	rosa_transform_matches;
	sup	so = NULL;

	/* no security association info to manage */
	if (!tinfo || ttype >= RTSA_MAX) {
		return;
	}

	/* collect SA info for route message */
	switch (ttype) {
	case RTSA_AH:
		so = &so_ah;
		*flags |= RTF_AUTH;
		*rtaddr |= RTA_AH;
		break;
	case RTSA_ESP:
		so = &so_esp;
		*flags |= RTF_CRYPT;
		*rtaddr |= RTA_ESP;
		break;
	case RTSA_AHTUNNEL:
		so = &so_ahtunnel;
		*flags |= RTF_AHTUNNEL;
		*rtaddr |= RTA_AHTUNNEL;
		break;
	case RTSA_ESPTUNNEL:
		so = &so_esptunnel;
		*flags |= RTF_ESPTUNNEL;
		*rtaddr |= RTA_ESPTUNNEL;
		break;
	default:
#ifdef DEBUG
		if (debug) {
			printf("getaddr_ipsec(): (internal) request to modify unknown transform type.\n");
		}
#endif
		syslog(LOG_ERR, "getaddr_ipsec(): (internal) request to modify unknown transform type.");
		exit(1);
	}


	/* define the chosen sockaddr_ipsec with the chosen rosa tuple info */
	so->sipsec.sipsec_len = sizeof(struct sockaddr_ipsec);
	so->sipsec.sipsec_type = ttype;
	so->sipsec.sipsec_spi = tinfo->spi;
	so->sipsec.sipsec_src = tinfo->src.s_addr;
	so->sipsec.sipsec_dst = tinfo->dst.s_addr;

}  /* end getaddr_ipsec() */
#endif /* IPSEC */


/*
 *  delete all arp entries EXCEPT for multicast LLINFO entries.
 */
static struct sockaddr_inarp blank_sin = {sizeof (blank_sin), AF_INET };

void
flushArp()
{
	int mib[6];
	size_t needed;
	char *malloc(), *lim, *buf, *next;
	struct rt_msghdr *rtm;
	int rtm_addrs = RTA_DST;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = AF_INET;
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_LLINFO;

	if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0) {
#ifdef TESTMAIN
		printf("flushArp(): sysctl failed on needed.\n");
#else
		syslog(LOG_ERR, "flushArp(): sysctl failed on needed.\n");
#endif
		exit(1);
	}
	if ((buf = malloc(needed)) == NULL) {
#ifdef TESTMAIN
		printf("flushArp(): malloc failed\n");
#else
		syslog(LOG_ERR, "flushArp(): malloc failed\n");
#endif
		exit(1);
	}
	if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0) {
#ifdef TESTMAIN
		printf("flushArp(): sysctl failed on buffer\n");
#else
		syslog(LOG_ERR, "flushArp(): sysctl failed on buffer\n");
#endif
		exit(1);
	}

	/* loop through the routing entries in the buf
	 * loop until lim is hit.
	 */
	initaddrs();
	lim = buf + needed;
	for (next = buf; next < lim; next += rtm->rtm_msglen) {
		rtm = (struct rt_msghdr *) next;
		/* zero out arp_sin_m and put dest ip in it
		*/
		bcopy(&blank_sin, &arp_sin_m, sizeof(struct sockaddr_inarp));
		bcopy( &((struct sockaddr_inarp *)(rtm + 1))->sin_addr.s_addr,
		       &(arp_sin_m.arp.sin_addr.s_addr),
		       sizeof(long));

		/* fill in that routing table entry by doing a GET
		*/
		if (rtmsg('g', 0, rtm_addrs, 1)) {
#ifdef TESTMAIN
			fprintf(stderr, "arp: RTM_GET failed in flush()\n");
#endif
			continue;
		}
		/* if address is multicast, skip it
		*/
		if (IN_MULTICAST(ntohl(arp_sin_m.arp.sin_addr.s_addr))) {
			continue;
		}
		/* else delete it
		*/
		else if (rtmsg('d', 0, rtm_addrs, 1)) {
#ifdef TESTMAIN
			fprintf(stderr, "arp: RTM_DELETE failed in flush()\n");
#endif
		}
	}
	/* free the malloced space
	*/
	free(buf);
}


/*
 * Interpret an argument as a network address of some kind,
 * returning 1 if a host address, 0 if a network address.
 */
static int
getaddr(int which, struct in_addr addr)
{
	sup su;

	switch (which) {
	case RTA_DST:
		su = &so_dst;
		break;
	case RTA_GATEWAY:
		su = &so_gate;
		break;
	case RTA_NETMASK:
		su = &so_mask;
		break;
	case RTA_GENMASK:
		su = &so_genmask;
		break;
	case RTA_IFP:
		su = &so_ifp;
		break;
	case RTA_IFA:
		su = &so_ifa;
		break;
	default:
		/*NOTREACHED*/
	}
	su->sa.sa_len = sizeof(struct sockaddr_in);
	su->sa.sa_family = AF_INET;

	if (addr.s_addr == 0) {
		switch (which) {
		case RTA_DST:
			(void) getaddr(RTA_NETMASK, addr);
			break;
		case RTA_NETMASK:
		case RTA_GENMASK:
			su->sa.sa_len = 0;
		}
		return (0);
	}

	su->sin.sin_addr.s_addr = addr.s_addr;

	return 1;
}

/*
 *  see if the destHost is contained in route table, and return nextHop.
 */
int route_table_contains(struct in_addr destHost, struct in_addr *nextHop)
{
	initaddrs();
	getaddr(RTA_DST, destHost);

	if (rtmsg('g', 0, RTA_DST,0) < 0) {
		return 0;
	} else {
		if (so_dst.sin.sin_addr.s_addr==destHost.s_addr) {
			nextHop->s_addr = so_gate.sin.sin_addr.s_addr;
			return 1;
		} else /* may be default or clone route */ {
			return 0;
		}
	}
}

int is_link_route(struct in_addr destHost) {
	initaddrs();
	getaddr(RTA_DST, destHost);
	return(rtmsg('g', 0, RTA_DST, 0) >= 0
			&& (m_rtmsg.m_rtm.rtm_flags & RTF_LLINFO));
}

int is_host_route(struct in_addr destHost, struct in_addr * nextHop) {
	initaddrs();
	getaddr(RTA_DST, destHost);
	if (rtmsg('g', 0, RTA_DST,0) >= 0){
		if ( (m_rtmsg.m_rtm.rtm_flags & RTF_GATEWAY) &&
		     (m_rtmsg.m_rtm.rtm_flags & RTF_HOST)  ){
			nextHop->s_addr=so_gate.sin.sin_addr.s_addr;
			return 1;
		} else {
			printf("flag of %s is %x\n", inet_ntoa(destHost),
				m_rtmsg.m_rtm.rtm_flags);
		}
	}
	return 0;
}

/*
 * write a routing message.
 *
 * cmd - 'a' - ADD
 *	 'c' - CHANGE		(to modify route SA's)
 *	 'd' - DELETE
 *	 'g' - GET
 *
 * errors:
 *	EEXIST - entry exists
 *	ESRCH - request to delete non-existance entry
 *	ENOBUFS - no buffers in kernel
 *	EIPSEC - some problem with SA creation, deletion or modification
 *		 (only with -DIPSEC.)
 *
 * flags
 *     RTF_STATIC
 *     RTF_HOST
 *     RTF_GATEWAY
 *     RTF_REJECT
 *     RTF_BLACKHOLE
 *     RTF_TUNNEL
 *				(ipsec flags next)
 *     RTF_AUTH
 *     RTF_CRYPT
 *     RTF_AHTUNNEL
 *     RTF_ESPTUNNEL
 *
 * returns:
 *	-1: failed
 *	 0: ok
 */
static int
rtmsg(int cmd, int flags, int rtm_addrs, int arpflag)
{
	static int seq;
	int rlen;
	char *cp = m_rtmsg.m_space;
	int l;

	errno = 0;
	bzero((char *)&m_rtmsg, sizeof(m_rtmsg));

	if (cmd == 'a') {
		cmd = RTM_ADD;
#ifdef IPSEC
	} else if (cmd == 'c' ) {
		cmd = RTM_CHANGE;
#endif
	} else if (cmd == 'g') {
		cmd = RTM_GET;
	} else {
		cmd = RTM_DELETE;
	}
/* set the route flags ACCORDING to the type of route
*/
#define rtm m_rtmsg.m_rtm
	rtm.rtm_type = cmd;
	rtm.rtm_flags = flags;
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_seq = ++seq;
	rtm.rtm_addrs = rtm_addrs;
	rtm.rtm_rmx = rt_metrics;
	rtm.rtm_inits = rtm_inits;

	if (rtm_addrs & RTA_NETMASK)
		mask_addr(rtm_addrs);
	/* NEXTADDR tests the rtm_addrs variable
	 * to see if a flag is set.  if so, it puts
	 * the socket info (2nd param) into the routing message
	 */
	if (!arpflag) {
		NEXTADDR(RTA_DST, so_dst);
		NEXTADDR(RTA_GATEWAY, so_gate);
	}
	else {
		NEXTADDR(RTA_DST, arp_sin_m);
		NEXTADDR(RTA_GATEWAY, arp_sdl_m);
	}
	NEXTADDR(RTA_NETMASK, so_mask);
	NEXTADDR(RTA_GENMASK, so_genmask);
	NEXTADDR(RTA_IFP, so_ifp);
	NEXTADDR(RTA_IFA, so_ifa);
#ifdef IPSEC
	NEXTADDR(RTA_AH, so_ah);
	NEXTADDR(RTA_ESP, so_esp);
	NEXTADDR(RTA_AHTUNNEL, so_ahtunnel);
	NEXTADDR(RTA_ESPTUNNEL, so_esptunnel);
#endif
	/*
	 * XXX
	 * Should not be necessary; NEXTADDR keeps l up-to-date.
	 */
	rtm.rtm_msglen = l = cp - (char *)&m_rtmsg;

#if 0
	if (verbose)
		print_rtmsg(&rtm, l);
	if (debugonly)
		return (0);
#endif

	if (cmd == RTM_GET) {
		routeLoopback(1);
	}
	if ((rlen = write(routesocket, (char *)&m_rtmsg, l)) < 0) {
		syslog(LOG_ERR, "writing to routing socket: %m");
		return (-1);
	}
	if (cmd == RTM_GET) {
		int pid = getpid();
		
		for (;;) {
			l =  read(routesocket, (char *)&m_rtmsg, sizeof(m_rtmsg));
			if (l <= 0 || (rtm.rtm_seq == seq && rtm.rtm_pid == pid)) {
				routeLoopback(0);
				if (l < 0) {
					(void) syslog(LOG_ERR,
					  	  "route: read from routing socket: %s\n",
						    strerror(errno));
					return(l);
				}
				break;
			}
  			/*
			 * XXX
			 * Race condition!  Kernel may be stuffing something at
			 * us that we merrily throw overboard because it's not
			 * a response to the get.
			 *
			 * Fix?  Maybe we need to open another routing socket
			 * when invoking RTM_GET (uck).  Log it for now.
			 */
			mip_log(MLOG_ERR, "rtmsg dropped a packet of type %d\n",
					rtm.rtm_type);
		}
	}
#undef rtm
	return (0);
}

static void
mask_addr(int rtm_addrs)
{
	int olen = so_mask.sa.sa_len;
	char *cp1 = olen + (char *)&so_mask;

	for (so_mask.sa.sa_len = 0; cp1 > (char *)&so_mask; )
		if (*--cp1 != 0) {
			so_mask.sa.sa_len = 1 + cp1 - (char *)&so_mask;
			break;
		}
	if ((rtm_addrs & RTA_DST) == 0)
		return;
}


#ifdef TESTMAIN

/* stub function. put in mnd/mipd config.c files
*/
int noDefaultroute(){
	extern struct mn_state mns;
	if(mns.state==NOWHERE) {
		return TRUE;
	} else {
		/* should have default route */
		return FALSE;
	}
}

/*
 * TBD stub:
 * move to *config.c
 *
 * map ifname to associated index and ip addr
 */
void
mapIfname(char *ifname, int *index, unsigned long *ifaddr)
{
	*index = 6;
	/* *index = 7;	/* */

	*ifaddr = inet_addr("131.254.50.19");	/* MN to irisa host */
	/* *ifaddr = inet_addr("204.203.70.114"); /* MN to (new) bridgeport */
	/* *ifaddr = inet_addr("204.203.67.178");	/* MN to bridgeport */
	/* *ifaddr = inet_addr("204.203.67.213");	/* MN to guinness */
}

main()
{
	struct in_addr mn, coa, interface, ipsec_interface, bridgeport;
	struct in_addr clone, mask;
	char mac[6];
#ifdef ROUTESA
	struct routeSA	rosa, mvifrosa;
#endif

	/* set local IP addresses for testing.
	 * NEED if ip and subnet mask (clone) version of that
	 */
	interface.s_addr = inet_addr("131.254.50.19");
	ipsec_interface.s_addr = inet_addr("204.203.70.114");
	clone.s_addr = inet_addr("131.254.0.0");
	mask.s_addr = inet_addr("255.255.0.0");
	bridgeport.s_addr = inet_addr("204.203.70.97");

	/* init routing
	*/
	routeInit();

	/* test 1. add ha/tunnel route
	*/
	mn.s_addr = inet_addr("1.1.1.1");
	coa.s_addr = inet_addr("2.2.2.2");
#ifdef TUNNEL
	ha_route_add_tunnel(mn, coa);
	ha_route_delete_tunnel(mn, coa);
#endif

#ifdef HOSTROUTE
	/* test 2, add/delete host route
	*/
	route_add_host(mn, coa, interface);
	route_delete_host(mn, coa);
#endif

#ifdef DEFAULTROUTE
	/* test 3, add/delete mn default route
	*/
	mn_route_add_default(coa, interface, 0,  NULL, "wlp0");
	mn_route_delete_default(coa, 0);
#endif

#ifdef CLONEROUTE
	/* test 4. clone route add and delete.  we will delete it
	 * first (so we assume it is in).
	 */
	clone_route_add(clone, interface, mask);
	clone_route_delete(clone);
	reject_route_add(clone, interface, mask);
#endif
#ifdef LINKROUTE
	/* arp style entry - arp will fill in if on
	*/
	route_add_link(mn , NULL,  "wlp0");
	route_delete_link(mn , NULL);

	/* we fill it in ourselves
	*/
	route_add_link(mn , (char *)ether_aton("1:1:1:2:2:2"),  "wlp0");
	route_delete_link(mn , NULL);
#endif
#ifdef FLUSHARP
	flushArp();
#endif

#ifdef PROXYARP
	/*
gcc -g -o rs -DDEBUG -DPROXYARP -DIPSEC -DTESTMAIN routesock.c
	*/

	route_add_proxy(mn , (char *)ether_aton("1:1:1:2:2:2"),  "wlp0");
	route_delete_proxy(mn);

	route_add_proxy(mn , (char *)ether_aton("1:1:1:3:3:3"),  "wlp0");
	ha_route_add_tunnel(mn, coa, NULL, NULL);
	ha_route_delete_tunnel(mn, coa);
	route_delete_proxy(mn);
#endif /* PROXYARP */

#ifdef IPSEC
#ifdef ROUTESA
	/*
gcc -g -Wall -DIPSEC -DDEBUG -DROUTESA -DTESTMAIN routesock.c
	*/
	bzero(&rosa, sizeof(rosa));
	bzero(&mvifrosa, sizeof(mvifrosa));

	rosa.tup[ROSA_AH].spi = 0x400;
	rosa.tup[ROSA_AH].src.s_addr = ipsec_interface.s_addr;
	rosa.tup[ROSA_AH].dst.s_addr = bridgeport.s_addr;

	rosa.tup[ROSA_ESP].spi = 0x300;
	rosa.tup[ROSA_ESP].src.s_addr = ipsec_interface.s_addr;
	rosa.tup[ROSA_ESP].dst.s_addr = bridgeport.s_addr;

	mvifrosa.tup[ROSA_AH].spi = 0x402;
	mvifrosa.tup[ROSA_AH].src.s_addr = ipsec_interface.s_addr;
	mvifrosa.tup[ROSA_AH].dst.s_addr = bridgeport.s_addr;

	mvifrosa.tup[ROSA_ESP].spi = 0x302;
	mvifrosa.tup[ROSA_ESP].src.s_addr = ipsec_interface.s_addr;
	mvifrosa.tup[ROSA_ESP].dst.s_addr = bridgeport.s_addr;

	/* ------------------------------------ */	/* add tunnel */
	mn.s_addr = inet_addr("40.0.0.1");
	ha_route_delete_tunnel(mn, coa);
	ha_route_add_tunnel(mn, coa, NULL, NULL);

	rosa.transform = (AHSEC|ESPSEC);	/* */
	mvifrosa.transform = AHSEC;	/* */
	mn.s_addr = inet_addr("40.0.0.2");
	ha_route_delete_tunnel(mn, coa);
	ha_route_add_tunnel(mn, coa, &rosa, &mvifrosa);

	/* ------------------------------------ */	/* add host */
	mn.s_addr = inet_addr("10.0.0.1");
	route_delete_host(mn, coa);
	route_add_host(mn, coa, interface, NULL);

	rosa.transform = (AHSEC);	/* */
	mn.s_addr = inet_addr("10.0.0.2");
	route_delete_host(mn, coa);
	route_add_host(mn, coa, interface, &rosa);

	rosa.transform = (AHSEC|ESPSEC);	/* */
	mn.s_addr = inet_addr("10.0.0.3");
	route_delete_host(mn, coa);
	route_add_host(mn, coa, interface, &rosa);

	/* ------------------------------------ */	/* add link */
	mn.s_addr = inet_addr("20.0.0.1");
	route_delete_link(mn, NULL);
	mac[0] = mac[2] = mac[4] = 0xa;
	mac[1] = mac[3] = mac[5] = 0;
	route_add_link(mn, mac, "wlp0", NULL);

	rosa.transform = (ESPSEC);	/* */
	mn.s_addr = inet_addr("20.0.0.2");
	route_delete_link(mn, NULL);
	mac[0] = mac[2] = mac[4] = 0;
	mac[1] = mac[3] = mac[5] = 0xb;
	route_add_link(mn, mac, "wlp0", &rosa);

	rosa.transform = (AHSEC|ESPSEC);	/* */
	mn.s_addr = inet_addr("20.0.0.3");
	route_delete_link(mn, NULL);
	mac[0] = mac[2] = mac[4] = 0xc;
	mac[1] = mac[3] = mac[5] = 0;
	route_add_link(mn, mac, "wlp0", &rosa);

	/* ------------------------------------ */	/* add default */
	mn_route_delete_default(bridgeport, 1);
	mn_route_add_default(bridgeport, interface, 0, NULL, "wlp0", NULL, 0);

	rosa.transform = (AHSEC);	/* */
	mn_route_delete_default(bridgeport, 1);
	mac[0] = mac[2] = mac[4] = 0;
	mac[1] = mac[3] = mac[5] = 0xd;
	mn_route_add_default(bridgeport, interface, 1, mac, "wlp0", &rosa, 1);

	rosa.transform = (AHSEC|ESPSEC);	/* */
	mn_route_delete_default(bridgeport, 1);
	mac[0] = mac[2] = mac[4] = 0xe;
	mac[1] = mac[3] = mac[5] = 0;
	mn_route_add_default(bridgeport, interface, 1, mac, "wlp0", &rosa, 0);

	mvifrosa.transform = (AHSEC|ESPSEC);	/* */
	mn_route_delete_default(bridgeport, 1);
	mac[0] = mac[2] = mac[4] = 0;
	mac[1] = mac[3] = mac[5] = 0xf;
	mn_route_add_default(bridgeport, interface, 1,
		mac, "wlp0", &mvifrosa, 1);

#endif	/* ROUTESA */
#endif	/* IPSEC */
	return 0;
}
#endif /* TESTMAIN */
