/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * icmpsend.c
 *
 * Used to send icmp advertisement messages.
 * Note that we have to create an icmp header here.
 *
 * functions:
 *	sendBeacon - send icmp router advertisement
 *	static in_cksum - compute ip csum for icmp header
 *	getSequence - compute next sequence number for router advertisement
 */
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
/*#include <netinet/ip_var.h> */
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

#include "mip_icmp.h"
#include "mip_log.h"
#include "icmpstats.h"
#ifdef IPSEC
#include "keylist.h"
#include "route.h"
#endif
#ifdef MND
#include "alist.h"
#include "mnstate.h"
#endif

/*#define BJORN_DEBUG*/

/* prototypes
*/ 
static 	in_cksum(u_short *, int);
unsigned short getSequence(u_short *);
char 	*configLocation();
unsigned char *configIcmpAuth(int *);
int 	getBeaconRate();
char	*getNetauthKey();
char	*getAdhocKey();

extern int beaconSd;
static int resendTimer;

static int current_lifetime;


/*
 * sendBeacon --
 *
 * See RFC 1256 for ICMP router adverts, etc.
 */

void
sendBeacon(router, dst, coa, mse_lifetime, identity, bothAgents, seq, mac_addr)
struct sockaddr_in *router;	/* router in ADVERT packet */
struct sockaddr_in *dst;        /* local ip dest for packet */
struct sockaddr_in *coa;	/* COA for FA */
int mse_lifetime;		/* lifetime to mobile host */
int identity;			/* true if fa, else ha */
int bothAgents;			/* TRUE if both ha AND fa */
u_short seq;
u_char *mac_addr;
{
	register struct icmp_advert *icp;
	register struct mip_mseicmp *mse;
	register struct mip_mse_icmp_mac *msemac;
	register struct mip_mse_icmp_info *mseinfo;
	register struct mip_mse_icmp_auth *mseauth;
	int rc;
	extern int debug;
	int offset=0;
	u_short flags = 0;
	u_char outpack[MAXICMPPACKET];
	char *location;
	u_char *key;
	int spi;
	int infosize = 0;
	char ifname[MAXIFNAME];
	int doArp;

#ifdef BJORN_DEBUG
	int sd;
	int on;
	struct protoent *proto;

	on = 1;

	fprintf( stderr, "BJORN_DEBUG, about to create my own socket\n");
	proto = getprotobyname("icmp");
	sd = socket(AF_INET, SOCK_RAW, proto->p_proto );
	if ( sd < 0 ){
	        perror( "socket failed");
		exit(1);
	}
	rc = setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (char*)&on,
                sizeof(on));
        if (rc < 0){
                perror("setsockopt SO_BROADCAST failed" );
                exit(1);
        }
	rc = setsockopt(beaconSd, SOL_SOCKET, SO_BROADCAST, (char*)&on,
                sizeof(on));
        if (rc < 0){
                perror("setsockopt SO_BROADCAST failed" );
                exit(1);
        }

	rc = sendto( sd, outpack, 122, 0, (struct sockaddr *)dst,
		sizeof(struct sockaddr));

	if ( rc < 0 ){
                perror( "BJORN_DEBUG my own sendto failed" );
                exit(1);
	}
	else{
		fprintf( stderr, "BJORN_DEBUG, my private sendto successful\n" );
	}
	close(sd);


	
	fprintf( stderr, "BJORN_DEBUG, about to send packet to %s\n",
		inet_ntoa( dst->sin_addr ));
	rc = sendto( beaconSd, outpack, 122, 0, (struct sockaddr*)dst,
		sizeof(struct sockaddr));
	if ( rc < 0 ){
		perror( "BJORN_DEBUG sendto failure" );
		exit(1);
	}
	else{
		fprintf( stderr, "BJORN_DEBUG sendto successful\n" );
	} 
#endif

	/* convert router string to ip address */

	/* pointer to packet buffer */
	icp = (struct icmp_advert *)outpack;

	/* setup icmp part of packet
	*/
	icp->icmp_type = ICMP_ROUTERADVERT;
	icp->icmp_code = 16;
	icp->icmp_cksum = 0;
	icp->icmp_naddrs = 1;
	icp->icmp_asize = 2;

	/* the "lifetime" of the advertisement
	*/
	icp->icmp_adlifetime = htons(3 * getBeaconRate());

	/* the router is most likely to be the mobile interface address
	 */
	icp->icmp_router = router->sin_addr.s_addr;
	icp->icmp_preference = htonl(0);

	/* setup mse part of packet */
	mse = (struct mip_mseicmp *) (outpack + sizeof(struct icmp_advert));
	mse->mse_type = MIPAGENTAD;
	/* note that all flags are off by default */
	if (identity == SELF_MN) {
		mse->mse_len = 6;
		SET(flags, MNBIT);
		mse->mse_flags = htons(flags);
		offset = -4;
		icmp_stats.mn_beacons_sent++;
	}
	/* set beacon fields for HA and/or FA characteristics.  if agent
	 * plays both roles, the FA characteristics will be used 
	 * as they are a superset of the HA characteristics in this context.
	 */
	else if (identity==SELF_FA || identity == SELF_HA ) {
		if (identity == SELF_HA) {
			SET(flags, HABIT);
			mse->mse_len = 6;
			offset = -4;   /* discount coa from offset size */
		}
		if (identity==SELF_FA || bothAgents) {
			SET(flags, FABIT);
			mse->mse_len = 10;   /* FA has coa */
			mse->mse_coa = coa->sin_addr.s_addr;
			offset = 0;
		}
		mse->mse_flags = htons(flags);

		if (bothAgents) {
			icmp_stats.combined_beacons_sent++;
		}
		else if (identity==SELF_HA) {
			icmp_stats.ha_beacons_sent++;
		}
		else {
			icmp_stats.fa_beacons_sent++;
		}
	}
	else {
		syslog( LOG_ERR, "sendBeacon: Don't know who I am:%d?",identity);
		exit(1);
	}
		
	/* each interface has associated sequence number 
	 */
	mse->mse_sequence = htons(seq);
	mse->mse_lifetime = htons(mse_lifetime);

	/* offset is size so far into packet (outpack) */
	offset += sizeof(struct icmp_advert) + sizeof(struct mip_mseicmp);

	/* send mac address if arp is off.
	 */
	doArp = getConfigArp(&icp->icmp_router);
	if (!doArp) {
		static i=0;
		msemac = (struct mip_mse_icmp_mac *) (outpack + offset);
		msemac->mse_type = MIPMAC;
		msemac->mse_len = sizeof(struct mip_mse_icmp_mac) - 2;
		memcpy(msemac->mse_mac, mac_addr, 6);
		/* update total size, need 2 for type/length */
		offset += sizeof(struct mip_mse_icmp_mac);
	}

	/* check for location option and if it exists, put it
	 * in the packet
	 */
	location = configLocation();
	if (location) {
		char *ptr;
		infosize = strlen(location);
		mseinfo = (struct mip_mse_icmp_info *) ((char *)outpack +
			offset);
		mseinfo->mse_type = MIPAGENTINFO;
		mseinfo->mse_len = infosize;
		/* copy the data into the packet buffer */
		ptr = ((char *)mseinfo) + 2; 
		memcpy(ptr, location, infosize);
		/* if odd pad and zero out the last byte */
		if (infosize % 2) {
			unsigned char z = 0;
			memcpy((ptr + infosize), &z, 1);
			mseinfo->mse_len++;
			infosize++;
		}
		/* update total size, need 2 for type/length */
		offset += infosize + 2;
	}

	if (key = getNetauthKey()) {
		int spi = htonl(getNetauthSpi());
		mseauth = (struct mip_mse_icmp_auth *) 
			((char *) outpack + offset);
		mseauth->mse_type = MIPAGENTAUTH;
		mseauth->mse_length = sizeof(struct mip_mse_icmp_auth) - 2;
		mseauth->flags = 0;
		SET(mseauth->flags, NETAUTH);
		mseauth->pad = 0;
		mseauth->tssecs = htonl(get_ntptime_secs());
		mseauth->tsfract = htonl(get_ntptime_fract());
#ifdef MND
		mseauth->agent = (mns.agent) ? (mns.agent->bip.s_addr) : 0;
#else /* MIPD */
		mseauth->agent = 0;
#endif
        	bcopy(&spi, &mseauth->spi, sizeof(int));
		/* do auth. over all payload up to here including
		 * first two fields of mseauth. timestamp + pad
		 * are included in hash.
		 */
 		storeMd5((char *)outpack, offset + 
			MIPAGENTAUTH_OFFSET, mseauth->hash, key);
		if (debug) {
			printf("md5 csum over %d bytes\n",
				offset + MIPAGENTAUTH_OFFSET);
		}
		offset += mseauth->mse_length + 2;
	}

	if (key = getAdhocKey()) {
		int spi = htonl(getAdhocSpi());
		mseauth = (struct mip_mse_icmp_auth *) 
			((char *) outpack + offset);
		mseauth->mse_type = MIPAGENTAUTH;
		mseauth->mse_length = sizeof(struct mip_mse_icmp_auth) - 2;
		mseauth->flags = 0;
		SET(mseauth->flags, ADHOCAUTH);
		mseauth->pad = 0;
		mseauth->tssecs = htonl(get_ntptime_secs());
		mseauth->tsfract = htonl(get_ntptime_fract());
#ifdef MND
		mseauth->agent = (mns.agent) ? (mns.agent->bip.s_addr) : 0;
#else /* MIPD */
		mseauth->agent = 0;
#endif
        	bcopy(&spi, &mseauth->spi, sizeof(int));
		/* do auth. over all payload up to here including
		 * first two fields of mseauth. timestamp + pad
		 * are included in hash.
		 */
 		storeMd5((char *)outpack, offset + 
			MIPAGENTAUTH_OFFSET, mseauth->hash, key);
		offset += mseauth->mse_length + 2;	
	}

	/* compute ICMP checksum here */
	icp->icmp_cksum = in_cksum((u_short *)icp, offset);

	/* send the beacon
	*/
	errno = 0;
	rc = sendto(beaconSd, (char *)outpack, offset, 0, 
		(struct sockaddr *)dst, sizeof(struct sockaddr));
	if (debug) {
		printf("Wrote icmp advert, %d chars, ret=%d\n", offset, rc);
		dumpme(outpack, offset); 
		printf("End of dump\n");
		printf("send packet to %s\n", 
			inet_ntoa( dst->sin_addr ));
		printf("rc, errno %d %d,%s\n", rc, errno, strerror(errno) );
	}

	if (rc < 0 || rc != offset)  {
		if (rc < 0) {
			/* if out of buffers continue, try again later
			*/ 
			if (errno == ENOBUFS)
				return;
			/* linux is not thrilled about sending ICMP
			 * router adverts ...  This is a bug that needs
			 * to be fixed, probably by pushing the socket
			 * mechanism down to include the IP hdr
			 */
			if (errno == ENETUNREACH) {
				return;
			}
		}
		/* else fatal error
		*/
		syslog( LOG_ERR, "sendBeacon(): sendto failed (%m).");
		exit(1);
	}
	if (debug){
		fprintf( stderr, "beacon sent successfully\n" );
	}

}

/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 *	(from BSD ping code)
 */
static
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

/*
 * generate next mse sequence number for router advert message
 */
unsigned short
getSequence(u_short *seq)
{
	u_short next = *seq;	/* return previous value */
	/* increment for next round
	*/
	if (next == 0xffff) {
		*seq = 256;	/* non reboot rollover */
	}
	else {
		*seq = *seq + 1;
	}
	return(next);
}
