
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <net/if.h>


#include <errno.h>
#include <signal.h>

#include "mip_log.h"

#ifdef IPSEC
#include <netsec/ipsec.h>
#include "ipsec.h"
extern int ipsecKeysInuse;
#endif

extern int debug;


void sendOneUdp(int , char *, int , struct in_addr, u_short);

/*
 * open mobile-ip udp port  
 *
 * returns:
 *	int s - udp socket descriptor
 */
int
initUdp(unsigned short udp_port)
{
	int	s,
		on = 1,
		rc;
	struct sockaddr_in server;

	/* create INTERNET, udp datagram socket
	*/
	s = socket(AF_INET, SOCK_DGRAM, 0); 
	if ( s < 0 ) {
		syslog( LOG_ERR, "initUdp(): socket (%m)");
		exit(1);
	}

	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(s, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initUdp(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;   /* ok from any machine */
	server.sin_port = htons(udp_port);     /* specific port */

	/* configure socket to sendup i/f index so we can determine
	 * interface packet came in on.  This calls requires us
	 * to use recvMsg(2)
	 */

#ifdef IPSEC
	if (ipsecKeysInuse) {
		int	ipsec_level = IPSEC_LEVEL_REQUIRE;

		if (setsockopt(s, SOL_SOCKET, SO_SECURITY_PLAINTEXT_ROUTE,
			(char *)&ipsec_level, sizeof(int)) < 0) {
			syslog(LOG_ERR, "initUdp(): sockopt SO_SECURITY_PLAINTEXT_ROUTE failed");
			if (debug) {
				printf("initUdp(): sockopt SO_SECURITY_PLAINTEXT_ROUTE failed");
			}
			exit(1);
		}
	}
#endif /* IPSEC */

	/* bind protocol to socket
	*/
	if (bind(s, (struct sockaddr *)&server, sizeof(server))) {
		syslog(LOG_ERR, "initUdp(): bind (%m)");
		exit(1);
	}
	return(s);
}

void
sendOneUdp(int s, char *buf, int size, struct in_addr whoto, u_short uport)
{
	int rc;
	struct sockaddr_in to;

	/* send the packet to the mobile udp port
	*/
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = whoto.s_addr;
	to.sin_port = uport;

	if (debug) {
		printf("sendOneUdp, dst %s, port %d\n", 
			inet_ntoa(whoto.s_addr), ntohs(uport));
		dumpme(buf, size);
	}
	rc = sendto(s, buf, size, 0,  (struct sockaddr *)
		&to, sizeof (struct sockaddr_in));
	if (rc < 0) {
		/* notice error, though it is not fatal */
		syslog( LOG_ERR, "sendOneUdp(): sendto failed (%m).");
	}
}


#ifdef TESTMAIN

/*
 * test the udp port 434 i/f hack, i.e., packets sent to
 *	port 434 have the ip address of the incoming i/f appended to
 *	them.  It is in the form of a network byte order in_addr; 
 *	i.e., a long.
 * parameters:
 *	int count - number of times to send/recv test packet
 */
testIfHack(int count)
{
	struct t {
		int x;
		int y;
		int z;
	} m;
	struct t *p;
	int rc;
	int size;
	struct sockaddr_in to;
	struct sockaddr_in from;   
	int s = initUdp(434);
	char buf[1024];
	char buf2[1024];
	int i;
	struct msghdr msg;

       /* since we are sending this to self, we don't need
        * to byte swap
	*/
	m.x = 1;
	m.y = 2;
	m.z = 3;

	/* send the packet to the  udp port 434
	*/
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = inet_addr("127.0.0.1");
	to.sin_port = htons(434);

	for (i = 0; i < count; i++) {
		struct iovec i;
		rc = sendto(s, &m, sizeof(struct t), 0,  (struct sockaddr *)
			&to, sizeof (struct sockaddr_in));
		if (rc < 0) {
			perror("sendto");
			exit(1);
		}
		/* read the packet back from the udp port 434
		*/
		size = sizeof(struct sockaddr_in);
		msg.msg_name = (caddr_t) &from;
		msg.msg_namelen = size;
		i.iov_base = buf;
		i.iov_len = 1024;
		msg.msg_iov = &i;
		msg.msg_iovlen = 1;
		msg.msg_control = buf2;
		msg.msg_controllen = 1024;
#ifdef SAVEENERGY
		rc = RECVMSG(s, &msg, 0);
#else
		rc = recvmsg(s, &msg, 0);
#endif
		if (rc < 0) {
			perror("recvmsg");
			exit(1);
		}
		{
			struct cmsghdr *cmsg;
			struct sockaddr_dl *sdl;
			cmsg = (struct cmsghdr *) msg.msg_control;
			if (cmsg->cmsg_len != 20 
			  && cmsg->cmsg_type != 20) {
				printf("cmsg failed\n");
				return;
			}
			sdl = (struct sockaddr_dl *) ((char *)cmsg + 12);
			printf("family (AF_LINK) %x  link %d\n", 
				sdl->sdl_family, sdl->sdl_index);
		}
		printf("addr=%x, port=%d, rc=%d\n", from.sin_addr.s_addr,
			ntohs(from.sin_port), rc);
		p = (struct t *) buf;
		printf("x %d, y %d, z %d\n", p->x, p->y, p->z);
		if (rc > sizeof(struct t)) {
			struct in_addr s;
			u_long l;
			char *pt = buf + sizeof(struct t);
			l = * ((u_long *)pt);
			s.s_addr = l;
			printf("MSG GREW: if addr %x\n", s.s_addr);
		}
	}
}

/* write a MOCK registration packet to a given ip
 * address to port 434
 */
sendOneUdp_test(char *who)
{
	struct t {
		int x;
		int y;
		int z;
	} m;
	struct t *p;
	int rc;
	int size;
	struct sockaddr_in to;
	struct sockaddr_in from;   
	int s = initUdp(434);
	char buf[1024];
	int i;

       /* since we are sending this to self, we don't need
        * to byte swap
	*/
	m.x = 1;
	m.y = 2;
	m.z = 3;

	/* send the packet to the  udp port 434
	*/
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = inet_addr(who);
	to.sin_port = htons(434);

	rc = sendto(s, &m, sizeof(struct t), 0,  (struct sockaddr *)
		&to, sizeof (struct sockaddr_in));
	if (rc < 0) {
		perror("sendto");
		exit(1);
	}
	close(s);
}

recvOneUdp_test()
{
	struct t {
		int x;
		int y;
		int z;
	} m;
	struct t *p;
	int rc;
	int size;
	struct sockaddr_in to;
	struct sockaddr_in from;   
	int s = initUdp(434);
	char buf[1024];
	int i;

	/* read the packet back from the udp port 434
	*/
	size = sizeof(struct sockaddr_in);
#ifdef SAVEENERGY
	rc = RECVFROM(s, buf, 1024, 0, (struct sockaddr *)&from, &size);
#else
	rc = recvfrom(s, buf, 1024, 0, (struct sockaddr *)&from, &size);
#endif
	if (rc < 0) {
		perror("sendto");
		exit(1);
	}
	printf("addr=%x, port=%d, rc(size)=%d\n", from.sin_addr.s_addr,
		ntohs(from.sin_port), rc);
	p = (struct t *) buf;
	printf("x %d, y %d, z %d\n", p->x, p->y, p->z);
	if (rc > sizeof(struct t)) {
		struct in_addr s;
		u_long l;
		char *pt = buf + sizeof(struct t);
		l = * ((u_long *)pt);
		s.s_addr = l;
		printf("MSG GREW: if addr %x\n", s.s_addr);
	}
}


/*
 * test writing a packet to port 434.  
 * The kernel will mysteriously grow the packet
 * and slap the input i/f ip address on the end
 */
main(int argc, char **argv)
{
	int count;
#ifdef TESTIFHACK
	if (argc != 2) {
		return;
	}
	count = atoi(argv[1]);
	testIfHack(count);
#else 
	if (strcmp(argv[1], "-s") == 0) {
		sendOneUdp_test(argv[2]); 		
	}
	else if (strcmp(argv[1], "-r") == 0) {
		recvOneUdp_test();
	}
#endif
}
#endif

