/*
 * Copyright (c) 1996,1997,1999 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/ethernet.h>
#include <linux/if.h>
/* #include <linux/if_arp.h> */
#include <linux/if_ether.h>
#include <linux/ipx.h>
#include <linux/wireless.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "mip_log.h"
#ifndef TESTMAIN
#include "mnconfig.h"
#endif

#define WAVELAN_ADDR_SIZE 6 


/* holds mac addresses of interest, sorted by most recent reference */ 
static u_char spy_address[ IW_MAX_SPY ][ sizeof(struct sockaddr) ];
static int num_spys;  /* number of valid addresses in spy_address[] */
#ifdef TESTMAIN
static char interface[]="eth0";
static int ifsize = 5;
int debug = 1;
#else
static char *interface = mn_config.ifname;
static int ifsize = MAXIFNAME;
extern int debug;
#endif

#define OLD_WAVELAN	1
#define IEEE_LUCENT     2
#define DONTKNOW	-1

/*
 * zot knows this is a whacky thing to do.  basic idea
 * is to use range in wireless extensions and try to *guess*
 * what kind of wireless i/f we have.
 * 
 * 1: old wavelan
 * 2: IEEE/lucent wavelan
 * -1: i don't know
 *
 * note: do NOT close the descriptor in here.
 */
static int
wl_determine_if_type(int sd, char *ifname)
{
	struct iwreq          wrq;
	float                 freq;
	struct iw_range       range;
	int                   k;

	strncpy(wrq.ifr_name, ifname, IFNAMSIZ);
	wrq.u.data.pointer = (caddr_t) &range;
	wrq.u.data.length = 0;
	wrq.u.data.flags = 0;
	if(ioctl(sd, SIOCGIWRANGE, &wrq) < 0) {
		return(DONTKNOW);
	}
	switch( range.max_qual.qual ) {
	case 15:  /* old wavelan */
		return(OLD_WAVELAN);
		break;
	case 92:  /* IEEE Lucent */
		return(IEEE_LUCENT);
		break;
	}

	return(DONTKNOW);
}

static int 
wl_mac_to_signal( int sd, int if_type, struct sockaddr *mac)
{
	int rc;
	int i;
	char buffer[ (sizeof(struct iw_quality) + sizeof(struct sockaddr))
			* IW_MAX_SPY ]; 
	struct ifreq ifr;
	struct iwreq wrq;
	struct iw_quality *qual;
	struct sockaddr *hwa;
	
	/* read signal strength cache.  Note: we need
	 * the name of the interface to do this
	 * The interface name must be correct 
	 * in /etc/mnd.conf
	 */
	strncpy( wrq.ifr_ifrn.ifrn_name, interface, ifsize );
	wrq.u.data.pointer = (caddr_t)buffer;
	rc = ioctl( sd, SIOCGIWSPY, &wrq );
	if ( rc < 0 ){
		static int static_flag = 0;
		/* just log this error once.  It is not a fatal
		 * error
		 */
		if (static_flag == 0) {
			static_flag = 1;
			mip_log( MLOG_ERR, "SIOCGIWSPY ioctl errno=%s, - no signal strength available on interface %s", sys_errlist[errno], interface);
		}
		return (-1);
	}
	num_spys = wrq.u.data.length;
	qual = (struct iw_quality*)(buffer+(sizeof(struct sockaddr)*num_spys));
	hwa = (struct sockaddr*)buffer;

	/* check if the mac we want is in the cache, if so return
	 * the signal strength 
	 */
	for ( i=0; i< num_spys; i++ ){
		if ( !memcmp(mac->sa_data, hwa[i].sa_data,
			WAVELAN_ADDR_SIZE )){
				if ( debug ){
					fprintf( stderr, "mac found in cache,
						qual= %d, level= %d, 
						noise= %d\n", qual[i].qual,
						qual[i].level, qual[i].noise );
				}
				/* quality is better than signal strength
				 * if it varies over a "bigger" threshold
				 * this is a whacky theory.
				 */
				if (if_type == IEEE_LUCENT) {
					return (qual[i].qual);
				}
				return (qual[i].level);
		}
	}
	/* mac not found */
	if ( num_spys < IW_MAX_SPY ){
		if (debug){
			fprintf( stderr, "mac not found, adding to cache\n" );
		}
		memcpy((char*)hwa[num_spys].sa_data, (char*)mac->sa_data, 
			WAVELAN_ADDR_SIZE );
		wrq.u.data.length = num_spys + 1;
		rc = ioctl( sd, SIOCSIWSPY, &wrq );
		if ( rc < 0 ){
			mip_log( MLOG_ERR, "wl_mac_to_signal(): ioctl:%s 
				failure while adding mac", sys_errlist[errno] );
			return (-1);
		}

	}
	else {	/* not enough space in cache, slide cache down to make space */
		if ( debug ){
			fprintf( stderr, "mac not found, full cache. Tossing
				%02X:%02X:%02X:%02X:%02X:%02X\n",
				(u_char)hwa[7].sa_data[0],
				(u_char)hwa[7].sa_data[1],
				(u_char)hwa[7].sa_data[2],
				(u_char)hwa[7].sa_data[3],
				(u_char)hwa[7].sa_data[4],
				(u_char)hwa[7].sa_data[5] );
		}
		memmove( buffer+sizeof(struct sockaddr), buffer, 
			sizeof(struct sockaddr) * (IW_MAX_SPY-1));
		memcpy(hwa[0].sa_data, (char*)mac->sa_data, WAVELAN_ADDR_SIZE );
		rc = ioctl( sd, SIOCSIWSPY, &wrq );
		if ( rc < 0 ){
			mip_log( MLOG_ERR, "wl_mac_to_signal(): ioctl:%s
				failure while adding mac", sys_errlist[errno] );
			return (-1);
		}
	}
	return (-1);
}

static int 
get_mac_addr (unsigned long int ipa, unsigned short *mac)
{
        FILE *fp;
        char line[200];
        char ip[100];
        int type;
        int flags;
        char hwa[100];
        char junk[100];
        struct ether_addr *ea;
        unsigned long int ipaddr;
        int i;

        if ( (fp = fopen("/proc/net/arp", "r")) == (FILE *) NULL ) {
                return ( -1 );
        }

        if ( fgets( line, sizeof(char) * 200, fp ) != (char *) NULL ) {
                for ( ; fgets( line, sizeof(char) * 200, fp ); ) {
                        (void) sscanf( line, "%s 0x%x 0x%x %s %s\n",
                                        ip, &type, &flags, hwa, junk);
                        ipaddr = inet_addr( ip );
                        if ( ipaddr == ipa ) {
                                ea = ether_aton( hwa );
                                for ( i = 0; i < 6; i++ ) {
                                        mac[i] = ea->ether_addr_octet[i];
                                }
                                fclose( fp );
                                return ( 0 );
                        }
                }
        }
        fclose( fp );
        return ( -1 );
}

/*
 * given an ip address (for an agent or mn) we want
 * to ask the wireless driver for a signal strength
 *
 * we need to map the ip address to a MAC address
 * How that is done may be interesting ... current assumption
 * is that /proc/net/arp exists  (try cat /proc/net/arp).
 * There is also an ioctl here that might be used, but the
 * calling sequence for it should be checked.
 */

int wl_ip_to_signal( struct in_addr bip)
{
	int sd;
	int rc;
	int sig_strength;
	struct sockaddr host;
	struct arpreq arpr;
	struct sockaddr_in *sin;
        unsigned short mac[ 6 ];
        int i;
	int if_type;  /* radio interface type */

	sd = socket( AF_INET, SOCK_DGRAM, 0 );
	if ( sd < 0 ){
		mip_log( MLOG_ERR, "wl_ip_to_signal(): socket:%s failure",
			sys_errlist[errno] );
		exit(1);
	}
	sin = (struct sockaddr_in*)&arpr.arp_pa;
	sin->sin_family = AF_INET;
	sin->sin_port = 0;
	sin->sin_addr.s_addr = bip.s_addr;

        if ( get_mac_addr( bip.s_addr, mac ) != 0 ) {
                mip_log( MLOG_ERR, "get_mac_addr() failed - /proc/net/arp doesn't exist?!" );
                close( sd );
                return (100);
        }
        arpr.arp_flags = 0;
        for ( i = 0; i < 6; i++ ) {
                arpr.arp_ha.sa_data[i] = mac[i];
        }

#ifdef OLDWAY
	/*  this mechanism may exist, but the calling sequence
	 *  should be rechecked, as it is presumed to be buggy
	 *  e.g., it may be necessary to specific the interface to
	 *  get a single arp address.
	 */
	rc = ioctl( sd, OLD_SIOCGARP, &arpr );
	if ( rc < 0 ){
		mip_log( MLOG_ERR, "wl_mac_to_signal(): ioctl failed:%s arp entry
			not found", sys_errlist[errno] );
		close(sd); 
		return (100);  /* we say 100% when we don't know */
	}
#endif
	if ( debug ){
		fprintf( stderr, "converted to mac= %02X:%02X:%02X:%02X:%02X:%02X\n",
			(u_char)arpr.arp_ha.sa_data[0],
			(u_char)arpr.arp_ha.sa_data[1],
			(u_char)arpr.arp_ha.sa_data[2],
			(u_char)arpr.arp_ha.sa_data[3],
			(u_char)arpr.arp_ha.sa_data[4],
			(u_char)arpr.arp_ha.sa_data[5] );
	}
	/* guess as to interface, type
	*/
	if_type = wl_determine_if_type(sd, interface);
	/* if IEEE_LUCENT, we are actually using quality,
	 * else signal strength
	 */
	sig_strength = wl_mac_to_signal(sd, if_type,  &arpr.arp_ha);

	close(sd);

	if ( debug ){
		fprintf( stderr, "sig strength=%d, iftype=%d \n", 
			sig_strength, if_type );
	}
	/* if call fails, we assume 100%.  This
	 * allows us to deal with ethernet interfaces or
	 * wireless interfaces that lack signal strength.
 	 * mnd will just fall back to a starvation-based
	 * handoff mechanism (1st beacon wins).
	 */
	if ( sig_strength == -1 ){  
		return 100;
	}

	/* ok, we have i/f that has signal strength
	 * we round up in case of error or negative value
	 * which can happen to "quality"
	 */
	if (sig_strength < 5 ){
		return 1;
	}

	/* smooth the magic number if over a certain
	 * value to PERFECT, else we leave it alone.
	 */
	if (if_type == IEEE_LUCENT) {
		if (sig_strength > 60)
			sig_strength = 100;
	}
	else if (sig_strength > 40) {
		sig_strength = 100;
	}
	return(sig_strength);
}

		
#ifdef TESTMAIN
main( int argc, char **argv)
{
	int sig_strength;
	struct in_addr agent_ip;
	
	inet_aton((const char*)argv[1], &agent_ip );
	printf( "seeking arp address for %s\n", inet_ntoa(agent_ip)); 

	sig_strength = wl_ip_to_signal(agent_ip);
	printf( "signal strength for host= %d\n", sig_strength );
}
#endif
		
